%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%   $RCSfile: diff.pl,v $
%%  $Revision: 1.19 $
%%      $Date: 2003/05/12 19:56:20 $
%%     Author: Detmar Meurers (dm@ling.osu.edu)
%%    Purpose: Code for computing a diff of two feature structures
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% $Log: diff.pl,v $
%% Revision 1.19  2003/05/12 19:56:20  dm
%% added noadderrs around add_to/3 call in get_diff_input/6
%% and added real header to file
%%

:- use_module(library(lists),[select/3]). % and nonmember/2 from sics.pl
:- use_module(library(avl),[avl_to_list/2,avl_store/4,avl_delete/4,
                            empty_avl/1,avl_fetch/3]).

% ------------------------------------------------------
% top-level predicate, using results saved using save_results_on
% as defined in save_results.pl

diff(NrOrDescA,NrOrDescB) :-
   diff(NrOrDescA,[],NrOrDescB,[]).

% Paths can be entered either as f:g:h:i or as [f,g,h,i]
% This does not take into account potential residues yet!
diff(NrOrDescA,PathA,NrOrDescB,PathB) :-
   get_diff_input(NrOrDescA,PathA,FSA,_ResidueA,DupsA,PathsA),
   get_diff_input(NrOrDescB,PathB,FSB,_ResidueB,DupsB,PathsB),
   get_diff(FSA,FSB,IdentA,IdentB,UnfA,UnfB,DupsA,DupsB,PathsA,PathsB),
   pref_struc_on,   % we want to see diffs as structures (first) not as trees
   set_title(diff:NrOrDescA),
   quiet_save_results_off,	% don't save again what we've already got
   grisu_pp_fs(FSA,flagAss(IdentA,UnfA),0),  % GBP - can we find a better MGType here than 0?
   set_title(diff:NrOrDescB),
   grisu_pp_fs(FSB,flagAss(IdentB,UnfB),0),
   save_results_restore,	% switch back to previous value
   clear_title,
   pref_struc_restore.

% --------------------------------
% auxiliary:

get_diff_input(NrOrDesc,Path,FS,Residue,Dups,Paths) :-
   (   number(NrOrDesc)
   ->  saved_result(NrOrDesc,Desc)
   ;   Desc=NrOrDesc),
   call_residue((secret_noadderrs_toggle(Mode),
		 add_to(Desc,NewFS),
		 % deref(Tag,bot,Ref0,SVs0),
		 % extensionalise(Ref0,SVs0),
		 secret_adderrs_toggle(Mode)),
		Residue),
   path_as_list(Path,PathList),
   deref(NewFS,TFS,Type,AttPos),
   pathval(PathList,NewFS,AttPos,Type,TFS,FS,bot),
   sharing_paths_assocs(FS,Dups,Paths).

path_as_list(Path,List) :-
   (  is_list(Path)
   -> List=Path
   ;  path_list(Path,List) ).
   
path_list((H:R),[H|T]):- !,
   path_list(R,T).
path_list(X,[X]).

% --------------------------------
% get_diff/4: the main predicate doing the work
%

get_diff(FSA,FSB,IdentA,IdentB,UnfA,UnfB,DupsA,DupsB,PathsA,PathsB) :-
   empty_avl(Empty),
   diff_fs(FSA,FSB,Empty,_Vis,
           Empty,IdentA,Empty,IdentB,
           Empty,UnfA,Empty,UnfB,
           [],_,
           DupsA,DupsB,PathsA,PathsB).

diff_fs(FSA,FSB,Vis0,Vis,
	IdentA0,IdentA,IdentB0,IdentB,
        UnfA0,UnfA,UnfB0,UnfB,
	Diff0,Diff,
        DupsA,DupsB,PathsA,PathsB) :-
   % deref(FSA,RefA,SVsA),
   % deref(FSB,RefB,SVsB),
   (   avl_fetch(FSA,Vis0,Diff0-Diff1) % Seen this node before?
   ->  % must be a tag:
       Vis=Vis0,IdentA1=IdentA0,IdentB1=IdentB0,
       UnfA1=UnfA0,UnfB1=UnfB0,Diff=Diff1
   ;   % unseen node (struc or tag):
       avl_store(FSA,Vis0,X-X,Vis1),      % remember seeing this node
       fs2keyed_feats(FSA,KFeatsA,TypeA), % get features+type of node
       fs2keyed_feats(FSB,KFeatsB,TypeB), %         ""
       diff_type(TypeA,TypeB,Diff0,Diff1),
       % check sharings if node is a tag in at least one of the structures:
       (   avl_fetch(FSA,DupsA,SharingPathsA) 
       ->  diff_reentrancy(SharingPathsA,DupsB,PathsB,Diff1,Diff2)
       ;   (   avl_fetch(FSB,DupsB,SharingPathsB)
	   ->  diff_reentrancy(SharingPathsB,DupsA,PathsA,Diff1,Diff2)
	   ;   Diff2=Diff1
	   )
       ),
       diff_fs_feats(KFeatsA,KFeatsB,Vis1,Vis2,
                     IdentA0,IdentA1,IdentB0,IdentB1,
                     UnfA0,UnfA1,UnfB0,UnfB1,
                     Diff2,Diff,
                     DupsA,DupsB,PathsA,PathsB),
       remember_tag_diff(FSA,DupsA,Diff0,Diff,Vis2,Vis)
   ),
   % If something is locally different (i.e. Diff0 =/= Diff1?)
   % we need to show this node in red. IdentA/B contain all other nodes.
   mark_identical(Diff0,Diff2,FSA,IdentA1,IdentA,
		              FSB,IdentB1,IdentB),
   % If something that's part of this structure or its substructure
   % changed (i.e., Diff0 =/= Diff) then we need to show this substructure
   mark_identical(Diff0,Diff,FSA,UnfA1,UnfA,
                             FSB,UnfB1,UnfB).

% GBP - ported; was svs2keyed_feats/3
fs2keyed_feats(FS,KeyedFeats,Type) :-
  get_type(FS,Type),
  approps(Type,FRs,_),
  build_keyed_feats(FRs,FS,KeyedFeats).

% Remember if this is a tag and inside not everything identical:
remember_tag_diff(FSA,DupsA,Diff0,Diff,Vis0,Vis) :-
   avl_fetch(FSA,DupsA,_), !,
   (   Diff0==Diff
   ->  Vis=Vis0
   ;   avl_delete(FSA,Vis0,_,Vis1),
       avl_store(FSA,Vis1,X-[diff_below_tag|X],Vis)
   ).
remember_tag_diff(_,_,_,_,Vis,Vis).
   
diff_reentrancy(SharingPathsX,DupsY,PathsY,Diff0,Diff) :-
  % SharingPathsX = list of paths that this node structure shares with
  % Pick one, to get its tag in structure Y
   SharingPathsX=[SharingPath|_], 
   (  (avl_fetch(SharingPath,PathsY,SharingRefY), 
       avl_fetch(SharingRefY,DupsY,SharingPathsY))
   -> true
   ;  SharingPathsY=[] ),
   % Iff the same directories share in X and Y, it's no difference
   (   SharingPathsX=SharingPathsY 
   ->  Diff=Diff0
   ;   %format(user_output,'~p ~p~n',[SharingPathsX,SharingPathsY]),
       Diff=[structure_sharing|Diff0]
   ).
   
mark_identical(Old,New,FSA,AssocA0,AssocA,
	       FSB,AssocB0,AssocB) :-
   (   Old==New
   ->  avl_store(FSA,AssocA0,_,AssocA),
       avl_store(FSB,AssocB0,_,AssocB) 
   ;   AssocA=AssocA0,AssocB=AssocB0
   ).

diff_type(TypeA,TypeB,Diff0,Diff) :-
   (  TypeA==TypeB              
   -> Diff=Diff0                
   ;  Diff=[types|Diff0]         % A and B have different types
   ).                           
 
diff_fs_feats([],KeyedFeatsB,Vis,Vis,
             IdentA,IdentA,IdentB,IdentB,
             UnfA,UnfA,UnfB,UnfB,
             Diff0,Diff,
             _DupsA,_DupsB,_PathsA,_PathsB) :-
   (  KeyedFeatsB==[]
   -> Diff=Diff0
   ;  Diff=[b_feats|Diff0]              % B has features not in A
   ).

diff_fs_feats([VA|VsA],FeatsB0,Vis0,Vis,
             IdentA0,IdentA,IdentB0,IdentB,
             UnfA0,UnfA,UnfB0,UnfB,
             Diff0,Diff,
             DupsA,DupsB,PathsA,PathsB):-
   diff_feat(VA,FeatsB0,FeatsB1,Vis0,Vis1,
             IdentA0,IdentA1,IdentB0,IdentB1,
             UnfA0,UnfA1,UnfB0,UnfB1,
             Diff0,Diff1,
             DupsA,DupsB,PathsA,PathsB),
   diff_fs_feats(VsA,FeatsB1,Vis1,Vis,
               IdentA1,IdentA,IdentB1,IdentB,
               UnfA1,UnfA,UnfB1,UnfB,
               Diff1,Diff,
               DupsA,DupsB,PathsA,PathsB).

diff_feat(fval(F,VA,_),FeatsB0,FeatsB,Vis0,Vis,
          IdentA0,IdentA,IdentB0,IdentB,
          UnfA0,UnfA,UnfB0,UnfB,
          Diff0,Diff,
          DupsA,DupsB,PathsA,PathsB) :-
   select(fval(F,VB,_),FeatsB0,FeatsB), !,
   diff_fs(VA,VB,Vis0,Vis,
           IdentA0,IdentA,IdentB0,IdentB,
           UnfA0,UnfA,UnfB0,UnfB,
           Diff0,Diff,
           DupsA,DupsB,PathsA,PathsB).

diff_feat(_,FeatsB,FeatsB,Vis,Vis,
          IdentA,IdentA,IdentB,IdentB,
          UnfA,UnfA,UnfB,UnfB,
          Diff0,[a_feats|Diff0], % A has features not in B
          _DupsA,_DupsB,_PathsA,_PathsB).

% --------------------------------------------------------------------
% Code to obtain path sets for reentrancies, i.e. this translates
% between a variable based encoding of reentrancies to an encoding
% of the paths which are equated.
%
% sharing_paths/8:
% a) for each tag collects: Tag-[path1,path2, ...]
%    where each path is the reverse list of features leading from the
%    root to the reentrancy
% b) for each reentrant path collects: path-Tag


show_sharing_paths(FS) :-
   sharing_paths_assocs(FS,Dups,_),
   avl_to_list(Dups,DupsList),
   print_result(DupsList).

sharing_paths_assocs(FS,Dups,Paths) :-
   empty_avl(Empty),
   sharing_paths(FS,Empty,Dups,Empty,_Vis,Empty,Paths,[]).

sharing_paths(FS,Dups0,Dups,Vis0,Vis,Paths0,Paths,Path):-
   % deref(FS,Ref,SVs),
   (   avl_delete(FS,Dups0,Where,Dups1) % seen this reentrancy before
   ->  avl_store(FS,Dups1,[Path|Where],Dups), % store ref->paths
       avl_store(Path,Paths0,FS,Paths),         % store path->ref
       Vis0=Vis
   ;   (   avl_fetch(FS,Vis0,PrevP) % seen this node before: aha, a reentrancy
       ->  avl_store(FS,Dups0,[Path,PrevP],Dups), % store ref->both paths
           avl_store(Path,Paths0,FS,Paths1),        % store path2->ref
           avl_store(PrevP,Paths1,FS,Paths),        % store path1->ref
           Vis0=Vis
       ;   % we never saw this node before:
           avl_store(FS,Vis0,Path,Vis1), % record that we now saw it
	   get_type(FS,Type),
	   (   (Type = (a_ _))   % Is this an atomic thing without features?
           ->  Dups0=Dups, Vis1=Vis, Paths0=Paths % yes: return 
               % no: start recursion into featurs:
           ;   approps(Type,FRs,_),
               build_keyed_feats(FRs,FS,KeyedFeats),
               sharing_paths_list(KeyedFeats,Dups0,Dups,Vis1,Vis,Paths0,Paths,Path)
         )
       )
   ).
         
sharing_paths_list([],Dups,Dups,Vis,Vis,Paths,Paths,_).
sharing_paths_list([fval(F,V,_Restr)|Vs],Dups0,Dups,Vis0,Vis,Paths0,Paths,Path):-
  sharing_paths(V,Dups0,Dups1,Vis0,Vis1,Paths0,Paths1,[F|Path]),
  sharing_paths_list(Vs,Dups1,Dups,Vis1,Vis,Paths1,Paths,Path).

print_result([]).
print_result([_Var-PathList|T]) :-
   reverse(PathList,Rev),
   print_pathlist(Rev),
   nl,
   print_result(T).

print_pathlist([]).
print_pathlist([Path|Paths]) :-
   print_path(Path),nl,
   print_pathlist(Paths).

print_path([]) :-
   print(root).
print_path([H|T]) :-
   print_path(T,H).

print_path([],X) :-
   print(X).
print_path([H|T],X) :-
   print_path(T,H:X).


