/*  $Id$

    This file is part of the

    GRALE: a grammar visualization tool

    Developed by Holger Wunsch and Detmar Meurers
    Copyright (C) 1998-2006 by the developers
    All rights reserved

    Seminar fr Sprachwissenschaft, Universitt Tbingen and
    Department of Linguistics, The Ohio State University

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/** @author Martin Lazarov */

#ifndef LRS_CONFIG_H
#define LRS_CONFIG_H

#include "LRS_Util.h"

#include <map>
#include <string>
#include <sstream>

#include <qcolor.h>

namespace LRS {

/** A configuration class for the LRS viewer */
class Config {
  std::map<std::string,std::string> _map;
public:
  /** Load the config data from a file */
  Config(const std::string& filename);
  
  /** Gets the value of a key. Throws Config::UnknowKeyError
        if the key does not have a value.
  */
  template <class T>
  T forKey(const std::string& key) const {
    if (!_map.count(key))
      throw UnknownKeyError(key);
    return forKey(key, T());
  }
  
  /** Gets the value of a key, or returns the given default,
        if the key is undefined.
  */
  template <class T>
  T forKey(const std::string& key, T defaultValue) const {
    if (!_map.count(key))
      return defaultValue;
    return convert<T>(_map.find(key)->second);
  }
  
  /** Converts a string to a value. Throws Config::ConvertionError
        if the conversion cannot be performed.
  */
  template <class T>
  static T convert(const std::string& s) {
    T value;
    std::istringstream iss(s);
    iss >> value;
    if (!(iss && iss.eof()))
      throw ConvertionError(s);
    return value;
  }
  
  struct UnknownKeyError : GeneralError {
    UnknownKeyError(const std::string& key)
      : GeneralError("unknown key:", key.c_str()) {}
  };
  
  struct ConvertionError : GeneralError {
    ConvertionError(const std::string& key)
      : GeneralError("convertion failed:", key.c_str()) {}
  };
};

/// Specialization for std::string.
template <>
inline
std::string Config::convert<std::string>(const std::string& s) {
  return s;
}

/// Specialization for QColor.
template <>
inline
QColor Config::convert<QColor>(const std::string& s) {
  QColor color(s.c_str());
  if (!color.isValid())
    throw ConvertionError(s + " (as color)");
  return color;
}

}

#endif // !LRS_CONFIG_H
