/*  $Id$

    This file is part of the

    GRALE: a grammar visualization tool

    Developed by Holger Wunsch and Detmar Meurers
    Copyright (C) 1998-2006 by the developers
    All rights reserved

    Seminar fr Sprachwissenschaft, Universitt Tbingen and
    Department of Linguistics, The Ohio State University

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/** @author Martin Lazarov */

#ifndef LRS_OBJECT_MODEL_H
#define LRS_OBJECT_MODEL_H

#include "LRS_Util.h"

#include <string>
#include <memory>

/**
  This file contains the definitions of several classes,
    which together comprise an object model for LRS expressions.
    
    @author Martin Lazarov
*/

namespace LRS {

struct TermVisitor;

/** The base class of all "terms". Provides support for the
      Visitor design pattern.
*/
struct Term : NonCopyable {
  virtual ~Term() {}
  virtual void accept(TermVisitor*) = 0;
};

/** A list of terms. Note that this class is not a term itself.
      All terms in the list are "owned" by it and are automatically
      destroyed when the list is destroyed.
*/
struct ListOfTerms : AutoPtrVector<Term> {
  void addTerm(Term* t)
    { push_back(t); }
};

/** An LRS expression (formula). Can be converted to XML. */
struct Expression : ListOfTerms {
  static std::string toXMLString(const std::string& lrsExpr);
  std::string toXMLString();
};

/** A base class for all identifiable terms: variables, meta-variables and
  functors.
*/
struct NamedTerm : Term {
  NamedTerm(const std::string& name)
    : _name(name) {}
  const std::string& name() const
    { return _name; }
protected:
  std::string _name;
};

/** A variable term */
struct _Var : NamedTerm {
  _Var(const std::string& name)
    : NamedTerm(name) {}
};

/** A base class for terms that can have subterms */
struct ComplexTerm : NamedTerm {
  ComplexTerm(const std::string& name, ListOfTerms* subTerms)
    : NamedTerm(name), _subTerms(subTerms) {}
  const ListOfTerms* subTerms() const
    { return _subTerms.get(); }
protected:
  std::auto_ptr<ListOfTerms> _subTerms;
};

/** A meta-variable term */
struct _MetaVar : ComplexTerm {
  _MetaVar(const std::string& name, ListOfTerms* subTerms)
    : ComplexTerm(name, subTerms) {}
};

/** A functor term */
struct _Functor : ComplexTerm {
  _Functor(const std::string& name, ListOfTerms* args, ListOfTerms* subTerms)
    : ComplexTerm(name, subTerms), _args(args) {}
  const ListOfTerms* arguments() const
    { return _args.get(); }
private:
  std::auto_ptr<ListOfTerms> _args;
};

/** A base class for contents */
struct Content : Term {
  Content(ListOfTerms* innerTerms) 
    : _innerTerms(innerTerms) {}
  const ListOfTerms* innerTerms() const
    { return _innerTerms.get(); }
protected:
  std::auto_ptr<ListOfTerms> _innerTerms;
};

/** Internal content */
struct _InCont : Content {
  _InCont(ListOfTerms* innerTerms)
    : Content(innerTerms) {}
};

/** External content */
struct _ExCont : Content {
  _ExCont(ListOfTerms* innerTerms)
    : Content(innerTerms) {}
};

/** ??? */
struct _SqCont : Content {
  _SqCont(ListOfTerms* innerTerms)
    : Content(innerTerms) {}
};

/** A helper class template providing the implementation
      of the various terms' accept methods.
*/
template <class BaseClass>
struct TTerm : BaseClass {
  template <class A1>
  TTerm(A1 a1) : BaseClass(a1) {}
  
  template <class A1, class A2>
  TTerm(A1 a1, A2 a2) : BaseClass(a1, a2) {}
  
  template <class A1, class A2, class A3>
  TTerm(A1 a1, A2 a2, A3 a3) : BaseClass(a1, a2, a3) {}
  
  void accept(TermVisitor* v);
};

// Instantiate the term classes from the class templete
typedef TTerm<_Var>     Var;
typedef TTerm<_MetaVar> MetaVar;
typedef TTerm<_Functor> Functor;
typedef TTerm<_InCont>  InCont;
typedef TTerm<_ExCont>  ExCont;
typedef TTerm<_SqCont>  SqCont;

/** A term visitor. For sample usage see the implementation of the
      toXMLString method of the Expression class.
*/
struct TermVisitor {
  virtual void visit(Var* t)         { visit(static_cast<NamedTerm*>(t)); }
  virtual void visit(MetaVar* t)     { visit(static_cast<ComplexTerm*>(t)); }
  virtual void visit(Functor* t)     { visit(static_cast<ComplexTerm*>(t)); }
  virtual void visit(InCont* t)      { visit(static_cast<Content*>(t)); }
  virtual void visit(ExCont* t)      { visit(static_cast<Content*>(t)); }
  virtual void visit(SqCont* t)      { visit(static_cast<Content*>(t)); }
  
  virtual void visit(Content*)       { }
  virtual void visit(ComplexTerm* t) { visit(static_cast<NamedTerm*>(t)); }
  virtual void visit(NamedTerm* t)   { visit(static_cast<Term*>(t)); }
  virtual void visit(Term*)          { }
  
  virtual ~TermVisitor() {}
};

template <class T>
void TTerm<T>::accept(TermVisitor* v) { v->visit(this); }

} // namespace LRS

#endif // !LRS_OBJECT_MODEL_H
