/*  $Id$

    This file is part of the

    GRALE: a grammar visualization tool

    Developed by Holger Wunsch and Detmar Meurers
    Copyright (C) 1998-2005 by the developers
    All rights reserved

    Seminar fr Sprachwissenschaft, Universitt Tbingen and
    Department of Linguistics, The Ohio State University

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


/** @file

    Implementation of the <code>Canvas</code>, with exception of all
    drawing routines, which can be found in drawer.cpp.

    @author Holger Wunsch */

#include <q3accel.h>
#include <qdir.h>
#include <qregexp.h>
#include <qstringlist.h>
//Added by qt3to4:
#include <QMouseEvent>
#include <QWheelEvent>

#include "global.h"
#include "canvas.h"
#include "node.h"

#include "LRS_MainWindow.h"
#include "LRS_ExpressionParser.h"

#if 0
#include <typeinfo>
#include <iostream>

void print_parents(QWidget* w) {
  if (w) {
    print_parents(w->parentWidget());
    std::cerr << typeid(w).name() << "/" << w->windowTitle().toStdString() << "/" << w->objectName().toStdString() << std::endl;
  }
}
#endif

static inline
ViewCanvas* view(Canvas* c) {
  return dynamic_cast<ViewCanvas*>(c->parentWidget()->parentWidget()->parentWidget());
}

Canvas::Canvas(Options *op, GStructureData *doc, GenericNode* rootNd, QWidget *parent, const char *name )
  :QWidget(parent, name)
{
  options=op;
  
  setAttribute(Qt::WA_PaintOutsidePaintEvent);
  setPalette(options->backgroundColor);

  document=doc;

  if (rootNd == NULL) {
    rootNode = document->data->at(0);
  }
  else
    rootNode = rootNd;

  toKlick.setAutoDelete(true);

  resetToClick();
  
  maxSize=QSize(100,100);
  offsetX=offsetY=0;

  dragging = false;
  setMouseTracking(true);

  highlight = NULL;
  foundMark = NULL;
  
  dispNodes = new IdTree(document->maxID() + 1);
  
  resize(1, 1);
}


Canvas::~Canvas() {
  delete dispNodes;
}

void Canvas::mousePressEvent(QMouseEvent *mo) {
  DisplayableNode* dn = 0;

  if (mo->button() == Qt::RightButton) {
    klickedOn = whereKlicked(mo->pos());
    if (klickedOn) {
      //popup = new KPopupMenu(this);
      popup = new Q3PopupMenu(this);
      switch (klickedOn->type) {
      case NODETYPE_TYPE_NODE: {
        char itemText[500];
        NodeAddress addrParent;
        
        klickedOn->addr.parentAddress(addrParent);

        dnd = dispNodes->findDisplayableNode(klickedOn->addr);

        sprintf(itemText, "Type '%s'", ((TypeNode*) dnd->node)->name);
        popup->insertItem(itemText);
        
        //print_parents(this);

        sprintf(itemText, "Show/Hide Type '%s'", ((TypeNode*) dnd->node)->name);
        popup->insertItem(itemText, view(this), SLOT(hideUnhidePopupSlot()));
        
        if ((dnd->flags & NODEFLAG_HIDDEN) && (dnd->flags & NODEFLAG_SHOWHIDDEN))
          popup->insertItem("Omit This Hidden Type", 
                            this, SLOT(hideNodePopupSlot()));
        popup->insertSeparator();

        dn = dispNodes->findDisplayableNode(addrParent);
        QStringList names = namesOfHiddenChildNodes(addrParent);
        // Index 0 becomes the search item below, therefore start with index one
        for (int i = 0; i < names.count(); i++)
          popup->insertItem(names[i], i + 1);
        popup->insertItem("Display All Hidden Types/Attributes on Level", 
                          this, SLOT(unhideLevelPopupSlot()));
        popup->insertItem("Omit All Hidden Types/Attributes on Level", 
                          this, SLOT(hideLevelPopupSlot()));

        popup->insertSeparator();

        sprintf(itemText, "Find Next Occurrence of '%s'", ((TypeNode*) dnd->node)->name);
        popup->insertItem(QString::fromLocal8Bit(itemText), 0);

        int index = popup->exec(mapToGlobal(mo->pos()),2);

        if (index == 0)
          view(this)->contextMenuSearch(((TypeNode*) dnd->node)->name, klickedOn->addr);
        else if (index >= 1) {
          // Index is one higher than the number of the child node
          dn->node->unhideChildNode(this, addrParent, index - 1);
          emit repaint();
        }

        break;
      }
      case NODETYPE_FEATVAL_NODE: {
        char itemText[500];
        NodeAddress addrParent;
        klickedOn->addr.parentAddress(addrParent);

        dn = dispNodes->findDisplayableNode(klickedOn->addr);
        dnd = dn;

        sprintf(itemText, "Attribute '%s'", ((FeatvalNode*) dnd->node)->feature);
        popup->insertItem(itemText);

        sprintf(itemText, "Show/Hide Attribute '%s'", ((FeatvalNode*) dnd->node)->feature);
        popup->insertItem(itemText, view(this), SLOT(hideUnhidePopupSlot()));
        
        if ((dn->flags & NODEFLAG_HIDDEN) && (dn->flags & NODEFLAG_SHOWHIDDEN))
          popup->insertItem("Omit This Hidden Attribute", 
                            this, SLOT(hideNodePopupSlot()));
        popup->insertSeparator();

        dn = dispNodes->findDisplayableNode(addrParent);
        QStringList names = namesOfHiddenChildNodes(addrParent);

        // Index 0 becomes the search item below, therefore start with index one
        for (int i = 0; i < names.count(); i++)
          popup->insertItem(names[i], i + 1);
        popup->insertItem("Display All Hidden Types/Attributes on Level", 
                          this, SLOT(unhideLevelPopupSlot()));
        popup->insertItem("Omit All Hidden Types/Attributes on Level", 
                          this, SLOT(hideLevelPopupSlot()));

        popup->insertSeparator();

        
        sprintf(itemText, "Find Next Occurrence of '%s'", ((FeatvalNode*) dnd->node)->feature);
        popup->insertItem(QString::fromLocal8Bit(itemText), 0);

        int index = popup->exec(mapToGlobal(mo->pos()),2);

        if (index == 0)
          view(this)->contextMenuSearch(((FeatvalNode*) dnd->node)->feature, klickedOn->addr);
        else if (index >= 1) {
          // Index is one higher than the number of the child node
          dn->node->unhideChildNode(this, addrParent, index - 1);
          emit repaint();
        }

        break;
      }
      case NODETYPE_LIST_NODE:
      case NODETYPE_TAIL_NODE:
        dnd = dispNodes->findDisplayableNode(klickedOn->addr);
        popup->insertItem("List");
        popup->insertItem("Show/Hide List", 
                          view(this), SLOT(hideUnhidePopupSlot()));
        popup->popup(mapToGlobal(mo->pos()),2);
        break;
      case NODETYPE_SET_NODE:
      case NODETYPE_REST_NODE:
        dnd = dispNodes->findDisplayableNode(klickedOn->addr);
        popup->insertItem("Set");
        popup->insertItem("Show/Hide Set", 
                          view(this), SLOT(hideUnhidePopupSlot()));
        popup->popup(mapToGlobal(mo->pos()),2);
        break;
      case NODETYPE_FUNCTION_NODE: {
        dn = dispNodes->findDisplayableNode(klickedOn->addr);
        dnd = dn;

        popup->insertItem("Function");
        popup->insertItem("Show/Hide Function", 
                          view(this), SLOT(hideUnhidePopupSlot()));

        popup->insertSeparator();

        char itemText[500];
        sprintf(itemText, "Find Next Occurrence of '%s'", ((FunctionNode*) dn->node)->functor);
        popup->insertItem(QString::fromLocal8Bit(itemText), 0);

        int index = popup->exec(mapToGlobal(mo->pos()),2);

        if (index == 0)
          view(this)->contextMenuSearch(((FunctionNode*) dn->node)->functor, klickedOn->addr);

        break;
      }
      case NODETYPE_REF_NODE: {
        dn = dispNodes->findDisplayableNode(klickedOn->addr);
        dnd = dn;

        popup->insertItem("Tag");
        if (!(dn->flags & NODEFLAG_EXPANDED))
          popup->insertItem("Expand Tag", parentWidget()->parentWidget(), SLOT(expandLocalPopupSlot()));
        popup->insertItem("Show/Hide Tag", 
                          view(this), SLOT(hideUnhidePopupSlot()));

        popup->insertSeparator();

        char itemText[500];
        sprintf(itemText, "Find Next Occurrence of '%d'", ((RefNode*) dn->node)->ref);
        popup->insertItem(QString::fromLocal8Bit(itemText), 0);

        int index = popup->exec(mapToGlobal(mo->pos()),2);

        if (index == 0) {
          char refStr[50];
          sprintf(refStr, "%d", ((RefNode*) dn->node)->ref);
          view(this)->contextMenuSearch(refStr, klickedOn->addr);
        }

        break;
      }
      case NODETYPE_REENT_NODE:
        dnd = dispNodes->findDisplayableNode(klickedOn->addr);
        popup->insertItem("Reentrancy");
        popup->insertItem("Show/Hide Reentrancy", 
                          view(this), SLOT(hideUnhidePopupSlot()));
        popup->popup(mapToGlobal(mo->pos()),2);
        break;
      case NODETYPE_TREE_NODE: {
        dn = dispNodes->findDisplayableNode(klickedOn->addr);
        dnd = dn;
        
        popup->insertItem("Tree");
        popup->insertItem("Show/Hide Structure", this, SLOT(expandCollapseSlot()));
        popup->insertItem("Show/Hide Tree Node", 
                          view(this), SLOT(hideUnhidePopupSlot()));

        popup->insertSeparator();

        char itemText[500];
        sprintf(itemText, "Find Next Occurrence of '%s'", ((TreeNode*) dn->node)->label);
        popup->insertItem(QString::fromLocal8Bit(itemText), 0);

        int index = popup->exec(mapToGlobal(mo->pos()),2);

        if (index == 0)
          view(this)->contextMenuSearch(((TreeNode*) dn->node)->label, 
            klickedOn->addr);

        break;
      }
      case NODETYPE_ATOM_NODE: {
        dn = dispNodes->findDisplayableNode(klickedOn->addr);
        dnd = dn;
        popup->insertItem("Atom");
        popup->insertItem("Show/Hide Atom", 
                          view(this), SLOT(hideUnhidePopupSlot()));

        popup->insertSeparator();

        char itemText[500];
        sprintf(itemText, "Find Next Occurrence of '%s'", ((AtomNode*) dn->node)->atom);
        popup->insertItem(QString::fromLocal8Bit(itemText), 0);

        int index = popup->exec(mapToGlobal(mo->pos()),2);

        if (index == 0)
          view(this)->contextMenuSearch(((TreeNode*) dn->node)->label, 
            klickedOn->addr);

        break;
      }
      case NODETYPE_REL_NODE: {
        dn = dispNodes->findDisplayableNode(klickedOn->addr);
        dnd = dn;
        popup->insertItem("Relation");
        popup->insertItem("Show/Hide Relation", 
                          view(this), SLOT(hideUnhidePopupSlot()));

        popup->insertSeparator();

        char itemText[500];
        sprintf(itemText, "Find Next Occurrence of '%s'", ((RelationNode*) dn->node)->functor);
        popup->insertItem(QString::fromLocal8Bit(itemText), 0);

        int index = popup->exec(mapToGlobal(mo->pos()),2);

        if (index == 0)
          view(this)->contextMenuSearch(((RelationNode*) dn->node)->functor, 
            klickedOn->addr);

        break;
      }
      case NODETYPE_STRUC_NODE: {
        dn = dispNodes->findDisplayableNode(klickedOn->addr);
        dnd = dn;
        popup->insertItem("Feature Structure");
        popup->insertItem("Show/Hide Feature Structure", 
                          view(this), SLOT(hideUnhidePopupSlot()));

        popup->exec(mapToGlobal(mo->pos()),2);
        break;
      }
      }
    }
  }
  else if (mo->button() == Qt::LeftButton) {
    klickedOn = whereKlicked(mo->pos());
    if(klickedOn) {
      if (mo->state() & Qt::ShiftModifier) {
        switch (klickedOn->type) {
        case NODETYPE_TYPE_NODE:
          dn = dispNodes->findDisplayableNode(klickedOn->addr);
          view(this)->contextMenuSearch(((TypeNode*) dn->node)->name, klickedOn->addr);
          break;
        case NODETYPE_FEATVAL_NODE:
          dn = dispNodes->findDisplayableNode(klickedOn->addr);
          view(this)->contextMenuSearch(((FeatvalNode*) dn->node)->feature, klickedOn->addr);
          break;
        case NODETYPE_FUNCTION_NODE:
          dn = dispNodes->findDisplayableNode(klickedOn->addr);
          view(this)->contextMenuSearch(((FunctionNode*) dn->node)->functor, klickedOn->addr);
          break;
        case NODETYPE_REL_NODE:
          dn = dispNodes->findDisplayableNode(klickedOn->addr);
          view(this)->contextMenuSearch(((RelationNode*) dn->node)->functor, klickedOn->addr);
          break;
        case NODETYPE_TREE_NODE:
          dn = dispNodes->findDisplayableNode(klickedOn->addr);
          view(this)->contextMenuSearch(((TreeNode*) dn->node)->label, klickedOn->addr);
          break;
        case NODETYPE_REF_NODE:
          dn = dispNodes->findDisplayableNode(klickedOn->addr);

          if (mo->state() & Qt::ShiftModifier) {
            char refStr[50];
            sprintf(refStr, "%d", ((RefNode*) dn->node)->ref);
            view(this)->contextMenuSearch(refStr, klickedOn->addr);
          }

          break;
        case NODETYPE_ATOM_NODE:
          dn = dispNodes->findDisplayableNode(klickedOn->addr);
          view(this)->contextMenuSearch(((AtomNode*) dn->node)->atom, klickedOn->addr);
          break;
        }
      }
      else {
        dn = dispNodes->findDisplayableNode(klickedOn->addr);
        if (dn && dn->node->nodetype == NODETYPE_ATOM_NODE) {
          std::string lrs = ((AtomNode*) dn->node)->atom;
          if (lrs.compare(0, strlen(LRS_PREFIX), LRS_PREFIX) == 0) {
            std::string expr = lrs.substr(strlen(LRS_PREFIX));
            std::string conf = (qApp->applicationDirPath() + "/conf/lrsviewer.conf").toStdString();
            try {
              if (LRS::MainWindow* lrsViewer = new LRS::MainWindow(expr, conf, view(this))) {
                lrsViewer->setCaption("LRS");
                lrsViewer->show();
              }
            }
            catch (const LRS::ExpressionParser::SyntaxError& e) {
              QMessageBox::critical(view(this), "LRS syntax error", e.what());
              qWarning("LRS syntax error: |%s|", e.what());
            }
            catch (...) {
              QMessageBox::critical(view(this), "Unknown exception", "LRS viewer error");
            }
          }
        }
        if (dn)
          switch (klickedOn->type) {
          case NODETYPE_TREE_NODE:
            if (dn->flags & NODEFLAG_HIDDEN)
              hideUnhide(dn);
            else
              expandCollapse(dn);
            break;
          case NODETYPE_REF_NODE:
            if (dn->flags & NODEFLAG_HIDDEN)
              hideUnhide(dn);
            else
              expandCollapse(dn);
            break;
          default:
            // Uncomment to disable hiding of whole feature structures
            //	  if (klickedOn->type != NODETYPE_STRUC_NODE) {
            if (dn->flags & NODEFLAG_HIDDEN)
              hideUnhide(dn);
            else if (mo->modifiers() & Qt::ControlModifier)
              hideUnhide(dn);
              //	  }
          }
      }
    }
  }
  if (mo->button() == Qt::MidButton) {
    dragX = mo->pos().x();
    dragY = mo->pos().y();
    dragging = true;
  }
}


void Canvas::mouseDoubleClickEvent(QMouseEvent *mo) {
  DisplayableNode* dn;

  if (mo->button() == Qt::LeftButton) {
    klickedOn = whereKlicked(mo->pos());
    if(klickedOn) {  
      dn = dispNodes->findDisplayableNode(klickedOn->addr);
      if ((dn->flags & NODEFLAG_HIDDEN) == 0)
        hideUnhide(dn);
    }
  }
}


void Canvas::mouseMoveEvent(QMouseEvent *mo) {
    int x = mo->pos().x();
    int y = mo->pos().y();
    
    /*    char buf[1000];
          klickedOn = whereKlicked(mo->pos());
          if (klickedOn) {
          klickedOn->addr.sprint(buf);
          view(this)->statusBar->message(buf);
          }*/

    
    if (dragging) {
      //int dx = dragX - x;
      //int dy = dragY - y;

      dragX = x;  
      dragY = y;

      //int nOx = offsetX + dx;
      //int nOy = offsetY + dy;
      //TODO:??
      //if(v_scroll->value() != nOy)
        //v_scroll->setValue(nOy);
      //if(h_scroll->value() != nOx)
        //h_scroll->setValue(nOx);
    }
    else {
      klickedOn = whereKlicked(mo->pos());
      if (klickedOn != NULL) {
        DisplayableNode* nd = dispNodes->findDisplayableNode(klickedOn->addr);

        // If the node the user points on is a type node, featval node, or struc node, 
        // calculate a list containing the type and all
        // features of the structure node in order to display additional
        // information about the structure
        if ((nd->node->nodetype == NODETYPE_TYPE_NODE) ||
            (nd->node->nodetype == NODETYPE_FEATVAL_NODE) ||
            (nd->node->nodetype == NODETYPE_STRUC_NODE)) {
          QStringList typeAndFeats;
          NodeAddress addr;
          StrucNode* snd;
          
          if (nd->node->nodetype == NODETYPE_STRUC_NODE)
            addr = klickedOn->addr;
          else
            klickedOn->addr.parentAddress(addr);
          
          snd = (StrucNode*) (dispNodes->findDisplayableNode(addr)->node);
          typeAndFeats = typeAndFeatsList(snd);
          
          QString statusMsg;
          statusMsg.append(typeAndFeats[0]);
          if (typeAndFeats.count() > 1) {
            statusMsg.append(": ");
            for (int i = 1; i < typeAndFeats.count(); i++) {
              statusMsg.append(typeAndFeats[i]);
              statusMsg.append(" ");
            }
          }
          view(this)->setStatusBarMessage((const char*) statusMsg.local8Bit());
        }
        else
          view(this)->setStatusBarMessage("");
          
        if (options->highlightStructures) {
          if (highlight == NULL) {
            highlight = nd;
            repaint(true);
          }
          else if (highlight != nd) {
            highlight = nd;
            repaint(true);
          }
        }
      }
      else {
        view(this)->setStatusBarMessage("");
        if (options->highlightStructures) {
          if (highlight != NULL) {
            highlight = NULL;
            repaint(true);
          }
        }
      }
    }
}


void Canvas::mouseReleaseEvent(QMouseEvent *mo) {
  if (mo->button() == Qt::MidButton) {
    dragging = false;
  }
}


Canvas::KlickBox* Canvas::whereKlicked(const QPoint &p) {
  KlickBox *klick;
  for(klick = toKlick.first(); klick != NULL; klick = toKlick.next()) {
    if (klick->box.contains(p)) {
      return klick;
    }
  }

  return NULL;
}



void Canvas::hideUnhide(DisplayableNode* nd, bool redraw) {
  if (nd->flags & NODEFLAG_HIDDEN)
    nd->flags &= ~NODEFLAG_HIDDEN;
  else
    nd->flags |= NODEFLAG_HIDDEN;
 
  // repaint ...
  if ( redraw ) {
    erase();
    emit redrawSignal();
  }
}


void Canvas::hideLevelPopupSlot(void) {
  NodeAddress addrParent;
  klickedOn->addr.parentAddress(addrParent);
  
  DisplayableNode* dn = dispNodes->findDisplayableNode(addrParent);
  if (dn) {
    dn->node->hideUnhideChildNodes(this, addrParent, true);
    emit redrawSignal();
  }
}


void Canvas::unhideLevelPopupSlot(void) {
  NodeAddress addrParent;
  klickedOn->addr.parentAddress(addrParent);
  
  DisplayableNode* dn = dispNodes->findDisplayableNode(addrParent);
  if (dn) {
    dn->node->hideUnhideChildNodes(this, addrParent, false);
    emit redrawSignal();
  }
}


void Canvas::hideNodePopupSlot(void) {
  DisplayableNode* dn = dispNodes->findDisplayableNode(klickedOn->addr);
  if (dn) {
    dn->flags &= ~NODEFLAG_SHOWHIDDEN;
    emit redrawSignal();
  }
}
  


void Canvas::expandCollapse(DisplayableNode* nd, bool redraw) {
  if (nd->flags & NODEFLAG_EXPANDED)
    nd->flags &= ~NODEFLAG_EXPANDED;
  else
    nd->flags |= NODEFLAG_EXPANDED;

  if (redraw) {
    erase();
    emit redrawSignal();
  }
}

void Canvas::expandCollapseSlot() {
  expandCollapse(dispNodes->findDisplayableNode(klickedOn->addr), true);
}

void Canvas::setRootNode(GenericNode* node) {
  if (node)
    rootNode = node;
  else
    rootNode = document->data->at(0);

  highlight = NULL;
  foundMark = NULL;

  emit redrawSignal();
}

Canvas::KlickBox* Canvas::getKlickedOn() {
  return klickedOn;
}

IdTree* Canvas::getDispNodes(void) {
  return dispNodes;
}

void Canvas::reset() {
  delete dispNodes;
  dispNodes = new IdTree(document->maxID() + 1);
  emit redrawSignal();
  highlight = NULL;
  foundMark = NULL;
}
  
void Canvas::resetToClick(void) {
  toKlick.clear();
}

void Canvas::updateOptions(const Options& op) {
  options->backgroundColor = op.backgroundColor;
  options->normalColor = op.normalColor;
  options->typeColor = op.typeColor;
  options->textColor = op.textColor;
  options->errorColor = op.errorColor;
  options->treeColor = op.treeColor;
  options->treeLineColor = op.treeLineColor;
  options->leafColor = op.leafColor;
  options->hiddenColor = op.hiddenColor;
  options->markerColor = op.markerColor;
  options->foundColor = op.foundColor;
  options->differentNodeColor = op.differentNodeColor;
  options->arclabelColor = op.arclabelColor;
  options->arclabelBgColor = op.arclabelBgColor;

  options->normalFont = op.normalFont;
  options->typeFont = op.typeFont;
  options->textFont = op.textFont;
  options->treeFont = op.treeFont;
  options->leafFont = op.leafFont;
  options->errorFont = op.errorFont;
  options->arclabelFont = op.arclabelFont;
}
  
void Canvas::ensureVisible(const NodeAddress* addr) {
  char id[ID_LENGTH];
  int idpos = 0;
  char idb[ID_LENGTH];
  int i;
  NodeAddress a;
  bool expand;
  GenericNode* node;
  for (unsigned int pos = 0; pos < addr->len; pos++) {
    id[pos] = (addr->addr[pos] & 0x0F) + '0';
    if (addr->addr[pos] & 0x40) {
      pos++;
      id[pos] = 0x00;
      break;
    }
  }
  node = document->findNodeById(id);
  if (node->nodetype == NODETYPE_TREE_NODE)
    expand = false;
  else
    expand = true;
    
  for (int pos = addr->len - 1; pos >= 0; pos--) {
    idb[idpos] = (addr->addr[pos] & 0x0F) + '0';
    if (addr->addr[pos] & 0x80) {
      for (i = 0; i <= idpos; i++)
        id[i] = idb[idpos - i];
      id[i] = 0x00;
      NodeAddress ad(a, id);
      node = document->findNodeById(id);
      dnd = dispNodes->addDisplayableNode(ad, node);
      dnd->flags &= ~(NODEFLAG_HIDDEN);
      if (expand && (pos > 0))
        dnd->flags |= NODEFLAG_EXPANDED;
      a = ad;
      idpos = 0;
    }
    else
      idpos++;
  }
}

QPoint Canvas::nodeOrigin(const NodeAddress* addr) {
  KlickBox *klick;
  for(klick = toKlick.first(); klick != NULL; klick = toKlick.next()) {
    if (klick->addr == *addr)
      return QPoint((klick->box.left() + offsetX > 20 ? klick->box.left() + offsetX - 20 : 0), (klick->box.top() + offsetY > 20 ? klick->box.top() + offsetY - 20 : 0));
  }
  return QPoint(0, 0);
}


      
void Canvas::markFound(const NodeAddress* addr) {
  if (foundMark != NULL)
    foundMark->flags &= ~(NODEFLAG_FOUND);
  if (addr == NULL) {
    foundMark = NULL;
    repaint(false);
  }
  else {
    dnd = dispNodes->findDisplayableNode(*addr);
    dnd->flags |= NODEFLAG_FOUND;
    foundMark = dnd;

    update();
  }
}


GStructureData* Canvas::getDocument(void) const {
  return document;
}


/** Returns a list that contains the type and all attributes of a
    feature structure (a <code>StrucNode</code>).
    @param node the <code>StrucNode</code>
    @return a list containing the type as its first element followed by all
      attributes. */
QStringList Canvas::typeAndFeatsList(StrucNode* node) {
  QStringList list;
  List* featvals;
  
  list.append(QString::fromLocal8Bit(node->type->name));

  featvals = node->featvals;
  while (featvals) {
    list.append(QString::fromLocal8Bit(((FeatvalNode*) featvals->item)->feature));
    featvals = featvals->tail;
  }

  return list;
}

  
QStringList Canvas::namesOfHiddenChildNodes(NodeAddress& addr) {
  QStringList names;
  DisplayableNode* dnd = dispNodes->findDisplayableNode(addr);
  char name[500];
  
  if (dnd) {
    if (dnd->node->nodetype == NODETYPE_STRUC_NODE) {
      StrucNode* nd = (StrucNode*) dnd->node;
      List* featvals;

      dnd = dispNodes->addDisplayableNode(NodeAddress(addr, nd->type), nd->type);
      if ((dnd->flags & NODEFLAG_HIDDEN) && !(dnd->flags & NODEFLAG_SHOWHIDDEN)) {
        sprintf(name, "Show Type: %s", nd->type->name);
        names.append(QString::fromLocal8Bit(name));
      }
      
      featvals = nd->featvals;
      while (featvals) {
        dnd = dispNodes->addDisplayableNode(NodeAddress(addr, featvals->item), featvals->item);
        if ((dnd->flags & NODEFLAG_HIDDEN) && !(dnd->flags & NODEFLAG_SHOWHIDDEN)) {
          sprintf(name, "Show Attribute: %s", ((FeatvalNode*) featvals->item)->feature);
          names.append(QString::fromLocal8Bit(name));
        }
        featvals = featvals->tail;
      }
    }
  }

  return names;
}

DisplayableNode* Canvas::getDispNode() {
  return dnd;
}
