/*  $Id$

    This file is part of the

    GRALE: a grammar visualization tool

    Developed by Holger Wunsch and Detmar Meurers
    Copyright (C) 1998-2005 by the developers
    All rights reserved

    Seminar fr Sprachwissenschaft, Universitt Tbingen and
    Department of Linguistics, The Ohio State University

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/**
   @file

   GRALE main program file.

   @author Holger Wunsch */

/** @mainpage

    <center>
    <h2>GRALE: a grammar visualization tool - Technical Documentation</h2>
    <h3>Version 1.4.x</h3>

    Developed by Holger Wunsch and Detmar Meurers<br>
    Copyright (C) 1998-2005 by the developers<br>
    All rights reserved<br>

    Seminar fr Sprachwissenschaft, Universitt Tbingen and<br>
    Department of Linguistics, The Ohio State University


    For a documentation of the GRALE interface language syntax, see the
    <a href="../interface-language-ebnf/ebnf.txt">interface-language EBNF</a>
    </center>
*/

#include <iostream>

#include <qglobal.h>
#if QT_VERSION >= 0x030000
#include <q3filedialog.h>
#else
#include <qfiledlg.h>
#endif
#include <qtextcodec.h>

#include <string.h>
#include <stdlib.h>

#include <set>
#include <vector>
#include <string>

//#include <kapp.h>
#include <qapplication.h>

#include <qstringlist.h>

//#include <kmessagebox.h>
#include <qmessagebox.h>

//#include <kstddirs.h>
//#include <kcmdlineargs.h>
//#include <kaboutdata.h>

#include "mainw.h"
#include "options.h"
#include "gstructuredata.h" 
#include "gdatamanager.h"

#include "config.h"
#include "global.h"

#include "gdataexchange.h"


/** The global application options. */
Options *globalOptions;


/** <code>true</code> if an options was changed during a GRALE session, <code>false</code> otherwise. */
bool optionChanged;


/** The main window. */
MainWindow *mainw;


/** The application object */
//KApplication *application;


/** The application-wide data exchange manager. */
GDataExchange* dataxc;


/** The application-wide data manager. */
GDataManager* dataMgr;


/** If this variable is set to <code>true</code>, GRALE will dump all
    data coming in on its socket connection to the file <code>GRALE.dmp</code>
    in the directory where GRALE was started. */
bool dumpData;


/** If this variable is set to <code>true</code>, GRALE will accept connections on
    an internet socket instead of a file socket. */
bool useTCP;

const char* resourcePath;

struct _CmdLineArgs {
  struct Options {
    const char* option;
    const char* description;
    int dummy;
  };
  
  std::set<std::string>     _opts;
  std::vector<std::string>  _args;
  
  _CmdLineArgs(int argc, char** argv, Options* opts = 0) {
    for (int i = 1; i < argc; ++i) {
      std::string arg = argv[i];
      if (arg[0] == '-') {
        if (arg == "-help" ||
            arg == "-h" ||
            arg == "--help" ||
            arg == "--h" ||
            arg == "-?")
        {
          // print usage and exit
          std::cout << APP_NAME   << " Version " << APP_VERSION << std::endl;
          std::cout << DEVELOPERS << std::endl;
          std::cout << COPYRIGHT  << std::endl << std::endl;
          std::cout << "Options:" << std::endl;
          if (opts) while (opts->option) {
            std::cout << "  " << opts->option << '\t' << opts->description << std::endl;
            ++opts;
          }
          exit(1);
        }
        _opts.insert(arg.substr(1));
      }
      else {
        _args.push_back(arg);
      }
    }
  }
  
  unsigned count() const {
    return _args.size();
  }
  
  const char* arg(unsigned pos) const {
    return _args.at(pos).c_str();
  }
  
  bool isSet(const std::string& opt) const {
    return _opts.find(opt) != _opts.end();
  }
};

static _CmdLineArgs::Options allowedArgs[] = {
  {"dump", "Dump incoming data to " DATADUMP_FILENAME, 0},
  {"net", "Use a network connection instead of\na local connection to transfer data", 0},
  {"+GRALE file(s)", "One or more GRALE data files to be opened", 0},
  {0, 0, 0},
};


/** Program entry point. */
int main( int argc, char *argv[] )
{
  QApplication app(argc, argv);
  
  QApplication::setOrganizationName("SfS");
  QApplication::setOrganizationDomain("sfs.uni-tuebingen.de");
  QApplication::setApplicationName(APP_NAME);
  
  char* resPath = getenv("GRALE_RESOURCE_PATH");
  if (resPath)
    resourcePath = strdup(resPath);
  else
    resourcePath = strdup((app.applicationDirPath() + "/pics").toStdString().c_str());

  if (char* sPort = getenv("GDE_PORTRANGE_LO")) {
    GDE_PORTRANGE_LO = atoi(sPort);
    if (sPort = getenv("GDE_PORTRANGE_HI")) {
      GDE_PORTRANGE_HI = atoi(sPort);
    }
    else {
      GDE_PORTRANGE_HI = GDE_PORTRANGE_LO + 100;
    }
  }
  
  {
    QIcon appIcon(QString(resourcePath) + "/GRALE.xpm");
    app.setWindowIcon(appIcon);
  }

  _CmdLineArgs cmdLineArgs(argc, argv, allowedArgs);
  _CmdLineArgs *args = &cmdLineArgs;
  
  globalOptions = new Options;
  globalOptions->writeOptionToConfig();

  dataMgr = NULL;  // The data manager is initialized in the constructor of the main window.

  mainw         = new MainWindow;
  mainw->show();
  
  dataxc = new GDataExchange(dataMgr);
  QObject::connect(dataxc, SIGNAL(connectionStateSignal(unsigned char, char*, int)),
                   mainw, SLOT(connectionStateSlot(unsigned char, char*, int)));
  
  if (args->isSet("dump"))
    dumpData = true;

  if (args->isSet("net"))
    useTCP = true;
  
  // always use tcp
  // until we get the file server on windows going
  useTCP = true;

  for (unsigned i = 0; i < args->count(); i++)
    dataMgr->loadDataFile(args->arg(i));

  if (useTCP == true)
    dataxc->setupInetServerSocket();
  else
    dataxc->setupFileServerSocket();
  
  optionChanged = false;

  // the main loop
  int ret = app.exec();
  
  // Never forget to remove window-structures from X-Windows
  // ... but NOT with Qt 2, where widgets are deleted automatically...
  //delete mainw;
  
  dataxc->freeResources();
  
  if (optionChanged) {
    if (QMessageBox::question(NULL, "Modified Options",
				    "You have changed several options during this GRALE session.\n\n"
				    "Would you like to make the new settings permanent?",
				    QMessageBox::Yes | QMessageBox::Default, QMessageBox::No
				    ) == QMessageBox::Yes)
      globalOptions->writeOptionToConfig();
  }
  
  delete globalOptions;
  free((void*)resourcePath);
  
  return ret;
}
