%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%   $RCSfile: grammar_interface.pl,v $
%%  $Revision: 1.14 $
%%      $Date: 2006/09/18 08:22:43 $
%%     Author: Detmar Meurers (dm@ling.osu.edu) with additions 
%%             by Stefan Mueller (stefan.mueller@uni-bremen.de)
%%    Purpose: Code to provide a user definable grammar interface
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 
%
% Define a way for user to specify what features are used 
% in a given grammar, e.g. for encoding the phonology and 
% the daughters of a phrase - and provide the defaults phon
% and dtrs if nothing is defined.
%
% -----------------------------------
%
% The following predicates can be specified in a grammar:
% 
% 1) For the feature reflecting the properties encoded in the
%    phrase structure (i.e. extra-logically):
% 
%    dtrs_feat(<featurename>).
%    phon_feat(<featurename>).
%
% 2) For finding the semantics (relevant for comparing analyses): 
%
%   cont_path(<path>). (where <path> is a Prolog list of <featurename>)
%   event_type(<typename>).
%   index_type(<typename>).
%
% 3) In principle the user can also define the following
%    list types and features, but the trale code largely
%    has those hardwired so it's not advisable to change 
%    it until we get around to changing all old code
%    to take these predicates into account.
%
% list_type(<typename>).
% ne_list_type(<typename>).
% hd_feat(<featurename>).
% tl_feat(<featurename>).
%
% -----------------------------------
%
% The internal predicates for querying these user definitions
% have the same predicate name, plus a trailing _, i.e.
% the dtrs feature can be obtained by calling dtrs_feat_/1.
% A default is provided for all these predicates, so that
% the return a value regardless of whether the user specified
% them or not.
% ------------------------------------------------------

dtrs_feat_(Feat) :-
   ( current_predicate(dtrs_feat,dtrs_feat(_)) ->
      dtrs_feat(Feat)
   ;   %% The default if the user hasn't defined dtrs_feat/1
      Feat = dtrs
   ).

phon_feat_(Feat) :-
   ( current_predicate(phon_feat,phon_feat(_)) ->
      phon_feat(Feat)
   ;   %% The default if the user hasn't defined phon_feat/1
      Feat = phon
   ).

% ------------------------------------------------------

list_type_(Type) :-
   ( current_predicate(list_type,list_type(_)) ->
      list_type(Type)
   ;   %% The default if the user hasn't defined list_type/1
      Type = list
   ).

ne_list_type_(Type) :-
   ( current_predicate(ne_list_type,ne_list_type(_)) ->
      ne_list_type(Type)
   ;   %% The default if the user hasn't defined ne_list_type/1
      Type = ne_list
   ).

e_list_type_(Type) :-
   ( current_predicate(e_list_type,e_list_type(_)) ->
      e_list_type(Type)
   ;   %% The default if the user hasn't defined e_list_type/1
      Type = e_list
   ).

hd_feat_(Feat) :-
   ( current_predicate(hd_feat,hd_feat(_)) ->
      hd_feat(Feat)
   ;   %% The default if the user hasn't defined hd_feat/1
      Feat = hd
   ).

tl_feat_(Feat) :-
   ( current_predicate(tl_feat,tl_feat(_)) ->
      tl_feat(Feat)
   ;   %% The default if the user hasn't defined tl_feat/1
      Feat = tl
   ).

cont_path_(Path) :-
   ( current_predicate(cont_path,cont_path(_)) ->
      cont_path(Path)
   ;   %% The default if the user hasn't defined cont_path/1
      Path = [synsem,loc,cont]
   ).

gtop_path_(Path) :-
   ( current_predicate(gtop_path,gtop_path(_)) ->
      gtop_path(Path)
   ;   %% The default if the user hasn't defined gtop_path/1
      Path = [synsem,loc,cont,hook,gtop]
   ).

ind_path_(Path) :-
   ( current_predicate(ind_path,ind_path(_)) ->
      ind_path(Path)
   ;   %% The default if the user hasn't defined ind_path/1
      Path = [synsem,loc,cont,hook,ind]
   ).

liszt_path_(Path) :-
   ( current_predicate(liszt_path,liszt_path(_)) ->
      liszt_path(Path)
   ;   %% The default if the user hasn't defined cont_path/1
      Path = [synsem,loc,cont,liszt]
   ).

hcons_path_(Path) :-
   ( current_predicate(hcons_path,hcons_path(_)) ->
      hcons_path(Path)
   ;   %% The default if the user hasn't defined cont_path/1
      Path = [synsem,loc,cont,h_cons]
   ).

event_type_(Type) :-
   ( current_predicate(event_type,event_type(_)) ->
      event_type(Type)
   ;   %% The default if the user hasn't defined event_type/1
      Type = event
   ).

index_type_(Type) :-
   ( current_predicate(index_type,index_type(_)) ->
      index_type(Type)
   ;   %% The default if the user hasn't defined index_type/1
      Type = index
   ).

% The LKB needs a certain abbreviated format for MRS representations.
% event -> e, ref_ind -> x,  event_or_index -> i, bot -> u

% (defparameter *handle-type* (vsym "handle"))
% (defparameter *event_or_index-type* (vsym "event_or_index"))
% (defparameter *event-type* (vsym "event"))
% (defparameter *ref-ind-type* (vsym "ref-ind"))

event_or_index_type_(Type) :-
   ( current_predicate(event_or_index_type,event_or_index_type(_)) ->
      event_or_index_type(Type)
   ;   %% The default if the user hasn't defined event_or_index_type/1
      Type = event_or_index
   ).

ref_ind_type_(Type) :-
   ( current_predicate(ref_ind_type,ref_ind_type(_)) ->
      ref_ind_type(Type)
   ;   %% The default if the user hasn't defined ref_ind_type/1
      Type = ref
   ).



pred_feat_(Feat) :-
   ( current_predicate(pred_feat,pred_feat(_)) ->
      pred_feat(Feat)
   ;   %% The default if the user hasn't defined pred_feat/1
      Feat = pred
   ).

ind_feat_(Feat) :-
   ( current_predicate(ind_feat,ind_feat(_)) ->
      ind_feat(Feat)
   ;   %% The default if the user hasn't defined ind_feat/1
      Feat = ind
   ).

per_feat_(Feat) :-
   ( current_predicate(per_feat,per_feat(_)) ->
      per_feat(Feat)
   ;   %% The default if the user hasn't defined per_feat/1
      Feat = per
   ).

per_type_(Type) :-
   ( current_predicate(per_type,per_type(_)) ->
      per_type(Type)
   ;   %% The default if the user hasn't defined per_type/1
      Type = per
   ).

quant_type_(Type) :-
   ( current_predicate(quant_type,quant_type(_)) ->
      quant_type(Type)
   ;   %% The default if the user hasn't defined quant_type/1
      Type = quant_rel
   ).

label_feat_(Feat) :-
   ( current_predicate(label_feat,label_feat(_)) ->
      label_feat(Feat)
   ;   %% The default if the user hasn't defined label_feat/1
      Feat = lbl
   ).

top_feat_(Feat) :-
   ( current_predicate(top_feat,top_feat(_)) ->
      top_feat(Feat)
   ;   %% The default if the user hasn't defined top_feat/1
      Feat = ltop
   ).

outscoped_feat_(Feat) :-
   ( current_predicate(outscoped_feat,outscoped_feat(_)) ->
      outscoped_feat(Feat)
   ;   %% The default if the user hasn't defined outscoped_feat/1
      Feat = outscoped
   ).

sc_arg_feat_(Feat) :-
   ( current_predicate(sc_arg_feat,sc_arg_feat(_)) ->
      sc_arg_feat(Feat)
   ;   %% The default if the user hasn't defined sc_arg_feat/1
      Feat = sc_arg
   ).

gen_pathes_(Pathes) :-
   ( current_predicate(gen_pathes,gen_pathes(_)) ->
      gen_pathes(Pathes)
   ;   %% The default if the user hasn't defined gen_pathes/1
      Pathes = [[synsem,loc,cat,head],[synsem,loc,cat,subcat],[synsem,loc,cont]]
   ).

scopable_description_(Scopable) :-
   ( current_predicate(scopable_description,scopable_description(_)) ->
      scopable_description(Scopable)
   ;   %% The default if the user hasn't defined scopable_description/1
      Scopable = []
   ).



%%% ==================================================================
% Local Variables:
% mode: prolog
% End:

