/*  $Id$

    This file is part of the

    GRammar vISUalization tool (GRISU)

    Developed by Holger Wunsch and Detmar Meurers
    Copyright (C) 1998-2005 by the developers
    All rights reserved

    Seminar fr Sprachwissenschaft, Universitt Tbingen and
    Department of Linguistics, The Ohio State University

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/**
   @file

   Declarations for the <code>GDataHolder</code> class.

   @author Holger Wunsch */


#ifndef __GDATAHOLDER_H
#define __GDATAHOLDER_H


#include <qptrlist.h>
#include <qwidget.h>

#include "gdataviewer.h"
#include "gstructuredata.h"
#include "gtypehierarchydata.h"
#include "vcanvas.h"
#include "gsvgwindow.h"


/* Supported data package types. */
#define DT_FEATURE_STRUCTURE 1  /**< The data package is a feature structure. */
#define DT_TYPE_HIERARCHY    2  /**< The data package is a type hierarchy. */


/** This class implements a general container class for all types
    of data packages supported by Grisu. It provides methods to query the type of
    data package contained in this holder, as well as type-safe accessor
    methods.

    @author Holger Wunsch */
class GDataHolder {
  /** A generic pointer to the data package. */
  void* data;


  /** The type of the data package, a value which can be one of the
      <code>DT_xxxx</code> constants. */
  int datatype;


  /** A list of viewers displaying the data package contained in this
      holder that are currently open. */
  QPtrList<GDataViewer>* viewers;


  /** Opens a viewer for feature structures.
      @param parent the parent widget.
      @param showFlags if <code>DRAW_AUTO</code> is passed in this argument, a tree will be
        displayed initially if tree data is available, otherwise a feature structure is displayed.
        <code>DRAW_STRUC</code> displays a feature structure, <code>DRAW_TREE</code> displays
        a tree.
      @return the new viewer. */
  ViewCanvas* openFeatureStructureViewer(QWidget* parent, int showFlags);

  
  /** Opens a viewer for type hierarchies.
      @param parent the parent widget.
      @param showFlags currently not used for type hierarchies.
      @return the new viewer. */
  GSVGWindow* openTypeHierarchyViewer(QWidget* parent, int showFlags);


 public:
  /** Creates a new data holder that contains a feature structure.
      @param data the feature structure data package. */
  GDataHolder(GStructureData* data);


  /** Creates a new data holder that contains a type hierarchy.
      @param data the type hierarchy data package. */
  GDataHolder(GTypeHierarchyData* data);


  /** Deletes the data holder and the contained data package. */
  ~GDataHolder(void);


  /** Returns the type of the data package contained in this holder.
      @return the type of the data package. */
  int type(void);


  /** Returns the name of the data package. */
  const QString getName(void);


  /** Returns a feature structure data package.
      @return a pointer to the feature structure data package, or <code>NULL</code>
      if the data package contained in this holder is not a feature structure. */
  GStructureData* featureStructure(void);


  /** Returns a type hierarchy data package.
      @return a pointer to the type hierarchy data package, or <code>NULL</code>
      if the data package contained in this holder is not a type hierarchy. */
  GTypeHierarchyData* typeHierarchy(void); 


  /** Opens a viewer for the data package
      @param parent the parent widget.
      @param showFlags data package dependent information about how the data is shown.
      
      For feature structures, this can be a <code>DRAW_xxxx</code> value. */
  GDataViewer* openViewer(QWidget* parent, int showFlags);


  /** Closes a viewer.
      @param index the index of the viewer. */
  void closeViewer(int index);


  /** Closes all viewers. */
  void closeAllViewers(void);


  /** Returns the number of open viewers for this data package.
      @return the number of open viewers. */
  int numViewers(void);


  /** Checks if a given viewer is belongs to this data holder.
      @param viewer the viewer to be checked
      @return <code>true</code> if the viewer belongs to this data holder,
      <code>false</code> otherwise. */
  bool hasViewer(GDataViewer* viewer);


  /** Saves the data package. This method supports saving of multiple
      data packages in one shot. 
      
      This method displays a "Save As" dialog if the <code>basename</code>
      argument is <code>NULL</code>. In this case, only the data package
      contained in this holder will be saved. The dialog will display the
      right choice of file extensions according to the data type of the
      data package to be saved.
      
      When the <code>basename</code> is non-null, this method will name
      the file by appending <code>number</code> to <code>basename</code>,
      then adding the default extension for the data package.
      @param parent the parent widget.
      @param basename the basic filename when used while saving multiple
        data packages, or <code>NULL</code>, if only this data package is
        to be saved.
      @param number the number of this data package. The value of
        <code>number</code> is ignored if <code>basename</code> is <code>NULL</code>. */
  void save(QWidget* parent, const QString& basename = QString::null, int number = -1);


  /** Saves a feature structure. See also <code>save</code>.
      @param parent the parent widget.
      @param basename the basic filename.
      @param number the number of this data package. */
  void saveFeatureStructure(QWidget* parent, const QString& basename, int number);


  /** Saves a type hierarchy. See also <code>save</code>.
      @param parent the parent widget.
      @param basename the basic filename.
      @param number the number of this data package. */
  void saveTypeHierarchy(QWidget* parent, const QString& basename, int number);


  /** Removes a viewer that was just closed from the list of currently open viewers.
      @param viewer a pointer to the viewer to be removed. */
   void viewerClosed(GDataViewer* viewer); 
};


#endif
