/*  $Id$

    This file is part of the

    GRammar vISUalization tool (GRISU)

    Developed by Holger Wunsch and Detmar Meurers
    Copyright (C) 1998-2005 by the developers
    All rights reserved

    Seminar fr Sprachwissenschaft, Universitt Tbingen and
    Department of Linguistics, The Ohio State University

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


/** @file

    Implementation of class <code>GDataManager</code>.

    @author Holger Wunsch */


#include <qwidget.h>
#include <qmessagebox.h>
#include <qfiledialog.h>
#include <kfiledialog.h>

#include "gdatamanager.h"
#include "dataparser.h"
#include "tools.h"
#include "global.h"

#include "gdatamanager.moc"


/** Creates a new empty data manager.
    @param parent pointer to a widget which will be the parent widget of dialog boxes
    displayed by the data manager, as well as the receiver of update messages. */
GDataManager::GDataManager(QWidget* parent):
  QObject() {
  data = new QPtrList<GDataHolder>;
  this->parent = parent;
}


/** Destroys all data packages managed by this data manager and this data
    manager itself. */
GDataManager::~GDataManager(void) {
  for (int i = data->count() - 1; i >= 0; i--)
    delete data->at(i);

  delete data;
}


/** Deletes this instance. This should be connected to <code>KApplication::shutDown()</code>,
    such that viewer windows can be properly deleted. Viewer windows, which are opened by
    instances of <code>GDataHolder</code>, are created as children of the application's main window, and
    deleted when the application shuts down. However, a data holder will also delete its
    dependent viewer windows when it is deleted itself. This would cause a double deletion of
    the viewer windows if the data holder instances are deleted after the main window. 
    Since viewer windows will automatically deregister themselves from the
    main window when deleted prior to the main window, all data holders and dependent
    viewers <b>must</b> be deleted first. By connecting this slot to 
    <code>KApplication::shutdown()</code>, this requirement can be fulfilled. */
void GDataManager::deleteSlot(void) {
  delete this;
}


/** Inserts a new data package in the data manager.
    @param newData the new data package. */
void GDataManager::addData(GDataHolder* newData) {
  data->append(newData);
}


/** Returns the data package indexed by <code>index</code>.
    @param index the index. */
GDataHolder* GDataManager::getData(int index) {
  return data->at(index);
}


/** Deletes a data package.
    @param index the index of the data package to be deleted.
    @param notify if this argument is <code>true</code>, a <code>dataModifiedSignal()</code>
    will be emitted. If <code>false</code>, no signal will be emitted. */
void GDataManager::deleteData(int index, bool notify) {
  delete data->at(index);
  data->remove(index);

  if (notify)
    emit dataModifiedSignal();
}


/** Deletes the data packages indexed by <code>indices</code>.
    @param indices the indices of the data packages to be deleted. */
void GDataManager::deleteData(QValueVector<int>& indices) {
  qHeapSort(indices.begin(), indices.end());
  for (int i = indices.size() - 1; i >= 0; i--)
    deleteData(indices[i], false);

  emit dataModifiedSignal();
}
    

/** Deletes all data packages. */
void GDataManager::deleteAllData(void) {
  for (int i = data->count() - 1; i >= 0; i--)
    deleteData(i, false);

  emit dataModifiedSignal();
}


/** Returns the number of available data packages. */
int GDataManager::numData(void) {
  return data->count();
}


/** Saves a data package to a file. 
    @param index the index of the data package to be saved. */
void GDataManager::saveData(int index) {
  getData(index)->save(parent);
};


/** Saves multiple data packages at once.
    @param indices a vector of the indices of the data packages to be saved. */
void GDataManager::saveData(QValueVector<int>& indices) {
  QString basename = getSaveFilename(parent, "Batch Save", false, "All files (*)", "*|All files (*)");
  
  if (!basename.isNull()) {
    for (int i = 0; i < indices.size(); i++)
      getData(indices[i])->save(parent, basename, i);
  }
}


/** This method is called by an instance of <code>GDataExchange</code> when
    new data has become available.
    
    It is not guaranteed that <code>buffer</code> remains valid after a call
    to this method.
    @param buffer the buffer that contains the incoming data. */
void GDataManager::processData(GBuffer* buffer) {
  char* line;

  while ((line = (char*) buffer->readLine()) != NULL) {
    
    // Dump all incoming data to a file if requested
    if (dumpData) {
      FILE* dumpFile = fopen(DATADUMP_FILENAME, "a");
      fputs(line, dumpFile);
      fclose(dumpFile);
    }

    // Special command: quit Grisu
    if (strncmp(line, "!quit", 5) == 0) {
      delete [] line;
      parent->close();
      return;
    }

    parseData(line);

    delete [] line;
  }
}


/** Displays a file dialog which allows the user to pick a file; the data
    contained in that file is loaded afterwards.
    @param parent the parent widget. */
void GDataManager::loadData(QWidget* parent) {
  if (globalOptions->useKFileDialog) {
    KFileDialog fileDlg(NULL, "*.grisu|Grisu data files (*.grisu)\n*|All files (*)", parent, "openfile", true);
    fileDlg.setCaption("Load Data"); 
    fileDlg.setMode(KFile::Files);
    fileDlg.setIcon(application->icon());
    if (fileDlg.exec() == QDialog::Accepted) {
      QStringList Filenames = fileDlg.selectedFiles();
      for (int i = 0; i < Filenames.count(); i++)
	loadDataFile(Filenames[i]);
    }
  }
  else {
    QFileDialog fileDlg("*.grisu", "Grisu data files (*.grisu);;All files (*)", parent, "openfile", true);
    fileDlg.setCaption("Load Data"); 
    fileDlg.setMode(QFileDialog::ExistingFiles);
    fileDlg.setIcon(application->icon());
    if (fileDlg.exec() == QDialog::Accepted) {
      QStringList Filenames = fileDlg.selectedFiles();
      for (int i = 0; i < Filenames.count(); i++)
	loadDataFile(Filenames[i]);
    }
  }
}


/** Loads a GRISU data file and creates a new view window, depending on the user-settings.
    @param fileName the name of the data file. */
void GDataManager::loadDataFile(const QString& fileName) {
  FILE* file = fopen((const char*) fileName.local8Bit(), "r");
  
  if (file) {
    parent->setCursor(waitCursor);
    char* line = new char[250000];
    GStructureData* newDoc;
    
    fgets(line, 250000, file);

    parseData(line);
    
    delete [] line;
    
    fclose(file);
    
    parent->setCursor(arrowCursor);
  }
  else {
    char buf[500];
    sprintf(buf, "Could not open file %s", (const char*) fileName.local8Bit());
    QMessageBox::critical(parent, "Save", buf);
  }
}


/** Reads a data package and creates a new document from it.
    Depending on the user setting, a new viewer is opened.
    @param line the line containing the data package. */
void GDataManager::parseData(char* line) {
  // SVG Data
  if (strncmp(line, "!show_svg_file", strlen("!show_svg_file")) == 0) {
    char* ptr = line + strlen("!show_svg_file");
    QString Filename;
    while (ptr && *ptr != '\"')
      ptr++;
    if (*ptr == '\"') {
      ptr++;
      while (ptr && *ptr != '\"') {
	Filename.append(*ptr);
	ptr++;
      }
      if (QFile::exists(Filename)) {
	GTypeHierarchyData* newData = new GTypeHierarchyData(Filename, Filename);
	GDataHolder* holder = new GDataHolder(newData);
	addData(holder);
	holder->openViewer(parent, 0);
      }
      else
	QMessageBox::information(parent, "Error", "SVG file not found.", QMessageBox::Ok);
    }
  }
  // Feature Structure Data
  else {
    GStructureData* newDoc = DataParser().parse(line);
    if (newDoc) {
      GDataHolder* holder = new GDataHolder(newDoc);
      addData(holder);
      if (globalOptions->autoOpenWindows) {
	if (newDoc->showDefault != 0)
	  holder->openViewer(parent, newDoc->showDefault);
	else {
	  if (newDoc->tree)
	    holder->openViewer(parent, DRAW_TREE);
	  else
	    holder->openViewer(parent, DRAW_STRUC);
	}
      }
    }
  }
  emit dataModifiedSignal();
  emit viewerSignal();
}


/** Returns the total number of open viewers (that is, all open viewers for
    all available data packages.
    @return the number of open viewers. */
int GDataManager::numViewers(void) {
  int count = 0;

  GDataHolder* holder = data->first();

  while (holder != NULL) {
    count += holder->numViewers();
    holder = data->next();
  }

  return count;
}


/** Closes all open viewers. */
void GDataManager::closeAllViewers(void) {
  
  GDataHolder* holder;

  for (int i = 0; i < data->count(); i++) {
    holder = data->at(i);
    holder->closeAllViewers();
  }
}
