%% ==================================================================
%% $Id: macro_utils.pl,v 1.12 2003/05/08 16:41:46 dm Exp $
% GBP - as far as I can tell, nothing in the rest of the system depends on this
%%
%% Utilities for working with TRALE macros
%%
%% Author: Detmar Meurers
%% Ported to SP4 by: Gerald Penn
%%  
%% Macros can be hierarchically organized by calling one macro
%% in the definition of another. The macro X occuring in the definition
%% of a macro Y then can be referred to as a supermacro of X.
%% And conversely, Y is a submacro of macro X.
%%
%% The notions of sub- and supermacro thus in one sense are parallel to the 
%% notion of sub- and supertype. BUT IT IS IMPORTANT TO KEEP IN MIND THAT
%% ONTOLOGICALLY MACROS AND TYPES ARE VERY DIFFERENT; in 
%% particular an object of a type will also be of one of its subtypes,
%% and of exactly one of its most specific subtypes. There is no 
%% equivalent to this with macro hierarchies, which are just
%% subsumption hierarchies of some descriptions that were given names.
%% Different from types, macros have no theoretical status; they just
%% serve to write down a theory more compactly.
%%
%%
%% In terms of the macro hiearchy comands below, note note that 
%% (parallel to types) the sub- and supermacro relations 
%% only include macros on the same level, not those embedded under 
%% features.
%%
%% This file provides the following top-level predicates, where
%% <(sub/super)macro> is the macro name (incl. its argument slots).
%% Many of the predicates exist in two version, one that returns
%% single results and can be backtracked into, and the other 
%% (same predicate name, but ending in s) which returns the list
%% of all results. Note that the list returned by the second
%% kind of predicates is sorted though. Also, it is worth noting
%% that the predicates returning a list will always succeed (they
%% return a [] in the case where setof would fail).
%%
%%
%% submacro(<macro>,<submacro>).
%% submacros(<macro>,<list(submacros)).
%%
%% supermacro(<macro>,<supermacro>).
%% supermacros(<macro>,<list(supermacros)>).
%%
%% show_submacros(<macro>).
%% show_all_submacros(<macro>).
%%
%% show_supermacros(<macro>).
%% show_all_supermacros(<macro>).
%%
%% show_all_macros.
%%    shows the entire macro hierarchy
%%
%% macro(<macro>)
%%    shows the most general satisfier of the description abbreviated 
%%    by the macro  (predicate provided by core ale.pl)
%%    
%% is_macro(<macro>)
%%    returns every macro that's defined (if tracked back into)
%% macros(<list(macro)>)
%%    returns list of macros that are defined
%%
%% most_specific_macro(<macro>)
%% most_specific_macros(<list(macro)>)
%%
%% most_general_macro(<macro>)
%% most_general_macros(<list(macro)>)
%%
%% isolated_macro(<macro>)
%% isolated_macros(<list(macro)>)
%%    Isolated macros are those that are most general and most
%%    specific at the same time, i.e. they neither occur in other 
%%    macro definitions nor are they defined in terms of other macros.
%%
%% ==================================================================

% SP4: library(assoc) is now library(avl)
:- use_module(library(avl),[avl_to_list/2,avl_store/4,avl_delete/4,
                              empty_avl/1,avl_fetch/3]).
%:- use_module(library(lists),[memberchk/2]). -- SP4: built-in

% -------------------------------------------------------------
% Fundamentals:

% --------------------------------------------------
% Supermacros

supermacro(M,SupM) :-
   supermacros(M,SupMs),
   member(SupM,SupMs).

% supermacros(+macroname,-list(supermacronames)) 
% get immediate supermacros of a macro, where a supermacro is a macro 
% that is referred to in the definition of a macro
supermacros(X,SortL) :-		% will backtrack for disjunctions!
   macro(X,RHS),
   parens2list(RHS,L),
   sort(L,SortL).

parens2list(@(X),[X]):- !.
parens2list((H,T),L) :- !,
   (H = @(NewH) 
   -> L=[NewH|NewT]
   ;  L=NewT
   ),
   parens2list(T,NewT).
parens2list((X;Y),L) :-		% disjunction backtracks here
   parens2list(X,L)
   ;
   parens2list(Y,L).
parens2list(_,[]).

% ---------------------

supermacros_assoc(M,Hier) :-
   empty_avl(Hier0),
   supermacros(M,SuperMs),
   supermacros_aux_assoc(SuperMs,M,Hier0,Hier).

supermacros_aux_assoc([],_,Hier,Hier).
supermacros_aux_assoc([M|Ms],SubM,Hier0,Hier) :-
   avl_store(M,Hier0,[SubM],Hier1),
   supermacros_aux_assoc(Ms,SubM,Hier1,Hier).

% ---------------------

% get any supermacros directly or indirectly included in a macros definition
all_supermacros_assoc(M,Hier) :-
   empty_avl(Hier0),
   all_supermacros_assoc(M,Hier0,Hier).

all_supermacros_assoc(M,Hier0,Hier) :-
   supermacros(M,SuperMs),
   all_supermacros_aux_assoc(SuperMs,M,Hier0,Hier).

all_supermacros_aux_assoc([],_,Hier,Hier).
all_supermacros_aux_assoc([M|Ms],SubM,Hier0,Hier) :-
   (   avl_delete(M,Hier0,OtherMs,Hier1) % if already occurred as supermacro
   ->  avl_store(M,Hier1,[SubM|OtherMs],Hier2) % then add macro to its submacros
   ;   avl_store(M,Hier0,[SubM],Hier1), % else record new supermacro with macro
       all_supermacros_assoc(M,Hier1,Hier2)% and explore supermacros of supermacro
   ),
   all_supermacros_aux_assoc(Ms,SubM,Hier2,Hier).

% --------------------------------------------------
% Submacros

submacro(Macro,SubMacro) :-
   supermacros(SubMacro,SupMacros),
   member(Macro,SupMacros).   

% submacros(+macroname,-list(submacronames))
% (can't use setof since we want [] in case there are no submacros)
submacros(Macro,SortSubMacros) :-
   (   var(Macro)
   ->  is_macro(Macro)
   ;   true
   ),
   findall(SubMacro,submacro(Macro,SubMacro),SubMacros),
   sort(SubMacros,SortSubMacros).
   
% ---------------------

submacros_assoc(M,Hier) :-
   empty_avl(Hier0),
   submacros(M,SubMs),
   avl_store(M,Hier0,SubMs,Hier).

% ---------------------

% get any submacros
all_submacros_assoc(M,Hier) :-
   empty_avl(Hier0),
   all_submacros_assoc(M,Hier0,Hier).

all_submacros_assoc(M,Hier0,Hier) :-
   submacros(M,SubMs),
   avl_store(M,Hier0,SubMs,Hier1),
   all_submacros_assoc_list(SubMs,Hier1,Hier).

all_submacros_assoc_list([],Hier,Hier).
all_submacros_assoc_list([M|Ms],Hier0,Hier) :-
   all_submacros_assoc(M,Hier0,Hier1),
   all_submacros_assoc_list(Ms,Hier1,Hier).

% -------------------------
% Macros in general

is_macro(M) :-			% note: macro/1 is used in ale.pl
   macro(M,_).
macros(SortL) :-
   findall(M,is_macro(M),L),
   sort(L,SortL).

% -------------------------
% Macros that aren't called by other macros:

most_specific_macro(M) :-
   submacros(M,[]).
most_specific_macros(SortL) :-
   findall(M,most_specific_macro(M),L),
   sort(L,SortL).

% -------------------------
% Macros that don't call any other macros

most_general_macro(M) :-
   supermacros(M,[]).
most_general_macros(SortL) :-
   findall(M,most_general_macro(M),L),
   sort(L,SortL).

% ---

% most general macros in hierarchy provided as assoc
most_general_in_hierarchy(Hier,L) :-
   avl_to_list(Hier,HierList),
   get_dominated_in_hierarchy(HierList,[],Specific),
   setof(M,Sub^(member(M-Sub,HierList),
	    \+ (memberchk(M,Specific))),L).

get_dominated_in_hierarchy([],R,R).
get_dominated_in_hierarchy([_-List|T],Result0,Result) :-
   append(List,Result0,Result1),
   get_dominated_in_hierarchy(T,Result1,Result).

% -------------------------
% Macros that neither occur in other macro definitions
% nor are defined in terms of other macros:

isolated_macro(M) :-
   most_specific_macro(M),
   most_general_macro(M).
isolated_macros(SortL) :-
   findall(M,isolted_macro(M),L),
   sort(L,SortL).

% -------------------------------------------------------------
% Output

% -------------------------
% supermacros

show_supermacros(Macro) :-
   ensure_macro_defined(Macro),
   supermacros_assoc(Macro,Hier),
   most_general_in_hierarchy(Hier,L),
   format('~nHierarchy of immediate supermacros of "~w":~n~n',[Macro]),
   print_assoc_hierarchy(L,Hier,true),nl.


show_all_supermacros(Macro) :-
   ensure_macro_defined(Macro),
   all_supermacros_assoc(Macro,Hier),
   most_general_in_hierarchy(Hier,L),
   format('~nHierarchy of all supermacros of "~w":~n~n',[Macro]),
   print_assoc_hierarchy(L,Hier,true),nl.

% -------------------------
% submacros

show_submacros(Macro) :-
   ensure_macro_defined(Macro),
   submacros_assoc(Macro,Hier),
   format('~nHierarchy of immediate submacros of "~w":~n~n',[Macro]),
   print_assoc_hierarchy([Macro],Hier,false),nl.


show_all_submacros(Macro) :-
   ensure_macro_defined(Macro),
   all_submacros_assoc(Macro,Hier),
   format('~nHierarchy of all submacros of "~w":~n~n',[Macro]),
   print_assoc_hierarchy([Macro],Hier,false),nl.

% -------------------------
% entire hierarchy

show_all_macros :-
   ensure_macros_present,
   most_general_macros(L),
   empty_avl(Hier0),
   all_submacros_assoc_list(L,Hier0,Hier),
   format('~nComplete macro hierarchy~n~n',[]),
   print_assoc_hierarchy(L,Hier,false),nl.

% -------------------------

print_assoc_hierarchy(L,Hier,MarkLeaf) :-
   (  L = []
   -> format(' None defined.~n',[])
   ;  empty_avl(Empty),
      print_assoc_hierarchy(L,Hier,1,Empty,_,MarkLeaf) ).

print_assoc_hierarchy([],_,_,Vis,Vis,_).
print_assoc_hierarchy([Macro|Sisters],Hier,Tab0,Vis0,Vis,MarkLeaf) :-
   (   avl_fetch(Macro,Hier,Submacros) 
   ->  Leaf=false
   ;   Leaf=true,Submacros=[]
   ),
   (   avl_fetch(Macro,Vis0,_) 	% seen this macro before
   ->  print_indent_macro(Macro,Tab0,Tab1,true,Leaf,MarkLeaf),
       Vis2=Vis0
   ;   avl_store(Macro,Vis0,_,Vis1),
       print_indent_macro(Macro,Tab0,Tab1,false,Leaf,MarkLeaf),
       print_assoc_hierarchy(Submacros,Hier,Tab1,Vis1,Vis2,MarkLeaf)
   ),
   print_assoc_hierarchy(Sisters,Hier,Tab0,Vis2,Vis,MarkLeaf).

print_indent_macro(X,Tab0,Tab,Multi,Leaf,MarkLeaf) :-
   tab(Tab0),
   (Multi -> write('&');true),
   (   (Leaf, MarkLeaf)
   ->  X=..[Functor|_],
       capitalize_atom(Functor,NewX)
   ;   NewX=X 
   ),
   write(NewX),nl,
   Tab is Tab0 + 3.

% -------------------------

% check_lmacros/0: tests logical variable macros for consistency

check_lmacros :- check_lmacros(false).
check_lmacros_verbose :- check_lmacros(true).

check_lmacros(Debug) :-
   fail_if_undefined((LHS := Desc)),
   functor(LHS,Functor,Arity),
   secret_noadderrs,				    
   findall(RLength,  (call_residue((add_to(Desc,Tag,bot),
				    deref(Tag,bot,Ref,SVs),
				    extensionalise(Ref,SVs)),Residue),
			length(Residue,RLength)),Results),
   secret_adderrs,
   length(Results,NumResults),
   (  Debug
   -> format('~N Testing: ~w/~w',[Functor,Arity])
   ;  write('.')
   ),
   (  NumResults == 0
   -> format(user_error,'~N~n**Error: ~w/~w inconsistent!~n~n',[Functor,Arity])
   ;  (  NumResults @> 1
       -> (  member(R,Results), R @> 0 
	  -> format('~NDisjunctive & Residues: ~w/~w has ~w solutions, number of residues: ~w',[Functor,Arity,NumResults,Results]) 
	  ;  format('~NDisjunctive:            ~w/~w has ~w solutions',[Functor,Arity,NumResults]))
       ;  (  member(R,Results), R @> 0 
	  -> format('~N              Residues: ~w/~w has ~w residues.',[Functor,Arity,R]) ; true)
     )
   ),
   fail
   ;
   true.

% -------------------------

ensure_macros_present :-
   (macro(_,_) 
   ->  true
   ;   format(user_error,'~N**Error: No macros defined!~n',[]),
       fail
   ).

ensure_macro_defined(Name) :-
   ensure_macros_present,
   (is_macro(Name) 
   ->  true
   ;   format(user_error,'~N**Error: Macro "~w" not defined.~n',[Name]),
       fail
   ).
