%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%   $RCSfile: pth.pl,v $
%%  $Revision: 1.26 $
%%      $Date: 2003/05/13 05:20:43 $
%%     Author: Stephan Kepser, Dale Gerdemann, Thilo Goetz,
%              Detmar Meurers (dm@ling.osu.edu)
%%    Purpose: Type hierarchy reader and additional signature 
%%             commands for Trale
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Type hierarchy reader taken from the type hierarchy compiler 
%% of the ConTroll system written by Stephan Kepser, Dale Gerdemann, 
%% and Thilo Goetz. Additional type hierarchy commands for Trale at
%% bottom of file written by Detmar Meurers.
%%
%% $Log: pth.pl,v $
%
% Bug corrected: sub/2 referred to as subs/2 in warning message.
% 12/9/06 - G. Penn
%
%% Revision 1.26  2003/05/13 05:20:43  dm
%% Added possibility of having only an ALE style signature declaration
%% using subs/2 and intro/2 without a controll style signature file.
%%
%% Revision 1.25  2003/05/13 05:08:20  dm
%% removed additional newline output in case of multiple inheritance
%%
%% Revision 1.24  2003/05/13 05:05:41  dm
%% changed indentation of printout to start at column 0 instead of 1
%%
%% Revision 1.23  2003/05/13 05:01:49  dm
%% added show_signature/1 and show_signature/0
%%
%% Revision 1.22  2003/05/08 16:37:16  dm
%% changed error messages to go to user_error
%%
%% Revision 1.21  2003/05/01 15:58:32  dm
%% removed grammar search path code since a simpler solution is to start
%% compilation of grammar from a file in the grammar directory (Prolog by
%% default interprets a filename relative to the directory of the file in
%% which it is loaded).
%%
%% Revision 1.20  2003/05/01 13:59:41  dm
%% Version with grammar search path.
%%
%% Revision 1.19  2003/05/01 04:07:49  dm
%% Add lookup of use_grammar_path/2 (if exists - which allows use of
%% pth.pl with just the core trale, i.e. without loading
%% trale_patches.pl) and use of absolute_file_name/2.
%%
%% Revision 1.18  2002/11/25 23:53:10  dm
%% Added a seen/0 call to the beginning of input_ths/4 to fix bug that
%% arose when user types Cntrl-C after Reading signature signature
%% file...  appears but before the name of the signature file read in
%% appears on the user output. In that case, for the next compilation one
%% got the error "Warning: no type hierarchy found" because the see
%% stream was still open.
%%
%% Revision 1.17  2002/11/22 23:37:02  dm
%% Added check for existence of files to signature compiler and cleaned
%% up this file, throwing away the old ConTroll/Troll type hierarchy
%% compiler, given that Trale only uses the type hierarchy reader and the
%% various commands to interact with the signature at the end of the
%% file.
%%
%% Revision 1.16  2002/11/22 22:51:25  dm
%% *** empty log message ***
%%
%% Revision 1.15  2002/08/19 20:01:04  dm
%% Rewrote show_all_supertypes and modified show_supertypes. Both now
%% print proper hierarchies.
%%
%% Revision 1.14  2002/08/19 06:05:44  dm
%% For output of multiple inheritance in show_all_subtypes
%% added & as marker and only output subtypes of such types the first
%% time we see them.
%%
%% Revision 1.13  2002/08/15 03:20:47  dm
%% Redid the show_approp with its own pretty printer and hook. This is
%% needed since the trale feature structures can't represent only the
%% appropriateness information (since they always contain also the
%% appropriate values of the appropriate values etc.)
%% Also redid show_supertypes/1.


:- module(type_hierarchy_compiler, [process_th/1, process_th/2,
       pth_set_option/2,
       supertypes/2,
       show_approp/1,
       show_subtypes/1,show_all_subtypes/1,show_signature/1,show_signature/0,
       show_supertypes/1,show_all_supertypes/1,
       read_ths/4]). % export for TRALE

% load prolog-dependent code and other support code
% SP4: append/3 is built-in, file_exists/1 is available from site
:- site:site_import(type_hierarchy_compiler,[t_reconsult/1,unix/1,file_exists/1]).
:- use_module(library(lists),[is_list/1]).

:- dynamic prime_type/3.
:- dynamic option/2.
:- dynamic num_gen/1.

:- op(900, fy, [a_]).

% Default options for Troll:
% 1) Remove unary branching types
% 2) Don't require to mark the occurence of a multiple inheritance

:- assert(option(unary_branching, remove)).
:- assert(option(must_mark_multi, no)).

% Key word which begin a type hierarchy: type_hierarchy
% The end of a type hierarchy is found, when we read a '.' (period)
% the ascii code of the period is 46

start_kw_letter(0't).
start_kw(0't,[121,112,101,95,104,105,101,114,97,114,99,104,121]). % = ype_hierarchy


% Top level predicate: read_ths/4

read_ths(FileOrFiles,SubL,AppL,MultL) :-
   (  is_list(FileOrFiles)
   -> Files = FileOrFiles
   ;  Files = [FileOrFiles] 
   ),
   (  check_existence(Files)
   -> input_ths(Files,SubL,AppL,MultL)
   ;  SubL=[],AppL=[],MultL=[]
   ).

% check_existence/1 checks, if all the files in ListOfFiles exist
check_existence([]) :-
   !,
   format(user_error,'~N~n**Error: signature([]) specified in grammar!',[]),
   format(user_error,'~n  At least one signature file must be specified.~n    Aborting.~n~n',[]),
   abort.

check_existence([File]) :-  !,
   (  file_exists(File) 
   -> true
   ;  format(user_error,'~N~n**Warning: No signature file "~p" was found.~n',[File]),
      format(user_error,'  Assuming that signature is provided in ALE style~n',[]),
      format(user_error,'  i.e. using sub/2 and intro/2 statements as part of the grammar.~n',[]),
      fail ).
check_existence(Files) :-
   check_existence_aux(Files).

check_existence_aux([]).
check_existence_aux([File|Files]) :-
   (  nonvar(File)
   -> true
   ;  format(user_error,'~N~n**Error: Found variable instead of file name in signature/1 declaration.',[]),
      format(user_error,'~n  Only (list of) Prolog atoms are acceptable file name(s).~n  Aborting.~n~n',[]),
      abort
   ),
   (  file_exists(File) 
   -> true
   ;  format(user_error,'~N~n**Error: File "~p" specified in signature/1 declaration does not exist!~n  Aborting.~n~n',[File]),
      abort ),
   check_existence_aux(Files).


% input_ths(+FileList, -SubsumesList, -AppopriateList, -MultiInheritList)
input_ths([File|Files], SubL, AppL, MultL):-
        seen, % close any files that remained open, e.g. when user just
              % happened to invoke Cntrl-C during the previous reading 
              % in process of a signature
        initialise_number_generator,
	input_first_file(File, SubL1, AppL1, MultL1),
	input_rest_files(Files, SubL1, SubL, AppL1, AppL, MultL1,
	    MultL2),
	sort(MultL2, MultL).

% first type hierarchy of first file must be root hierarchy
% input_first_file(+File, -SubsumesList, -AppropriateList, -MultiInherL)
input_first_file(File, SubL, AppL, MultL) :- 
	user:t_see(File),
	get_next_th(Flag, ThL),
	input_initial(Flag, ThL, SubL, AppL, MultL),
	seen,
	%%close(File),
	missing_th_warning(Flag,File),
%	write('EOF reached processing file '),
	write(File),
%	write('.'),
	nl.

% input_rest_files(+Files, +SubL0, -SubL, +AppL0, -AppL, +MultL0, -MultL)
input_rest_files([File|Files], SubL0, SubL, AppL0, AppL, MultL0, MultL) :-
	user:t_see(File),
	get_next_th(Flag, ThL),
	input_normal(Flag, ThL, SubL0, SubL1, AppL0, AppL1, MultL0, MultL1),
	seen,
	%%close(File),
	missing_th_warning(Flag,File),
%	write('EOF reached processing file '),
	write(File),
%	write('.'),
	nl,
	input_rest_files(Files, SubL1, SubL, AppL1, AppL, MultL1, MultL).
input_rest_files([], SubL, SubL, AppL, AppL, MultL, MultL).

missing_th_warning(eof,File) :-
	user:writeln([' *** Warning: no type hierarchy found in file ',File]),
	user:proceed_or_abort.
missing_th_warning(normal,_).

% input_initial(+Mode, +TypeHierarchyAsAsciiList, -SubL, -AppL, -MultL)
input_initial(eof, _, [],[], []).   % on eof do nothing
input_initial(normal, ThL, SubL, AppL, MultL) :-
	(initial_th(SubL1, AppL1, MultL1, ThL, []) ->
		(get_next_th(Flag, ThL1),
		 input_normal(Flag, ThL1, SubL1, SubL, AppL1, AppL, 
		    MultL1, MultL))
	;
		error(syntaxinit)
	).

% input_normal(+Mode, +TypeHierarchyAsAsciiList, +SubL0, -SubL,
%              +AppL0, -AppL, +MultL0, -MultL) 
input_normal(eof, _, SubL, SubL, AppL, AppL, MultL, MultL).
input_normal(normal, ThL, SubL0, SubL, AppL0, AppL, MultL0, MultL) :-
	(additional_th(SubL0, SubL1, AppL0, AppL1, MultL0, MultL1,
	    ThL, []) ->
		(get_next_th(Flag, ThL1),
		 input_normal(Flag, ThL1, SubL1, SubL, AppL1, AppL,
		    MultL1, MultL))
	;
		error(syntaxaded)
	).

% get_next_th(-THList, -EOFFlag)
% reads from current position in file, finds next type hierarchy and
% retunrs it as a List of Chars.
% THList is [], if end of file reached before next type hierarchy
% could be read in. In this case, EOFFlag is 'eof', else EOFFlag is
% 'normal'.

get_next_th(Flag, ThL) :-
	get_th_start(nocomment, Char),
	(end_of_file(Char) ->
		(ThL = [], Flag = eof)
	;
		(read_th(nocomment, ThL, Flag))
	).

% read_th(+Mode, -End, -Flag)
read_th(nocomment, End, Flag) :-
	get_code(X),  % SP4: get0/1 is extinct
	read_th_nocom(X, End, Flag).

% read_th_nocom(+Char, -End, -EOFFlag)
read_th_nocom(-1, [], eof) :- !. % eof
read_th_nocom(9, _, _) :- % tab
	error(tab).
read_th_nocom(37, [10|End], Flag) :-   % comment
	!,
	skip_code(10),  % skip to new line; SP4: skip/1 is extinct
	get_code(X),
	read_th_nocom(X, End, Flag).
read_th_nocom(46, [], normal) :- !. % period ('.')
read_th_nocom(X, End, Flag) :-     % any other char
	End = [X|End1],
	get_code(Y),
	read_th_nocom(Y, End1, Flag).

% get_th_start(+Mode, -EOFFlag)
% Read file from current position till eof or beginning of a type
% hierarchy. The beginning of a type hierarchy is detected by the key
% words in start_kw

get_th_start(comment, Flag) :-
	skip_code(10),  % skip till end of line, may cause problems, if the last
	get_th_start(nocomment, Flag). % line in a file is a comment which
                                       % is not properly ended with \n

get_th_start(nocomment, Flag) :-
	get_code(X),
	get_th_start_nocomment(X, Flag).

% get_th_start_nocomment(+Char, -EOFFlag)
get_th_start_nocomment(-1, -1) :- !.   % eof encountered

get_th_start_nocomment(37, Flag) :-     % Comment encountered
	!,
	get_th_start(comment, Flag).

get_th_start_nocomment(0't, Flag) :-    % Beginning of key word enc
	!,
	start_kw(0't, RestKw),
	get_th_start_kw(RestKw, Flag).

get_th_start_nocomment(_, Flag) :-
	get_code(X),
	get_th_start_nocomment(X, Flag).


% get_th_start_kw(+RestKeyWord, -EOFFlag)
get_th_start_kw([], 97) :- !.  % we found the whole key word, so we return
	                        % 97 = 'a', just something \= eof

get_th_start_kw(RestKw, Flag) :-
	get_code(X),
	get_th_start_kw(X, RestKw, Flag).

get_th_start_kw(-1, _, -1) :- !.

get_th_start_kw(37, _, Flag) :-
	!,
	get_th_start(comment, Flag).

get_th_start_kw(X, [X|T], Flag) :-
	!,
	get_th_start_kw(T, Flag).

get_th_start_kw(_, _, Flag) :-
	get_code(X),
	get_th_start_nocomment(X, Flag).

end_of_file(-1).


/**************************************************************************/
% error messages


% error/1
% called by: clause 2 of input_initial/5
%            clause 2 of input_normal/8
% handles syntax/parsing errors

error(Type) :-
	user:close_all_streams,
	message(Type, Message),
	format(user_error,'~n**Error in parsing the signature: ~w~n~n',[Message]),
	abort.

message(syntaxinit, 'Syntax error in the initial type hierarchy !').
message(syntaxaded, 'Syntax error in one of the additional type hierarchies !').
% SP4: quoted newline must be \n
message(tab, 'There is a tab in a type hierarchy file !\nReplace all tabs with spaces, e.g., using "M-x untabify" in emacs.').

error_syntax(String, Message) :-
   user:close_all_streams,
   format(user_error,'~n**Error in parsing the signature: ~w~n',[Message]),
   write(user_error,' Trying to parse:'), 
   write_next_n(String, 40),
   nl(user_error), 
   abort.

write_next_n([], _) :- !.
write_next_n(_, 0) :- 
	!,
	write(user_error,' (...)').
write_next_n([C|Cs], N) :-
	atom_codes(W, [C]),  % SP4: atom_chars/2 is now sth different - atom_chars/2 
	write(user_error,W), % in SP3 is called atom_codes/2 in SP4
	N1 is N - 1,
	write_next_n(Cs, N1).
	

% error_type_doubles/1
% called by: clause 1 of p_to_s_graph_thc/2

error_type_doubles(Doubles) :-
	user:close_all_streams,
	format(user_error,'~n**Error: Types multiply defined!~n',[]),
	format(user_error,'  The following types are defined more than once: ~w~n~n',[Doubles]),
	abort.


% error_apps/2
% called by: clause 2 of collect_apps/4

error_apps(Type, Feature) :-
	user:close_all_streams,
	format(user_error,'~n**Error: Multiple appropriateness specification!~n',[]),
	format(user_error,'  There is more than one appropriateness specification for type ~w with feature ~w!~n~n',[Type,Feature]),
	abort.


% error_double_feat/3
% called by: clause 2 of merge_apps_bu/3

error_double_feat(Feat, Val, Val1) :-
	user:close_all_streams,
	write(user_error,'**Error: No subsumtion of values!'), nl(user_error),
	write(user_error,'Feature '), write(user_error,Feat), write(user_error,' is specified to have value '),
	write(user_error,Val), write(user_error,' on the super type and value '),
	write(user_error,Val1), write(user_error,' on the sub type.'), nl(user_error),
	write(user_error,'But '), write(user_error,Val), write(user_error,' doesn''t subsume '),
	write(user_error,Val1), write(user_error,'!'), nl(user_error),
	abort.


% error_nonempty_appsfromtype/2
% called by: clause 2 of merge_apps_td/4

error_nonempty_appsfromtype(Type, Appsfromtype) :-
	user:close_all_streams,
	write(user_error,'There has occured a real bizar error:'), nl(user_error),
	write(user_error,'Doing top-down inheritance for type '), 
	write(user_error,Type), write(user_error,' having processed all appropriateness'),
	write(user_error,'conditions'), nl(user_error), 
	write(user_error,'which are inherited downwards '),
	write(user_error,'from the sub types'), nl(user_error),
	write(user_error,'there are still the following appropriateness conditions'),
	write(user_error,'from the type remaining:'), nl(user_error),
	write(user_error,Appsfromtype), nl(user_error),
	write(user_error,'In principle this could NOT happen.'), nl(user_error),
	write(user_error,'Please report this error to the developers of Troll!!'),
	nl(user_error), nl(user_error),
	abort.


% error_maynot_multi/3
% called by: clause 1 of check_two_supertypes/4

error_maynot_multi(MaxType, Type1, Type2) :-
	user:close_all_streams,
	write(user_error,'Fatal error: Unexpected multiple inheritance!'), nl(user_error),
	write(user_error,'The maximal type '), write(user_error,MaxType),
	write(user_error,' multiply inherits from both '), nl(user_error),
	write(user_error,Type1), write(user_error,' and '), write(user_error,Type2), write(user_error,'.'), nl(user_error),
	write(user_error,'But it is not specified to be a multiply inheriting'),
	write(user_error,' type!'), 
	nl(user_error), nl(user_error),
	abort.


% error_two_supertypes/3
% called by: clause 2 of check_two_supertypes/4

error_two_supertypes(MaxTypes, Type1, Type2) :-
	user:close_all_streams,
	write(user_error,'Fatal error: No complete partial order!'), nl(user_error),
	write(user_error,'The disjunction '), write(user_error,MaxTypes), 
	write(user_error,' of maximal types has the following two super types: '),
	nl(user_error), 
	write(user_error,Type1), write(user_error,' and '), write(user_error,Type2), write(user_error,'.'), nl(user_error),
	write(user_error,'But none subsumes the other!'), nl(user_error),
	write(user_error,'But we demand that a type hierarchy be a COMPLETE '),
	write(user_error,'partial order!'), 
	nl(user_error), nl(user_error),
	abort.


% error_incompatible_feat/4
% called by: clause 2 of merge_multi_apps/4

error_incompatible_feat(Type, Feat, Val1, Val2) :-
	user:close_all_streams,
	write(user_error,'Fatal error: Incompatible values (multiple'),
	write(user_error,' inheritance)!'), nl(user_error), 
	write(user_error,'Type '), write(user_error,Type),
	write(user_error,' is specified to be multiply inheriting.'), nl(user_error),
	write(user_error,'Feature '), write(user_error,Feat),
	write(user_error,' occurs on more than one super type of '),
	write(user_error,Type), write(user_error,'.'), nl(user_error),
	write(user_error,'But the values on the super types are incompatible:'),
	nl(user_error), 
	write(user_error,'On one super type, the value is '), 
	write(user_error,Val1), write(user_error,'.'), nl(user_error),
	write(user_error,'On another the value is '), 
	write(user_error,Val2), write(user_error,'.'), nl(user_error),
	nl(user_error),
	abort.


% error_incompatible_feat/3
% called by: clause 1 of proc_one_feat/5
%            clause 1 of add_one_feat/5

error_incompatible_feat(Feat, Val1, Val2) :-
	user:close_all_streams,
	write(user_error,'Fatal error: Incompatible values (removing types)!'),
	nl(user_error), 
	write(user_error,'Feature '), write(user_error,Feat),
	write(user_error,' has a specification for its values which is'),
	write(user_error,' incompatible:'), nl(user_error),
	write(user_error,Val1), write(user_error,' and '), 
	write(user_error,Val2), write(user_error,' don''t unify!'), nl(user_error), 
	nl(user_error),
	abort.



/****************************************************************************/
% warning messages

% warning_th_change/4
% called by: clause 3 of merge_apps_td/4

warning_th_change(Type, Feat, Val, Val1) :-
	write(user_error,'Warning: Approp specification changed!'), nl(user_error),
	write(user_error,'Type '), write(user_error,Type), write(user_error,' Feature '), 
	write(user_error,Feat), nl(user_error),
	write(user_error,' Original value: '), write(user_error,Val1), nl(user_error),
	write(user_error,' New value:      '), write(user_error,Val), nl(user_error), nl(user_error).


% multi_inherit_warn/2
% called by: pth/4

multi_inherit_warn(IsMulti, ShouldMulti) :-
	option(must_mark_multi, no),
	multi_inherit_warn_3(IsMulti, ShouldMulti).

multi_inherit_warn(IsMulti, ShouldMulti) :-
	option(must_mark_multi, yes),
	multi_inherit_warn_2(IsMulti, ShouldMulti).


multi_inherit_warn_2([], []) :- !.
multi_inherit_warn_2([], ShouldMulti) :-
	write(user_error,'Warning: No multiple inheritance where marked!'), nl(user_error),
	write(user_error,'         The following types were specified to be'),
	write(user_error,'multiply inheriting,'), nl(user_error),
	write(user_error,'         but actually don''t inherit multiply: '),
	write(user_error,ShouldMulti), write(user_error,'.'), nl(user_error),
	nl(user_error).

multi_inherit_warn_2([Type-SuperTypes|IsMultiRest], ShouldMulti) :-
	user:selectchk(Type, ShouldMulti, ShouldMulti1),
	!,
	write(user_error,'% Type '), write(user_error,Type), 
	write(user_error,' multiply inherits from '), write(user_error,SuperTypes),
	write(user_error,'.'), nl(user_error),
	multi_inherit_warn_2(IsMultiRest, ShouldMulti1).

multi_inherit_warn_2([Type-SuperTypes|IsMultiRest], ShouldMulti) :-
	nl(user_error), 
	write(user_error,'Fatal error: Unexpected multiple inheritance!'), nl(user_error),
	write(user_error,'Type '), write(user_error,Type),
	write(user_error,' multiply inherits from '), write(user_error,SuperTypes),
	write(user_error,'.'), nl(user_error),
	write(user_error,'But it was not specified to be a multiply inheriting type!'), 
	nl(user_error), 
	multi_inherit_abort_2(IsMultiRest, ShouldMulti).

multi_inherit_abort_2([], _ShouldMulti) :-
	abort.
multi_inherit_abort_2([Type-_SuperTypes|IsMultiRest], ShouldMulti) :-
	user:selectchk(Type, ShouldMulti, ShouldMulti1),
	!,
	multi_inherit_abort_2(IsMultiRest, ShouldMulti1).
multi_inherit_abort_2([Type-SuperTypes|IsMultiRest], ShouldMulti) :-
	write(user_error,'Type '), write(user_error,Type),
	write(user_error,' multiply inherits from '), write(user_error,SuperTypes),
	write(user_error,'.'), nl(user_error),
	write(user_error,'But it was not specified to be a multiply inheriting type!'), 
	nl(user_error), 
	multi_inherit_abort_2(IsMultiRest, ShouldMulti).


multi_inherit_warn_3([], []) :- !.
multi_inherit_warn_3([], ShouldMulti) :-
	write(user_error,'Warning: No multiple inheritance where marked!'), nl(user_error),
	write(user_error,'         The following types were specified to be'),
	write(user_error,'multiply inheriting,'), nl(user_error),
	write(user_error,'         but actually don''t inherit multiply: '),
	write(user_error,ShouldMulti), write(user_error,'.'), nl(user_error),
	nl(user_error).

multi_inherit_warn_3([Type-SuperTypes|IsMultiRest], ShouldMulti) :-
	user:selectchk(Type, ShouldMulti, ShouldMulti1),
	!,
	write(user_error,'% Type '), write(user_error,Type), 
	write(user_error,' multiply inherits from '), write(user_error,SuperTypes),
	write(user_error,'.'), nl(user_error),
	multi_inherit_warn_3(IsMultiRest, ShouldMulti1).

multi_inherit_warn_3([Type-SuperTypes|IsMultiRest], ShouldMulti) :-
	nl(user_error), 
	write(user_error,'% Type '), write(user_error,Type), 
	write(user_error,' multiply inherits from '), write(user_error,SuperTypes),
	write(user_error,'.'), nl(user_error),
	write(user_error,'% Please mark this with & in the type hierarchy.'), nl(user_error),
	multi_inherit_warn_3(IsMultiRest, ShouldMulti).


% option settings
pth_set_option(Option, Value) :-
	user:memberchk(Option, [must_mark_multi, unary_branching]),
	!,
	set_option_2(Option, Value).

pth_set_option(Option, _Value) :-
	% user:memberchk failed
	write(user_error,Option), write(user_error,' is not a valid option!'), nl(user_error),
	write(user_error,'Options are: unary_branching, must_mark_multi'), nl(user_error),
	!, fail.

set_option_2(unary_branching, Value) :-
	user:memberchk(Value, [remove, error, ignore]),
	!,
	retractall(option(unary_branching, _)),
	assert(option(unary_branching, Value)).

set_option_2(unary_branching, Value) :-
	% user:memberchk failed
	write(user_error,Value), 
	write(user_error,' is not a legal value for option unary_branching!'),
	nl(user_error),
	write(user_error,'Values permitted: remove, error, ignore'), nl(user_error),
	!, fail.

set_option_2(must_mark_multi, Value) :-
	user:memberchk(Value, [yes, no]),
	!,
	retractall(option(must_mark_multi, _)),
	assert(option(must_mark_multi, Value)).

set_option_2(must_mark_multi, Value) :-
	% user:memberchk failed
	write(user_error,Value), 
	write(user_error,' is not a legal value for option must_mark_multi!'), 
	nl(user_error),
	write(user_error,'Values permitted: yes, no'), nl(user_error),
	!, fail.


/**************************************************************************/

% The DCG for parsing the type hierarchy


mtype(Type, multi) -->
	"&",
	!,
	type_thc(Type).
mtype(Type, normal) -->
	type_thc(Type).

type_thc(Type) -->
	word_thc(TypeName),
	!,
	primes_thc(TypeName, Type).
%-hw----
type_thc(Term) -->
	pl_term_thc(Term), !.
%-------
type_thc(_Type, String, _Rest) :-
	error_syntax(String, '''Type'' expected.').

primes_thc(TypeName0, TypeName) -->
	"'",
	!,
	aux_primes_thc(1, TypeName0, TypeName).
primes_thc(TypeName, TypeName) --> [].

aux_primes_thc(N, TypeName0, TypeName) -->
	"'",
	!,
	{N1 is N + 1},
	aux_primes_thc(N1, TypeName0, TypeName).
aux_primes_thc(N,TypeName0,TypeName) --> [],
   {string_append(TypeName0,'_',TypeName1),
    user:concat(TypeName1,N,TypeName),
    (prime_type(TypeName0,TypeName,N) ->  % only needed for process_th
        true
      ; assert(prime_type(TypeName0,TypeName,N)))}.

%-hw---

pl_term_thc((a_ Term)) -->
	"(",
	"a_",
	whitespace,
	pl_inner_term_thc(Term),
      whitespace,
	")".

pl_inner_term_thc(Term) -->
      pl_term(Term, [], _).

% pl_term
% Make a Prolog term from a list of characters.
% Free variables are kept in a list of pairs of type Name-Var, such that
% sharing between variables in terms is preserved, although this feature is
% currently ignored by the rest of the system.
%
% LIMITATIONS:

% pl_term/5 expects one single term; conjunctions or disjunctions of terms
%   are currently not supported.
% pl_term/5 cannot parse list expressions.

pl_term(Term, FreeVarsIn, FreeVarsOut) -->
      "(",
      whitespace,
      pl_term(Term, FreeVarsIn, FreeVarsOut),
      whitespace,
      ")", !.
pl_term(Term, FreeVarsIn, FreeVarsOut) -->
      atom_var_thc(FunctorName, ground),
      "(",
      whitespace,
      pl_arg_terms(ArgTerms, FreeVarsIn, FreeVarsOut),
      whitespace,
      ")", !,
      {term_univ(Term, [FunctorName | ArgTerms])}.
pl_term(Term, FreeVars, FreeVars) -->
      atom_var_thc(FunctorName, ground), !,
      {term_univ(Term, [FunctorName])}.
pl_term(Var, FreeVarsIn, FreeVarsOut) -->
      atom_var_thc(VarName, var), !,
      {lookup_var(VarName, Var, FreeVarsIn, FreeVarsOut)}.

pl_arg_terms([ArgTerm | MoreArgTerms], FreeVarsIn, FreeVarsOut) -->
      pl_term(ArgTerm, FreeVarsIn, FreeVarsIn1),
      whitespace,
      ",",
      whitespace, !,
      pl_arg_terms(MoreArgTerms, FreeVarsIn1, FreeVarsOut).
pl_arg_terms([ArgTerm], FreeVarsIn, FreeVarsOut) -->
      pl_term(ArgTerm, FreeVarsIn, FreeVarsOut).

atom_var_thc(Word, ground) -->
     	[Char],
	    {user:is_lower(Char)},
	    !,
	    chars_to_word(Chars),
	    {atom_codes(Word, [Char|Chars])}.
atom_var_thc(Word, var) -->
      [Char],
      {user:is_upper(Char); Char =:= 95},
      !,
      chars_to_word(Chars),
      {atom_codes(Word, [Char | Chars])}.

lookup_var('_', _, FreeVars, FreeVars) :- !.  % Special handling for anonymous variables.
lookup_var(VarName, Var, FreeVarsIn, FreeVarsOut) :-
      lookup_var1(VarName, Var, FreeVarsIn, FreeVarsIn, FreeVarsOut).

lookup_var1(VarName, Var, [], FreeVarsIn, [VarName-Var | FreeVarsIn]).
lookup_var1(VarName, Var, [VarName-Var | _], FreeVars, FreeVars).
lookup_var1(VarName, Var, [_ | FreeVarsInT], FreeVarsIn, FreeVarsOut) :-
      lookup_var1(VarName, Var, FreeVarsInT, FreeVarsIn, FreeVarsOut).

term_univ(Term, [Functor]) :-
      functor(Term, Functor, 0), !.
term_univ(Term, [Functor | Args]) :-
      length(Args, Arity),
      functor(Term, Functor, Arity),
      term_univ_args(Term, Args, 1, Term).

term_univ_args(Term, [], _, Term).
term_univ_args(Term, [Arg | MoreArgs], Num, Term) :-
      arg(Num, Term, Arg),
      Num1 is Num + 1,
      term_univ_args(Term, MoreArgs, Num1, Term).

%-----

word_thc(Word) -->
	[Char],
	{user:is_lower(Char)},
	!,
	chars_to_word(Chars),
	{atom_codes(Word, [Char|Chars])}.

chars_to_word([Char|Chars]) -->
	[Char], {user:is_alnum(Char); Char =:= 95},
	!,
	chars_to_word(Chars).
chars_to_word([]) --> [].

lf_thc --> "\n".  % SP4: quoted newline must be \n

twhitespace -->
	"\t",  % SP4: quoted tab must be \t
	!,
	whitespace.
twhitespace -->
	" ",
	!,
	whitespace.

whitespace -->
	"\t", % SP4: quoted tab must be \t
	!,
	whitespace.
whitespace -->
	" ",
	!,
	whitespace.
whitespace --> [].

tabs_and_spaces(NTabs,NSpaces) -->
	tabs_and_spaces(0,0, NTabs, NSpaces).

tabs_and_spaces(Tabs_sofar0, Spaces_sofar, NTabs, NSpaces) -->
	"\t", % SP4: quoted tabs must be \t
	!,
	{Tabs_sofar is Tabs_sofar0 +1},
	tabs_and_spaces(Tabs_sofar, Spaces_sofar, NTabs, NSpaces).
tabs_and_spaces(Tabs_sofar, Spaces_sofar0, NTabs, NSpaces) -->
	" ",
	!,
	{Spaces_sofar is Spaces_sofar0 +1},
	tabs_and_spaces(Tabs_sofar, Spaces_sofar, NTabs, NSpaces).
tabs_and_spaces(NTabs, NSpaces, NTabs, NSpaces) --> [].

emptylines -->
	whitespace, lf_thc,
	!,
	emptylines.
emptylines --> [].

features(Type, BirthStamp, Rest0, Rest) -->
	twhitespace,
	word_thc(F),
	":",
	type_thc(V),
	!,
	features(Type, BirthStamp, [app(Type, F, V, BirthStamp)|Rest0], Rest).
features(_Type, _BirthStamp, AppL, AppL) -->
	whitespace,
	lf_thc,
	!.
features(_Type, _BS, _Appl0, _Appl, String, _RestString) :-
	error_syntax(String, '''Feature:Value'' expected.').




%initial_th//3 initial_th(-SubsumesList, -AppropsList, -MultiInheritList)
% Note the unusual format of each element in the SubsumesList:
% 	SuperType-Type
% this format is called P-Representation by O'Keefe. The dash (-) is
% used as the functor to feed this list into sort predicates.

initial_th(SubL, AppL, MultL) -->
	emptylines,
	"bot",            % start of first type hierarchy encountered
	!,
	{next_number(BirthStamp)},
	features(bot, BirthStamp, [], AppL1),
	emptylines,
	immediate_subtypes(bot, 0, 0, [], SubL, AppL1, AppL, [], MultL).

initial_th(_SubL, _AppL, _MultL, String, _) :-
	error_syntax(String, '"bot" expected.').

immediate_subtypes(SuperType, OTabs, OSpaces, SubL0, SubL, AppL0,
   AppL, MultL0, MultL) -->
	first_subtype(SuperType, OTabs, OSpaces, SubL0, SubL1,
		AppL0, AppL1, MultL0, MultL1, NTabs, NSpaces),
	!,
	next_subtypes(SuperType, NTabs, NSpaces, SubL1, SubL,
		AppL1, AppL, MultL1, MultL).

immediate_subtypes(_SuperType, _Tabs, _Spaces, SubL, SubL, AppL, AppL,
   MultL, MultL) -->
	[].


first_subtype(SuperType, OTabs, OSpaces, SubL0, SubL, AppL0, AppL,
   MultL0, MultL, NT, NS) -->
	tabs_and_spaces(NTabs, NSpaces),
	{(NTabs > OTabs ; (NTabs =:= OTabs, NSpaces > OSpaces))},  % deeper
	mtype(Type,Mode),                                        % level
	!,                                                % of embedding
	{NT = NTabs, NS = NSpaces,
	 (Mode == multi -> MultL1 = [Type|MultL0]
	  ; MultL1 = MultL0)},
	{next_number(BirthStamp)},
	features(Type, BirthStamp, AppL0, AppL1),
	emptylines,
	immediate_subtypes(Type, NTabs, NSpaces, [SuperType-Type|SubL0],
           SubL, AppL1, AppL, MultL1, MultL).


next_subtypes(SuperType, OTabs, OSpaces, SubL0, SubL, AppL0, AppL,
   MultL0, MultL) -->
	tabs_and_spaces(OTabs, OSpaces),       % old level of indentation
	mtype(Type, Mode),
	!,
	{next_number(BirthStamp)},
	features(Type, BirthStamp, AppL0, AppL1),
	emptylines,
	{(Mode == multi -> MultL1 = [Type|MultL0]
	 ; MultL1 = MultL0)},
	immediate_subtypes(Type, OTabs, OSpaces, [SuperType-Type|SubL0],
	   SubL1, AppL1, AppL2, MultL1, MultL2),
	next_subtypes(SuperType, OTabs, OSpaces, SubL1, SubL,
		AppL2, AppL, MultL2, MultL).

next_subtypes(_SuperType, _Tabs, _Spaces, SubL, SubL, AppL, AppL,
   MultL, MultL) -->
	[].

next_subtypes(_Type, _Tabs, _Spaces, _SubL0, _SubL, _AppL0, _AppL,
   _MultL0, _MultL, String, []) :-
	error_syntax(String, 'Indentation error.').


additional_th(SubL0, SubL, AppL0, AppL, MultL0, MultL) -->
	emptylines,
	type_thc(Type),  % get new type
	{is_known(Type, SubL0)},  % check that type was already introduced
	whitespace,
	lf_thc,
	emptylines,           % no feature decl. allowed
	!,
	immediate_subtypes(Type, 0, 0, SubL0, SubL, AppL0, AppL,
	   MultL0, MultL).

% is_known(+Type,+SubsumesList)
% is_known/2 is a kind of user:memberchk/2 exept that we have a special
% data structure here which makes life more complicated: SubsumesList
% has a Variable as its tail and its elements are all terms of the
% kind SuperType-Type

is_known(_, catch) :- !, fail. % this is the "base case": we reached
                               % the tail of the list which is a Var

is_known(T, [T-_|_]) :- !.
is_known(T, [_-T|_]) :- !.
is_known(T, [_|R]) :-
	is_known(T,R).

% next_number(-Number)
% generates the natural numbers giving back on each call the successor
% of the number of the previous call.

next_number(Number) :-
	retract(num_gen(Pred)), 
	!, % retract is nondeterminate, but there can be only one entry
	Number is Pred + 1,
	assert(num_gen(Number)).


% initialise_number_generator/0
% initialises the number generator by retracting all clauses of type
% num_gen/1 and than asserts num_gen(-1)

initialise_number_generator :-
	retractall(num_gen(_)),
	assert(num_gen(-1)).

% ------------------------------------------------------------------------
% Some output predicates for the user

% show_approp/1: displays the appropriate attributes of the type provided as argument

show_approp(Type) :-
	ensure_signature_present,
	ensure_type_defined(Type),
	user:approps(Type,FVs,_),
	pp_approps(Type,FVs).


% The ALE feature structure can't handle underspecified stuff like approps,
% so we here need to define something new:
pp_approps(Type,FVs) :-
   ( (user:current_predicate(portray_approps,user:portray_approps(_,_)),
      user:portray_approps(Type,FVs))
   -> true
   ;  format('~N~n~p~n',[Type]),
      pp_approps_feat_vals(FVs)).

% Provide a proper AVM output for text mode here TODO!!!
pp_approps_feat_vals([]) :- nl.
pp_approps_feat_vals([F:V|Rest]) :-
   user:write_feature(F,_Len),
   format('~p~n',[V]),
   pp_approps_feat_vals(Rest).

	
ensure_signature_present :-
   (user:fail_if_undefined(user:sub(_,_)) 
   ->  true
   ;   format(user_error,'~N**Error: No compiled signature present.~n',[]),
       fail
   ).

ensure_type_defined(Type) :-
   (user:type(Type) 
   ->  true
   ;   format(user_error,'~N**Error: Type ~w not defined in compled signature.~n',[Type]),
       fail
   ).


% show_subtypes/1: display immediate subtypes of type provided as argument
show_subtypes(Type) :-
	ensure_signature_present,
	ensure_type_defined(Type),
	user:immed_subtypes(Type,SubTypes),
	(SubTypes == [] ->
	    write(' Type "'),
	    write(Type),
	    write('" has no subtypes.'),
	    nl
	;
	    print_subtypes([Type-SubTypes],0,false)).


% print subtypes in style of signature input format
print_subtypes([],_,_).
print_subtypes([A-B|T],Depth,ShowApprops) :- % type-subtypelist
	!,
	tab(Depth),
	print_one_subtype(A,ShowApprops),
	NewDepth is Depth + 3,
	print_subtypes(B,NewDepth,ShowApprops),
	print_subtypes(T,Depth,ShowApprops).
print_subtypes([H|T],Depth,ShowApprops) :-
	tab(Depth),
	print_one_subtype(H,ShowApprops),
	print_subtypes(T,Depth,ShowApprops).

print_one_subtype(Typ,ShowApprops) :-
   (   Typ=multi(BareTyp) 
   ->  format('&~w',[BareTyp])
   ;   write(Typ),
       (  (ShowApprops, 
	      (user:fail_if_undefined(intro(Typ,FVs))
	      ;user:fail_if_undefined(subs(Typ,intro(_SubT,FVs)))) )
       ->  print_approps(FVs)
       ;   true
       )
   ),
   nl.

print_approps([]).
print_approps([F:V|T]) :-
   format(' ~w:~w',[F,V]),
   print_approps(T).


% show_signature/1: display all subtypes of type provided as argument
%                   including appropriateness conditions
show_signature(Type) :-
   show_all_subtypes(Type,true).

% show_signature/0: show entire hierarchy
show_signature :-		
   show_signature(bot).

% show_all_subtypes/1: displays all suptypes of the type provided as argument
show_all_subtypes(Type) :-
   show_all_subtypes(Type,false).

show_all_subtypes(Type,ShowApprops) :-
	ensure_signature_present,
	ensure_type_defined(Type),
	get_all_subtypes(Type,Result,[],_),
	(Result == [] ->
	    write(' Type "'),
	    write(Type),
	    write('" has no subtypes.'),
	    nl
	;
	    print_subtypes([Type-Result],0,ShowApprops)).

% get_all_subtypes(+X,-Y)
% X properly subsumes Y
get_all_subtypes(X,Result,Seen0,Seen) :-
	user:immed_subtypes(X,FirstResult),
	get_all_subtypes_aux(FirstResult,Result,[X|Seen0],Seen).


% SP4: memberchk/2 is built-in
%:- use_module(library(lists),[memberchk/2]).

get_all_subtypes_aux([],[],Seen,Seen).
get_all_subtypes_aux([H|T],[NewH-RestH|NewT],Seen0,Seen) :-
	(   memberchk(H,Seen0) 
	->  NewH=multi(H),RestH=[],Seen1=Seen0
	;   NewH=H,get_all_subtypes(H,RestH,Seen0,Seen1)
	),
	get_all_subtypes_aux(T,NewT,Seen1,Seen).


% show_supertypes/1: displays the immediate supertype of the type provided as argument

show_supertypes(Type) :-
	ensure_signature_present,
	ensure_type_defined(Type),
	supertypes(Type,SuperTs),
	show_supertypes_aux(SuperTs,Type,0,3,false).

show_supertypes_aux([],Type,SuperTab,_,Multi) :-
   (Multi->true;tab(SuperTab),write(Type),nl). % Multi=false -> has no supertypes
show_supertypes_aux([SupType|Rest],Type,SuperTab,Tab,Multi) :-
   tab(SuperTab),write(SupType),nl,
   tab(Tab),(Multi->write('&');true),write(Type),nl,
   show_supertypes_aux(Rest,Type,SuperTab,Tab,true).

% supertypes(+X,-Y)
% Y properly subsumes X and
% there is no type Z such that Y properly subsumes Z and Z properly subsumes X
% (definition taken from show_type/1 in ale.pl)
supertypes(Type,SuperTypes) :-
   ( setof(T,T2^(user:sub_type(T,Type),
		 T \== Type,
		 \+ (user:sub_type(T2,Type),
			T2 \== Type, T2 \== T,
			user:sub_type(T,T2))),SuperTypes)
   -> true
   ; SuperTypes = []
   ).

% -------------------------

% SP4: library(assoc) is now library(avl)
:- use_module(library(avl),[avl_store/4,avl_delete/4,
                              empty_avl/1,avl_fetch/3]).

% -------------------------
% show_all_supertypes/1: displays all supertypes of the type provided as argument


show_all_supertypes(Type) :-
   ensure_signature_present,
   ensure_type_defined(Type),
   empty_avl(Empty),
   all_supertypes(Type,Empty,Hier),
   print_assoc_hierarchy([bot],Hier,0,Empty,_).

all_supertypes(T,Hier0,Hier) :-
   supertypes(T,SuperTs),
   all_supertypes_list(SuperTs,T,Hier0,Hier).

all_supertypes_list([],_,Hier,Hier).
all_supertypes_list([T|Ts],SubT,Hier0,Hier) :-
   (   avl_delete(T,Hier0,OtherTs,Hier1) % if already occurred as supertype
   ->  avl_store(T,Hier1,[SubT|OtherTs],Hier2) % then add type to its subtypes
   ;   avl_store(T,Hier0,[SubT],Hier1), % else record new supertype with type
       all_supertypes(T,Hier1,Hier2)% and explore supertypes of supertype
   ),
   all_supertypes_list(Ts,SubT,Hier2,Hier).

print_assoc_hierarchy([],_,_,Vis,Vis).
print_assoc_hierarchy([Type|Sisters],Hier,Tab0,Vis0,Vis) :-
   (avl_fetch(Type,Hier,Subtypes) -> Leaf=false; Leaf=true;Subtypes=[]),
   (   avl_fetch(Type,Vis0,_) 	% seen this type before
   ->  print_indent_type(Type,Tab0,Tab1,true,Leaf),
       Vis2=Vis0
   ;   avl_store(Type,Vis0,_,Vis1),
       print_indent_type(Type,Tab0,Tab1,false,Leaf),
       print_assoc_hierarchy(Subtypes,Hier,Tab1,Vis1,Vis2)
   ),
   print_assoc_hierarchy(Sisters,Hier,Tab0,Vis2,Vis).

print_indent_type(X,Tab0,Tab,Multi,Leaf) :-
   tab(Tab0),
   (Multi -> write('&');true),
   (Leaf -> user:capitalize_atom(X,NewX);NewX=X),
   write(NewX),nl,
   Tab is Tab0 + 3.
