% Adapted to new featural pivot/root interpretation.
% 8/20/08 - G. Penn

% Added contribution constraints (with Michael Tao)
% 7/26/11 - G. Penn

:- op(1150,fx,semtype).
:- op(1150,fx,findom).
:- op(600,fx,^).
%taomod
:- op(600,yfx,[/]). % removed /~
:- multifile if/2, macro/2.
:- multifile fun/1.
:- multifile term_expansion/2.
:- multifile abolish_user_preds/1.

:- use_module(library(terms),[term_variables/2]).
:- use_module(library(lists),[is_list/1]).
:- use_module(library(ugraphs),[vertices_edges_to_ugraph/3]).
:- use_module(library(ordsets),[ord_subtract/3]).
:- use_module(library(chr),[find_constraint/2,findall_constraints/2]).
:- ensure_loaded(subterm).
:- ensure_loaded(semrec).
:- dynamic semtype/2,findom/3.

abolish_user_preds(all) :- abolish_sem_preds.
abolish_user_preds(sig) :- abolish_sem_preds.

abolish_sem_preds :-
  abolish((semtype)/1), abolish((findom)/1),
  retractall(semtype(_,_)), retractall(findom(_,_,_)).

% semantic type compilation
term_expansion(semtype(Decl),semtype(Decl)) :-
  (Decl = Literals:Type) -> compile_semtypes_assert(Literals,Type)
  ; error_msg((nl,write('sem: ill-formed semantic type declaration: '),write(Decl),nl)).

% finite domain compilation
term_expansion(findom(Decl),findom(Decl)) :-
  (Decl = Literals:Domain) -> compile_findoms_assert(Literals,Domain)
  ; error_msg((nl,write('sem: ill-formed finite domain declaration: '),write(Decl),nl)).

compile_semtypes_assert([],_) :- !.
compile_semtypes_assert([Lit|Lits],Type) :-
  !,compile_semtypes_assert(Lit,Type),
  compile_semtypes_assert(Lits,Type).
compile_semtypes_assert(Literal,ArrowOrVar) :-
  ( ArrowOrVar = var(ArrowType) -> SemType2 = var(ATypes2)
  ; ArrowType = ArrowOrVar, SemType2 = ATypes2
  ),
  arrow_to_list(ArrowType,ATypes2),
	% this warning doesn't catch multiple declarations where both variants
	% of vars are used, e.g., var(x): entity + x: var(entity).
  ([Literal,multiply,declared,with,semantic,types,SemType1,and,SemType2]
    if_warning_else_fail clause(semtype(Literal,SemType1),true) -> true
  ; [Literal,multiply,declared,as,literal,and,finite,domain]
      if_warning_else_fail clause(findom(Literal,_,_),true) -> true
  ; assert(semtype(Literal,SemType2))
  ).

arrow_to_list(V,Args) :-
  var(V),
  !,Args = V.
arrow_to_list((A -> B),[AList|Args]) :-
  !, [illegal,use,of,'var/1',in,semantic,type,declaration,(A -> B)]
       if_error (nonvar(A), functor(A,var,1)),
  arrow_to_list(A,AList),
  arrow_to_list(B,Args).
arrow_to_list(Result,Result) :-
  [illegal,use,of,'var/1',in,semantic,type,declaration,Result]
       if_error (nonvar(Result), functor(Result,var,1)).


compile_findoms_assert([],_) :- !.
compile_findoms_assert([Lit|Lits],Domain) :-
  !,compile_findoms_assert(Lit,Domain),
  compile_findoms_assert(Lits,Domain).
compile_findoms_assert(Literal,Domain) :-
  sort(Domain,SortedDomain),
  homogeneous_domain_type(SortedDomain,Type,Literal),
  ([Literal,multiply,declared,with,finite,domains,Domain1,and,SortedDomain]
    if_warning_else_fail clause(findom(Literal,Domain1,_),true) -> true
  ; [Literal,multiply,declared,as,literal,and,finite,domain]
      if_warning_else_fail clause(semtype(Literal,_),true) -> true
  ; assert(findom(Literal,SortedDomain,Type))
  ).

homogeneous_domain_type([],_,_).
homogeneous_domain_type([Lit|Lits],Type,Literal) :-
  clause(semtype(Lit,Type),true) -> ( functor(Type,var,1) -> error_msg((nl,write('sem: '),write(Literal),
                                                                        write(' declared on a domain with variable: '),
                                                                        write(Lit),nl))
                                    ; homogeneous_domain_type(Lits,Type,Literal,Lit)
                                    )
  ; error_msg((nl,write('sem: '),write(Literal),write(' declared on a domain with undeclared literal: '),
               write(Lit),nl)).

homogeneous_domain_type([],_,_,_).
homogeneous_domain_type([Lit|Lits],Type,Literal,LastLit) :-
  clause(semtype(Lit,LType),true) -> ( Type = LType -> homogeneous_domain_type(Lits,Type,Literal,Lit)
                                     ; error_msg((nl,write('sem: '),write(Literal),
                                                  write(' declared on a domain with literals of differing types: '),
                                                  write([LastLit,Lit]),nl))
                                     )
  ; error_msg((nl,write('sem: '),write(Literal),write(' declared on a domain with undeclared literal: '),
               write(Lit),nl)).

% ALE macros			
semcontrib(Expr) macro @sem(_,Expr,contrib).
sem(Expr) macro @sem(_,Expr,nocontrib).
sem(Term,Expr,Contrib) macro sem((a_ sem(X,Term,Pivot,Root,Anchor,LRSCode))) :-
  semvars(VsIn),		
  sem_expr(Expr,Term,X,Pivot,Root,ArgTypes,ResType,Anchor,VsIn,VsOut,Contrib,Goals,[]),
  goal_list_to_seq(Goals,ExprCode),
  LRSCode = (anchor(Anchor),   
	     node(Term,ArgTypes,ResType,Anchor),
	     ExprCode),
  semvars_update(VsOut).

sem_residuate(LF,LFTerm,Pivot,Root) macro sem_residuate(LF,(a_ sem(LFTerm,_)),(a_ sem(Pivot,_)),(a_ sem(Root,_))).

% ALE functions
fun sem(-,+).
fun sem_residuate(-,+,+,+,+).

sem((sem:(X,(a_ sem(Term,Anchor))),pivot:(a_ sem(Pivot,Anchor)),root:(a_ sem(Root,Anchor))),
    (a_ sem(X,Term,Pivot,Root,Anchor,LRSCode))) if
  prolog((call(LRSCode))).

sem_close((a_ sem(Term,Anchor))) if prolog((sem_aggregate(Term,Anchor),sem_close(Anchor))).

sem_residuate(FS,LF,LFTerm,Pivot,Root) if
  SemGoal,
  (prolog(ale_flag(semclose,off)) -> true ; sem_close(LFTerm)) :-
     (semantics SemPred),
     SemGoal =.. [SemPred,FS,(LF,sem:LFTerm,pivot:Pivot,root:Root)].

%sem_residuate((a_ sem(Term,Anchor)),(a_ sem(Term,LRSResidue))) if
%  prolog((collapse_elastic_nodes(Anchor),
%	  findall_constraints(Anchor,_,LRSResidue))).

%sem_residuate(_Term,Anchor,LRSResidue) :-
%  collapse_elastic_nodes(Anchor),
%  findall_constraints(Anchor,_,LRSResidue).

collapse_elastic_nodes(Anchor) :-
  findall_constraints(Anchor,_,LRSConstraints),
  elastic_triage(LRSConstraints,Nodes,LiteralandImmSubTerms,SubTermEdges),
  sort(Nodes,NodeSet),
  sort(LiteralandImmSubTerms,LiteralandImmSubTermSet),
  ord_subtract(NodeSet,LiteralandImmSubTermSet,NonLiteralImmSubTermSet),
  vertices_edges_to_ugraph([],SubTermEdges,STGraph),
  collapse_elastic_nodes_act(STGraph,NonLiteralImmSubTermSet).

collapse_elastic_nodes_act([],_).
collapse_elastic_nodes_act([Elastic-[UniqueMother]|ElasticNodes],NonLiteralImmSubTermSet) :-
  member_eq(Elastic,NonLiteralImmSubTermSet),  % can't use ord_member: order disturbed by unifications
  Elastic = UniqueMother, % this may not succeed (type conflicts, etc.)
  !,collapse_elastic_nodes_act(ElasticNodes,NonLiteralImmSubTermSet).
collapse_elastic_nodes_act([_|ElasticNodes],NonLiteralImmSubTermSet) :-
  collapse_elastic_nodes_act(ElasticNodes,NonLiteralImmSubTermSet).

elastic_triage([],[],[],[]).
elastic_triage([C#_|LRSConstraints],Nodes,LiteralandImmedSubTerms,SubTermEdges) :-
  elastic_triage(C,LRSConstraints,Nodes,LiteralandImmedSubTerms,SubTermEdges).

elastic_triage(node(N,_,_,_),LRSConstraints,[N|Nodes],LiteralandImmedSubTerms,SubTermEdges) :-
  !,elastic_triage(LRSConstraints,Nodes,LiteralandImmedSubTerms,SubTermEdges).
elastic_triage(literal(N,_,_,_),LRSConstraints,Nodes,[N|LiteralandImmedSubTerms],SubTermEdges) :-
  !,elastic_triage(LRSConstraints,Nodes,LiteralandImmedSubTerms,SubTermEdges).
elastic_triage(ist(M,N,_,_),LRSConstraints,Nodes,[M,N|LiteralandImmedSubTerms],SubTermEdges) :-
  !,elastic_triage(LRSConstraints,Nodes,LiteralandImmedSubTerms,SubTermEdges).
elastic_triage(st(M,N,_),LRSConstraints,Nodes,LiteralandImmedSubTerms,[M-N|SubTermEdges]) :-
  !,elastic_triage(LRSConstraints,Nodes,LiteralandImmedSubTerms,SubTermEdges).
elastic_triage(_,LRSConstraints,Nodes,LiteralandImmedSubTerms,SubTermEdges) :-
  elastic_triage(LRSConstraints,Nodes,LiteralandImmedSubTerms,SubTermEdges).
			   
%  findall(N-M,find_constraint(st(M,N,_),_),SEdges,IEdges),
%  findall(N-M,find_constraint(ist(M,N,_,_),_),IEdges),
%  vertices_edges_to_ugraph([],SEdges,SGraph),
%  vertices_edges_to_ugraph([],IEdges,IGraph),
%  reachable(Term,SGraph,SNodes),
%  collapse_elastic_nodes(SNodes,SGraph,IGraph).

%collapse_elastic_nodes([],_,_).
%collapse_elastic_nodes([N|Ns],SGraph,IGraph) :-
  

%  resolve_det(Term,SGraph,IGraph,_,_).

%resolve_det(Node,SGraph,IGraph,NodeTuple,SubTuples) :-
%  node_tuple(Node,NodeTuple),
%  reachable(Node,IGraph,ImmSubTerms),
%  top_sort(IGraph,RevTopoVertices),
%  reverse(RevTopoVertices,TopoVertices),
%  subvertices(TopoVertices,ImmSubTerms,TopoImmSubTerms),
%  resolve_det_immterms(TopoImmSubTerms,SGraph,IGraph,ImmTuples),
%  neighbours(Node,SGraph,SubRootTerms),
%  resolve_det_subterms(SubRootTerms,SGraph,IGraph,SubRootTuples,SubTuples,[NodeTuple|ImmTuples]),
  
  
%subvertices([],_,[]).
%subvertices([V|Vs],Restrictor,SubVertices) :-
%  select_eq(V,Restrictor,RestrictorRest) -> SubVertices = [V|SubVerticesRest],
%                                            subvertices(Vs,RestrictorRest,SubVerticesRest)
%  ; subvertices(Vs,Restrictor,SubVertices).
  

% Expression processing
sem_expr(M,N,_,_,_,_,_,_,Vs,Vs,_,[M=N|GoalsRest],GoalsRest) :-
  var(M),
  !.
%sem_expr(Expr/~C,N,X,Pivot,Root,ArgTypes,ResType,Anchor,VsIn,VsOut,[negcontrib(N,C,Anchor)|GoalsMid],GoalsRest) :-
%  !,%is_ale_fs(C), %is_ale_fs/1 doesnt exist yet and needs to be implemented
%  sem_expr(Expr,N,X,Pivot,Root,ArgTypes,ResType,Anchor,VsIn,VsOut,GoalsMid,GoalsRest).
sem_expr(Expr/C,N,X,Pivot,Root,ArgTypes,ResType,Anchor,VsIn,VsOut,Contrib,[contrib(N,C,Anchor)|GoalsMid],GoalsRest) :-
  !,%is_ale_fs(C), %is_ale_fs/1 doesnt exist yet and needs to be implemented
  sem_expr(Expr,N,X,Pivot,Root,ArgTypes,ResType,Anchor,VsIn,VsOut,Contrib,GoalsMid,GoalsRest).
sem_expr(^Expr,Root,X,Pivot,Root,ArgTypes,ResType,Anchor,VsIn,VsOut,Contrib,Goals,GoalsRest) :-
  !,
  sem_expr(Expr,Root,X,Pivot,Root,ArgTypes,ResType,Anchor,VsIn,VsOut,Contrib,Goals,GoalsRest).
sem_expr({Expr},Pivot,X,Pivot,Root,ArgTypes,ResType,Anchor,VsIn,VsOut,Contrib,Goals,GoalsRest) :-
  !,
  sem_expr(Expr,Pivot,X,Pivot,Root,ArgTypes,ResType,Anchor,VsIn,VsOut,Contrib,Goals,GoalsRest).
sem_expr(M:Expr,N,X,Pivot,Root,ArgTypes,ResType,Anchor,VsIn,VsOut,Contrib,[M=N|Goals],GoalsRest) :-
  !,var(M),
  sem_expr(Expr,N,X,Pivot,Root,ArgTypes,ResType,Anchor,VsIn,VsOut,Contrib,Goals,GoalsRest).
sem_expr(SubExprs,N,X,Pivot,Root,_,_,Anchor,VsIn,VsOut,Contrib,Goals,GoalsRest) :-
  is_list(SubExprs),
  !,sem_sub(SubExprs,N,X,Pivot,Root,Anchor,VsIn,VsOut,Contrib,Goals,GoalsRest).
sem_expr(+(Literal),N,X,Pivot,Root,ArgTypes,ResType,Anchor,VsIn,VsOut,Contrib,[contrib(N,X,Anchor),Lit|GoalsMid],GoalsRest) :-
  !,sem_expr_lit(Literal,N,X,Pivot,Root,ArgTypes,ResType,Anchor,VsIn,VsOut,Contrib,Lit,GoalsMid,GoalsRest).
sem_expr(-(Literal),N,X,Pivot,Root,ArgTypes,ResType,Anchor,VsIn,VsOut,Contrib,[Lit|GoalsMid],GoalsRest) :-
  !,sem_expr_lit(Literal,N,X,Pivot,Root,ArgTypes,ResType,Anchor,VsIn,VsOut,Contrib,Lit,GoalsMid,GoalsRest).
sem_expr(Literal,N,X,Pivot,Root,ArgTypes,ResType,Anchor,VsIn,VsOut,Contrib,Goals,GoalsRest) :-
  ( Contrib == contrib -> Goals = [contrib(N,X,Anchor),Lit|GoalsMid]
  ; % Contrib == nocontrib
    Goals = [Lit|GoalsMid]
  ),
  sem_expr_lit(Literal,N,X,Pivot,Root,ArgTypes,ResType,Anchor,VsIn,VsOut,Contrib,Lit,GoalsMid,GoalsRest).

sem_expr_lit(Literal,N,X,Pivot,Root,ArgTypes,ResType,Anchor,VsIn,VsOut,Contrib,Lit,GoalsMid,GoalsRest) :-
  functor(Literal,Functor,Arity),
  ( clause(semtype(Functor,var(ResType)),true) -> semvar(Arity,N,Functor,Lit,Anchor,VsIn,VsMid),
                                                  ArgTypes = []
  ; clause(semtype(var(Functor),ResType),true) -> semvar(Arity,N,Functor,Lit,Anchor,VsIn,VsMid),
                                                  ArgTypes = []
  ; clause(semtype(Functor,ATypes),true) -> Lit = literal(N,Functor,Arity,Anchor),
                                            length(ArgTypes,Arity),
                                            append(ArgTypes,ResType,ATypes),
                                            VsIn = VsMid
  ; clause(findom(Functor,Domain,ATypes),true) -> Lit = fd(N,Domain,Anchor),
                                                  length(ArgTypes,Arity),
                                                  append(ArgTypes,ResType,ATypes),
                                                  VsIn = VsMid
  ; error_msg((nl,write('sem: no type declared for '),write(Functor),nl))
  ),
  sem_imsub(0,Arity,Literal,N,X,Pivot,Root,Anchor,VsMid,VsOut,Contrib,GoalsMid,GoalsRest).


semvar(0,N,Functor,literal(N,var(Var),0,Anchor),Anchor,VsIn,VsOut) :-
  !,( get_assoc(Functor,VsIn,Var) -> VsOut = VsIn
    ; put_assoc(Functor,VsIn,Var,VsOut)
    ).
semvar(Arity,_,_,_,_,_,_) :-
  error_msg((nl,write('sem: semantic variable used with arity '),write(Arity),nl)).

sem_sub([],_,_,_,_,_,Vs,Vs,_,Gs,Gs).
sem_sub([SubExpr|SubExprs],N,X,Pivot,Root,Anchor,VsIn,VsOut,Contrib,
	[node(M,ArgTypes,ResType,Anchor),st(M,N,Anchor)|GoalsMid],GoalsRest) :-
  sem_expr(SubExpr,M,X,Pivot,Root,ArgTypes,ResType,Anchor,VsIn,VsMid,Contrib,GoalsMid,GoalsMid2),
  sem_sub(SubExprs,N,X,Pivot,Root,Anchor,VsMid,VsOut,Contrib,GoalsMid2,GoalsRest).

sem_imsub(A,A,_,_,_,_,_,_,Vs,Vs,_,Gs,Gs) :- !.
sem_imsub(I,A,Literal,N,X,Pivot,Root,Anchor,VsIn,VsOut,Contrib,
	  [node(M,ArgTypes,ResType,Anchor),ist(M,N,NewI,Anchor)|GoalsMid],GoalsRest) :-
  NewI is I + 1,
  arg(NewI,Literal,ImmSubExpr),
  sem_expr(ImmSubExpr,M,X,Pivot,Root,ArgTypes,ResType,Anchor,VsIn,VsMid,Contrib,GoalsMid,GoalsMid2),
  sem_imsub(NewI,A,Literal,N,X,Pivot,Root,Anchor,VsMid,VsOut,Contrib,GoalsMid2,GoalsRest).



% convert_lrs_to_delay(LRSCode,PGoal,WGoal,NVsIn,NVsOut)
convert_lrs_to_delay((Goal1,Goal2),PGoal,WGoal,NVsIn,NVsOut) :-
  convert_lrs_to_delay(Goal1,PGoal,WGoal2,NVsIn,NVsMid),
  convert_lrs_to_delay(Goal2,WGoal2,WGoal,NVsMid,NVsOut).
convert_lrs_to_delay(node(_,_,_,_),Goal,Goal,NVs,NVs).
convert_lrs_to_delay(anchor(_),Goal,Goal,NVs,NVs).
convert_lrs_to_delay(M=N,PGoal,WGoal,NVsIn,NVsOut) :-
  ( get_assoc(M,NVsIn,MSeenFlag) -> true ; MSeenFlag = wide),
  convert_metavar_act(MSeenFlag,M,N,PGoal,WGoal,NVsIn,NVsOut).
convert_lrs_to_delay(contrib(N,C,_),PGoal,WGoal,NVsIn,NVsOut) :-
  ( get_assoc(C,NVsIn,CSeenFlag) -> true ; CSeenFlag = wide),
  convert_contrib_act(CSeenFlag,N,C,PGoal,WGoal,NVsIn,NVsOut).
convert_lrs_to_delay(st(M,N,_),ststar_req(M,N,WGoal),WGoal,NVs,NVs).
convert_lrs_to_delay(ist(M,N,A,_),(ist_req(M,N,A,R),when(nonvar(R),WGoal)),WGoal,NVs,NVs).
convert_lrs_to_delay(literal(N,F,A,_),(literal_req(N,F,A,R),when(nonvar(R),WGoal)),WGoal,NVs,NVs) :-
  atom(F).  % no delays on object variables yet - existential quant syntax ^ taken by root

convert_metavar_act(unseen,M,N,(FreshVar=N,WGoal),WGoal,NVsIn,NVsOut) :-
  put_assoc(M,NVsIn,seen(FreshVar),NVsOut).
convert_metavar_act(seen(FreshVar),_,N,when_eq(FreshVar,N,WGoal),WGoal,NVs,NVs).
convert_metavar_act(wide,M,N,when_eq(M,N,WGoal),WGoal,NVs,NVs).

convert_contrib_act(unseen,N,C,contrib_req1(N,FreshVar,WGoal),WGoal,NVsIn,NVsOut) :-
  put_assoc(C,NVsIn,seen(FreshVar),NVsOut).
convert_contrib_act(seen(FreshVar),N,_,(contrib_req2(N,FreshVar,R),when(nonvar(R),WGoal)),WGoal,NVs,NVs).
convert_contrib_act(wide,N,C,(contrib_req2(N,C,R),when(nonvar(R),WGoal)),WGoal,NVs,NVs).



% --------------------------------------------------------------------------------

term_constraints(Term,Pattern,List) :-
  term_variables(Term,Vars),
  term_constraints_act(Vars,Pattern,List).

term_constraints_act([],_,[]).
term_constraints_act([V|Vars],Pattern,List) :-
  findall_constraints(V,Pattern,Prefix),
  append(Prefix,Rest,List),
  term_constraints_act(Vars,Pattern,Rest).
	
