%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Exhaustive subtype covering constraint compilation
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  Could optimise compiler by:
%  - using clause/2 or module-level variables for access to dynamic predicates
%  - precompiling code for attested branching factors that would eliminate
%    recursion through maps (represented as lists)
%  - tabulating minimal disjoint maps of deranged types for reuse while
%    checking supertypes
%  Could optimise run-time constraint-checking by:
%  - using generalise_maps/2 (see below)
%  - using a better data structure for FS's

% Bug corrected: compile_deranged_assert/0 was not collecting default minimal types.
% 8/20/08 - G. Penn

% Bug corrected: project_val/6 was binding feature map tags to FSs instead of interrupts
%  in some cases.
% 8/12/09 - G. Penn

:- multifile if_b/2.
:- discontiguous if_b/2.

:- multifile alec_catch_hook/2.
:- discontiguous alec_catch_hook/2.
alec_catch_hook(stc_deranged,Code) :-
  multi_hash(0,(stc_maps_init/2),Code,CodeMid),
  multi_hash(0,(project_map/5),CodeMid,[end_of_file]).

stc_maps_init(Type,FS) if_b SubGoals :-
  clause(deranged_maps(Type,Maps),true),
  tag_maps(Maps,TaggedMaps,SubGoals,[stc_maps_init_act(Type,FS,TaggedMaps)]).
  
tag_maps([],[],Code,Code).
tag_maps([Map|Maps],[TaggedMap|TaggedMaps],Code,CodeRest) :-
  tag_map(Map,TaggedMap,Code,CodeMid),
  tag_maps(Maps,TaggedMaps,CodeMid,CodeRest).

tag_map(Map,TaggedMap,Code,CodeRest) :-
  add_tags(Map,TaggedMap),
  inequate_tags(TaggedMap,[],Code,CodeRest).

add_tags([],[]).
add_tags([T|TRest],[T-_Tag|TaggedTRest]) :-
  add_tags(TRest,TaggedTRest).

inequate_tags([],_,Code,Code).
inequate_tags([T-Tag|ToTagRest],TMapAccum,Code,CodeRest) :-
  inequate_tags_act(TMapAccum,T,Tag,ToTagRest,TMapAccum,Code,CodeRest).

inequate_tags_act([],T,Tag,ToTagRest,TMapAccum,Code,CodeRest) :-
  inequate_tags(ToTagRest,[T-Tag|TMapAccum],Code,CodeRest).
inequate_tags_act([T2-Tag2|TMapRest],T1,Tag1,ToTagRest,TMapAccum,Code,
                  CodeRest) :-
  (\+ \+ unify_type(T1,T2,_))
  -> inequate_tags_act(TMapRest,T1,Tag1,ToTagRest,TMapAccum,Code,CodeRest)
   ; Code = [dif(Tag1,Tag2)|CodeMid],
     inequate_tags_act(TMapRest,T1,Tag1,ToTagRest,TMapAccum,
                       CodeMid,CodeRest).
% could be more selective here - if appropriateness conditions at the current
% type exclude equality, then no dif's are necessary.

% compile_deranged/0
% Add constraints to enforce exhaustive subtyping covering, e.g.:
%
%  bot
%   a f:bool g:bool
%     b f:plus g:minus
%     c f:minus g:plus
%   bool
%    plus
%    minus
%
% In TRALE, there are no objects of type a with f:minus g:minus or 
% f:plus g:plus, because all a-objects are of either type b or c.
% Types, such as a, whose feature values are not exhaustively covered by
% by their subtypes' feature values, are called *deranged* types.

compile_deranged :-
  touch('.alec_throw'),
  absolute_file_name('.alec_throw',AbsFileName),
  retractall(ale_compiling(_)),
  assert(ale_compiling(AbsFileName)),
  compile_deranged_act,
  retract(ale_compiling(_)).

compile_deranged_act :-
  alec_announce('Compiling subtype covering constraints...'),
  retractall(deranged(_,_)),
  retractall(deranged_maps(_,_)),
  retractall(maxcount(_,_,_)),
  abolish((stc_maps_init)/2), abolish((project_map)/5),
  compile_deranged_assert,
  assert(alec(stc_deranged)),
  compile('.alec_throw').

compile_deranged_assert :-
  findall(Super-Sub,(non_a_type(Super),
		     imm_sub_type(Super,Sub)),SubGEdges,SubGEdgesMid),
  findall(bot-Min,(current_predicate(intro,(_ intro _)),
		   (Min intro _),
		   \+ clause(sub_rhstype(Min),true)),SubGEdgesMid,SubGEdgesRest),
  findall(bot-Min,(clause(ext_or_intro_rhstype(Min),true),
		   \+ clause(sub_rhstype(Min),true)),SubGEdgesRest),
  vertices_edges_to_ugraph([],SubGEdges,SubG),
%  base_graph(SubG,BaseSubG),
  top_sort(SubG,RevSortedTypes),
  reverse(RevSortedTypes,SortedTypes),
  compute_maxcount(SortedTypes,SubG),
  compute_deranged(SortedTypes,SubG).

% compute_maxcount:  maxcount/2 counts the number of maximal types subsumed
%  by every type.  maxcount/2 of every maximal type is 1.  maxcount/2 of
%  a non-maximal type is the sum of the maxcounts of the types in its 
%  immediate subtype cover, as given by the base subsumption graph.

:- dynamic maxcount/3.

compute_maxcount([],_).
compute_maxcount([Type|Types],SubG) :-
  (Type = (a_ _) -> assert(maxcount(Type,1,[(a_ [])]))  % treat a_/1 atoms specially
  ; neighbours(Type,SubG,ImmedSubCover),
    ( ImmedSubCover == [] -> assert(maxcount(Type,1,[Type]))
    ; maxcount_cover(ImmedSubCover,[],MaxCover),
      length(MaxCover,MaxCount),
      assert(maxcount(Type,MaxCount,MaxCover))
    )
  ),
  compute_maxcount(Types,SubG).

% maxcount_map/2:  A map is a product of types, such as that found in a set of 
%  appropriate features.  maxcount_map/2 counts the number of maximal maps
%  subsumed by every map.  A maximal map is one that has a maximal type in
%  every dimension (feature).  maxcount_map/2 of a map is the product of the 
%  maxcount/2 of every dimension.

maxcount_map([],Count,Count).
maxcount_map([T|MapRest],Accum,Count) :-
  clause(maxcount(T,TCount,_),true),
  NewAccum is Accum * TCount,
  maxcount_map(MapRest,NewAccum,Count).

maxcount_map_list([],Total,Total).
maxcount_map_list([Map|Maps],Accum,Total) :-
  maxcount_map(Map,1,Count),
  NewAccum is Accum + Count,
  maxcount_map_list(Maps,NewAccum,Total).

maxcount_cover([],Accum,Cover) :-
  ord_union(Accum,Cover).  % ord_union/2 uses divide-and-conquer
maxcount_cover([Type|Types],Accum,Cover) :-
  clause(maxcount(Type,_,TCover),true),
  maxcount_cover(Types,[TCover|Accum],Cover).

% compute_deranged: deranged/2 indicates the status of every type for the 
%  purpose of classifying deranged types.  A deranged type is one whose
%  map of appropriate value restrictions is not exhaustively covered by the
%  maps of appropriate value restrictions of its maximally specific subtypes.
%  The status is one of:
%
%  normal:   non-deranged types - these can be treated as maximal when
%             checking a potentially deranged supertype, even if they aren't.
%  deranged(NormalCover,NCRest):  deranged - when checking potentially deranged
%              supertypes, these types must be replaced by a minimal covering 
%              of non-deranged subtypes given by NormalCover-NCRest.
%
%  Since atoms are downward-closed, we don't need to add an entry for them.
%  All maximal types are normal.

:- dynamic deranged/2.
:- dynamic deranged_maps/2.

compute_deranged([],_).
compute_deranged([Type|Types],SubG) :-  
  approps(Type,FRs,_),
  ( FRs == [] -> compute_deranged(Types,SubG) % Type has no approp 
                   % features so dont add deranged/2 entry - includes a_ atoms
  ; maximal(Type) -> assert(deranged(Type,normal)),
                     compute_deranged(Types,SubG)
  ; build_root_map(FRs,Root,MapFeats),
    neighbours(Type,SubG,Cover),
    normalise_cover(Cover,NormalCover),
    build_cover_maps(NormalCover,MapFeats,NMaps),
    minimise_disjoin_maps(NMaps,[],SubG,MDNMaps),
    atomic_equal(MDNMaps,Root,AEMaps),
    ( AEMaps == [] -> % no cover maps cover a_ atomic values - deranged
                      append(NormalCover,NCRest,NCAssert),
                      assert(deranged(Type,deranged(NCAssert,NCRest))),
                      generalise_maps(MDNMaps,GenMaps),
                      assert(deranged_maps(Type,GenMaps))
    ; length(AEMaps,N), length(APMaps,N),  % fix length of APMaps first for atomic_prune_close/1
      atomic_prune(Root,AEMaps,APRoot,APMaps),
      ( APRoot == [] ->  % no other values to cover - normal
                         assert(deranged(Type,normal))
      ; maxcount_map(APRoot,1,RootCount),    % use counts instead of types - this
        maxcount_map_list(APMaps,0,APCount), % is what makes this algorithm
                                             % polynomial in number of types
        (APCount < RootCount                  % Cover maps dont cover root map
        -> append(NormalCover,NCRest,NCAssert),  % so root type is deranged
           assert(deranged(Type,deranged(NCAssert,NCRest))),
           generalise_maps(MDNMaps,GenMaps),
           assert(deranged_maps(Type,GenMaps))
         ; % APCount = RootCount
           assert(deranged(Type,normal))
        )
      )
    ),
    compute_deranged(Types,SubG)
  ).

% atomic_prune/4: having identified maps that cover the a_/1 atoms of the root
%  map (AMaps), prune off the dimensions that have a_/1 atoms from both the
%  AMaps and the root.  The standard counting algorithm applies to what
%  remains (APMaps and APRoot).
atomic_prune([],_,[],APMaps) :-
  atomic_prune_close(APMaps).
atomic_prune([RType|RTypes],AMaps,RPTypes,APMaps) :-
  ( RType = (a_ _) -> atomic_prune_delete(AMaps,AMapsRest),
                      atomic_prune(RTypes,AMapsRest,RPTypes,APMaps)
  ; RPTypes = [RType|RPTypesRest],
    atomic_prune_copy(AMaps,APMaps,AMapsRest,APMapsRest),
    atomic_prune(RTypes,AMapsRest,RPTypesRest,APMapsRest)).

atomic_prune_delete([],[]).
atomic_prune_delete([[_Type|Types]|MapsRest],[Types|NewMapsRest]) :-
  atomic_prune_delete(MapsRest,NewMapsRest).

atomic_prune_close([]).
atomic_prune_close([[]|PMapsRest]) :-
  atomic_prune_close(PMapsRest).

atomic_prune_copy([],[],[],[]).
atomic_prune_copy([[Type|Types]|MapsRest],[[Type|PTypes]|PMapsRest],
                  [Types|NewMapsRest],[PTypes|NewPMapsRest]) :-
  atomic_prune_copy(MapsRest,PMapsRest,NewMapsRest,NewPMapsRest).

% atomic_equal/3: find those maps which, in every dimension for which the root
%  map has bot or an a_/1 atom, have types that exhaustively cover the root
%  map's restrictions.  a_/1 atoms must be treated specially because their
%  arguments are Prolog terms, and in the lattice of generalised atomic
%  formulae, every non-ground term has an infinite number of maximal subtypes
%  in its cover.  AEMaps are those maps that manage to encompass these
%  infinite covers (with infinite covers of their own).  In contrast to the
%  rest of the type hierarchy, no collection of finite covers will suffice.
%
%  If the root map does not subsume a_/1 atoms in any dimension, then all 
%  Maps are trivially included in AEMaps.
atomic_equal([],_,[]).
atomic_equal([Map|Maps],Root,AEMaps) :-
  atomic_equal_map(Root,Map,Map,Maps,Root,AEMaps).

atomic_equal_map([],[],Map,Maps,Root,[Map|AEMapsRest]) :-
  atomic_equal(Maps,Root,AEMapsRest).
atomic_equal_map([RType|RTypes],[MType|MTypes],Map,Maps,Root,AEMaps) :-
  ( RType = (a_ RX) -> MType = (a_ MX), variant(MX,RX)
  ; RType = bot -> ( MType = (a_ MX) -> var(MX)
                   ; true) % other types are allowed - it could be that
                           % a_/1 atoms do not occur in any maximal extension
                           % and therefore should never occur.
  ; true)
  -> atomic_equal_map(RTypes,MTypes,Map,Maps,Root,AEMaps)
   ; atomic_equal(Maps,Root,AEMaps).


% normalise_cover:  replace all deranged types in Cover with their non-deranged
%  subtype covers

normalise_cover([],[]).
normalise_cover([Type|Types],NormalCover) :-
  clause(deranged(Type,Status),true),
  (Status == normal
  -> NormalCover = [Type|NCRest],
     normalise_cover(Types,NCRest)
   ; % Status = deranged(NormalCover,NCRest),
     arg(1,Status,NormalCover),
     arg(2,Status,NCRest),
     normalise_cover(Types,NCRest)).

% minimise_disjoin_cover/4: for each type in cover, first eliminate all types
%  that it subsumes (or eliminate it if it is subsumed), then ensure that it
%  is disjoint with every other type (doesn't unify).  If it is not, create
%  a new minimal cover of it that isolates the intersection type, eliminates
%  it, and proceed with the remaining cover types instead.

minimise_disjoin_cover([],MDCover,_,MDCover).
minimise_disjoin_cover([C|Covers],Accum,SubG,MDCover) :-
  minimise_disjoin_cover_act(Accum,C,NewAccum,NewAccum,Covers,SubG,MDCover).

minimise_disjoin_cover_act([],C,NewAccum,NewAccumRest,Covers,SubG,MDCover) :-
  ( C = (a_ _) -> NewAccumRest = [C],  % cant dissolve this
                  minimise_disjoin_cover(Covers,NewAccum,SubG,MDCover)
  ; NewAccumRest = [],
    disjoin_cover_act(NewAccum,C,NewerAccum,NewerAccum,Covers,SubG,MDCover)
  ).
minimise_disjoin_cover_act([AccC|AccumRest],C,NewAccum,NewAccumRest,Covers,
                           SubG,MDCover) :-
  (sub_type(AccC,C)
  -> NewAccumRest = [AccC|AccumRest],
     minimise_disjoin_cover(Covers,NewAccum,SubG,MDCover)  % throw out C
   ; (sub_type(C,AccC)
     -> minimise_disjoin_cover_act(AccumRest,C,NewAccum,NewAccumRest,Covers,
                                   SubG,MDCover)         % throw out AccC
      ; NewAccumRest = [AccC|NewAccumRest2],
        minimise_disjoin_cover_act(AccumRest,C,NewAccum,NewAccumRest2,Covers,
                                   SubG,MDCover))).

disjoin_cover_act([],C,NewAccum,[C],Covers,SubG,MDCover) :-
  minimise_disjoin_cover(Covers,NewAccum,SubG,MDCover).
disjoin_cover_act([AccC|AccumRest],C,NewAccum,NewAccumRest,Covers,SubG,
                  MDCover) :-
  unify_type(AccC,C,UnifC)
  -> dissolve_map_dim(C,UnifC,SubG,NewCovers,Covers),
     NewAccumRest = [AccC|AccumRest],
     minimise_disjoin_cover(NewCovers,NewAccum,SubG,MDCover)
   ; NewAccumRest = [AccC|NewAccumRest2],
     disjoin_cover_act(AccumRest,C,NewAccum,NewAccumRest2,Covers,SubG,
                       MDCover).

% minimise_disjoin_maps:  for each map, first eliminate all maps that it
%  subsumes (or eliminate it if it is subsumed), then ensure that it is
%  disjoint with every other map (doesn't unify coordinate-wise).  If it
%  is not, create a new minimal cover of it that isolates the intersection,
%  eliminate the intersection, and proceed with the remaining maps instead.
%  This is a generalisation of minimise_disjoin_cover/4 to products of types.

minimise_disjoin_maps([],MDMaps,_,MDMaps).
minimise_disjoin_maps([Map|Maps],Accum,SubG,MDMaps) :-
  minimise_disjoin_maps_act(Accum,Map,NewAccum,NewAccum,Maps,SubG,MDMaps).

minimise_disjoin_maps_act([],Map,NewAccum,NewAccumRest,Maps,SubG,MDMaps) :-
  ( a_atoms(Map) -> NewAccumRest = [Map],  % cant dissolve these
                    minimise_disjoin_maps(Maps,NewAccum,SubG,MDMaps)
  ; NewAccumRest = [],
    disjoin_maps_act(NewAccum,Map,NewerAccum,NewerAccum,Maps,SubG,MDMaps)
  ).
minimise_disjoin_maps_act([AccMap|AccumRest],Map,NewAccum,NewAccumRest,Maps,
                          SubG,MDMaps) :-
  (subsume_map(AccMap,Map)
  -> NewAccumRest = [AccMap|AccumRest],
     minimise_disjoin_maps(Maps,NewAccum,SubG,MDMaps)  % throw out Map
   ; (subsume_map(Map,AccMap)
     -> minimise_disjoin_maps_act(AccumRest,Map,NewAccum,NewAccumRest,Maps,
                                  SubG,MDMaps)         % throw out AccMap
      ; NewAccumRest = [AccMap|NewAccumRest2],
        minimise_disjoin_maps_act(AccumRest,Map,NewAccum,NewAccumRest2,Maps,
                                  SubG,MDMaps))).

disjoin_maps_act([],Map,NewAccum,[Map],Maps,SubG,MDMaps) :-
  minimise_disjoin_maps(Maps,NewAccum,SubG,MDMaps).
disjoin_maps_act([AccMap|AccumRest],Map,NewAccum,NewAccumRest,Maps,SubG,
                 MDMaps) :-
  unify_map(AccMap,Map,UnifMap)
  -> dissolve_map(Map,UnifMap,SubG,NewMaps,Maps),
     NewAccumRest = [AccMap|AccumRest],
     minimise_disjoin_maps(NewMaps,NewAccum,SubG,MDMaps)
   ; NewAccumRest = [AccMap|NewAccumRest2],
     disjoin_maps_act(AccumRest,Map,NewAccum,NewAccumRest2,Maps,SubG,
                      MDMaps).

dissolve_map(Ts,Ss,SubG,NewMaps,NewMapsRest) :-
  dissolve_map_act(Ts,Ss,SubG,NewMapsbyDim),
  permute_dimensions(NewMapsbyDim,NewMapsbyType,NewMapsRest),
  select(Ss,NewMapsbyType,NewMaps).  % a_ atoms will unify their args here

dissolve_map_act([],[],_,[]).
dissolve_map_act([T|Ts],[S|Ss],SubG,[OtherSs|OtherDims]) :-
  dissolve_map_dim(T,S,SubG,OtherSsBag,[]),
  remove_duplicates(OtherSsBag,OtherSs),
  dissolve_map_act(Ts,Ss,SubG,OtherDims).

dissolve_map_dim(T,S,SubG,OtherSs,OtherSsRest) :-
  (variant(T,S)                % variant/2 for equality check with a_ atoms
  -> OtherSs = [T|OtherSsRest]
   ; (sub_type(T,S)
     -> ( T = (a_ _) -> OtherSs = [T|OtherSsRest]
        ; neighbours(T,SubG,Neibs),
          dissolve_map_neibs(Neibs,S,SubG,OtherSs,OtherSsRest)
        )
      ; OtherSs = [T|OtherSsRest])).

dissolve_map_neibs([],_,_,OtherSs,OtherSs).
dissolve_map_neibs([N|Neibs],S,SubG,OtherSs,OtherSsRest) :-
  dissolve_map_dim(N,S,SubG,OtherSs,OtherSsMid),
  dissolve_map_neibs(Neibs,S,SubG,OtherSsMid,OtherSsRest).

% permute_dimensions/3: Convert [[a1,a2,...,a_n1],[b1,b2,...,b_n2],...] to
%  [[a1,b1,...],[a1,b2,...],[a2,b1,...],[a2,b2,...],...[a_n1,b_n2,...]].
% This is exponential in the number of features per type, i.e., max(i)(ni).

permute_dimensions(NewMapsbyDim,NewMapsbyType,NewMapsRest) :-
  findall(Map,permute_dimensions_act(NewMapsbyDim,Map),
          NewMapsbyType,NewMapsRest).

permute_dimensions_act([],[]).
permute_dimensions_act([Dim|DimsRest],[T|MapRest]) :-
  member(T,Dim),  % this leaves a choice-point that findall/4 backtracks to
  permute_dimensions_act(DimsRest,MapRest).

% generalise_maps/2:  Given a list of mutually disjoint maps, add all of the
%  dimension-relative generalisations of any sublist, then remove all maps 
%  subsumed by another map (in all dimensions) on the list.  A dimension-
%  relative (or k-relative) generalisation of a set of maps, A, is a map in 
%  which, for one dimension, k, it has a type that is exhaustively covered by 
%  the k-dimensions of A and, for all other dimensions, i, has a type that is
%  the intersection (unification) of the i-dimensions of A.
%    It can be shown that if a map is exhaustively covered by a finite set of 
%  maps A, then it can be derived from A by a finite sequence of dimension-
%  relative generalisations.  Because these derivation sequences satisfy the
%  Church-Rosser property, each dimension in turn can be closed under its 
%  relative generalisations once.
%    Closing under exhaustive generalisation can allow us to kill coroutined
%  constraints on feature structures earlier by identifying more general maps 
%  that satisfy deranged type constraints.

generalise_maps(Maps,Maps).  % skip for now

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Run-time exhaustive type constraint enforcement
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

stc_maps_init_act(CType,AleFS,TMaps) :-
  deref(AleFS,TFS,Type,_),
  (Type = CType -> % can't be a_/1 atom - no features
                   % can't be maximal - not deranged
                   % must belong to module - otherwise no features
                   arg(4,TFS,Int),
                   project_map(Type,AleFS,VsMap,Condition,Int),
                   stc_check_maps(TMaps,VsMap,[],Condition,CType,AleFS)
  ; true % if the new type is deranged, then its own exhaustive
         % type constraints will take care of it.  If it is 
         % normal, then its appropriate value restrictions will
         % take care of it.
  ).

stc_maps(CType,AleFS,TMaps,VsMap) :-
  deref(AleFS,TFS,Type,_),
  ( Type = CType % can't be a_/1 atom - no features
                 % can't be maximal - not deranged
                 % must belong to module - otherwise no features
  -> arg(4,TFS,Int),
     update_map(VsMap,NewVsMap,Condition,nonvar(Int),Condition,CType,AleFS,
		TMaps,NewVsMap)
  ; true % if the new type is deranged, then its own exhaustive
  ).     % type constraints will take care of it.  If it is 
         % normal, then its appropriate value restrictions will
         % take care of it.

stc_check_maps([],VsMap,TMaps,Condition,CType,AleFS) :-
  TMaps \== [],     % more to try or else VsMap cannot have a maximal extension
  when(Condition,stc_maps(CType,AleFS,TMaps,VsMap)).
stc_check_maps([TMap|TMapsRest],VsMap,TMapAccum,Condition,CType,AleFS) :-
  (subsume_tmap(TMap,VsMap)   % VsMap is guaranteed to have a maximal extension
  -> true                     
   ; \+ \+ unify_tmap_check(TMap,VsMap) % VsMap has one now, but might lose it 
                                        %  later
     -> stc_check_maps(TMapsRest,VsMap,[TMap|TMapAccum],Condition,CType,
                         AleFS)
      ; stc_check_maps(TMapsRest,VsMap,TMapAccum,Condition,CType,AleFS)). 
                                        % throw this Map out and try the others

% project_map/5: create a feature value map from Vs, along with a conditional for
%  delaying until a change is observed.
project_map(Type,FS,VsMap,Condition,Int) if_b SubGoals :-
  clause(deranged(Type,deranged(_,_)),true),
  clause(tmodule(Type,Module),true), clause(marity(Module,Arity),true),
  functor(FS,Module,Arity),
  approps(Type,FRs,_),
  project_map_act(FRs,FS,VsMap,Condition,nonvar(Int),SubGoals).
  
project_map_act([],_,[],CondRest,CondRest,[]).
project_map_act([F:R|FRs],FS,[VType-VInt|VsMap],Condition,CondRest,SubGoals) :-
  clause(fcolour(F,N,_),true),
  (maximal(R) -> VType = R, Condition = CondMid, SubGoals = SubGoalsRest % maximal type cannot change
  % KNOWN BUG: actually, these shouldn't even be in the maps; there probably is a better way to
  %  partially evaluate these at compile-time to keep map size minimal.  In general, many features
  %  are known at compile-time never to be the cause of subtype cover violations.  These include,
  %  but are not limited to, all features with maximal appropriate value restrictions.
  ; SubGoals = [project_val(V,R,VType,VInt,Condition,CondMid)|SubGoalsRest]
  ),
  arg(N,FS,V),
  project_map_act(FRs,FS,VsMap,CondMid,CondRest,SubGoalsRest).

project_val(V,R,VType,VInt,Condition,CondMid) :- 
  deref(V,VTFS,VType0,_),
  ( maximal(VType0) -> Condition = CondMid, VType = VType0, VInt = VType
  ; VType0 == 0 -> VType = R, Condition = (nonvar(VCond);CondMid),
                   (R == bot -> '$put_attributes'(V,tfs(bot,VInt)), VCond = VInt
		   ; clause(tmodule(R,RM),true) -> VCond = V, clause(marity(RM,RMArity),true),
                                                   when(nonvar(VCond),arg(RMArity,VCond,VInt))
		   ; bind_mgsat(R,V,VInt), VCond = VInt
		   )
  ; VType0 = (a_ X) -> VType = VType0, delay_for_term(X,Condition,CondMid)
  ; VType = VType0, functor(VTFS,_tfs,Arity), arg(Arity,VTFS,VInt),
    Condition = (nonvar(VInt);CondMid)
  ).

delay_for_term(X,Cond,CondRest) :-
  term_variables(X,XVars),
  delay_for_term_act(XVars,Cond,CondRest).
delay_for_term_act([],Cond,Cond).
delay_for_term_act([V|Vs],(nonvar(V);CondMid),CondRest) :-
  delay_for_term_act(Vs,CondMid,CondRest).

% update_map/9: update the feature value map VsMap after one or more referencing and
%  full dereferencing events.  Also create a new conditional.  If there were only
%  referencing events, check the maps.
update_map([],[],CondRest,CondRest,Condition,CType,AleFS,TMaps,NewVsMap) :-
  stc_check_maps(TMaps,NewVsMap,[],Condition,CType,AleFS).
update_map([VType-VTag|VsMap],[NewVType-NewVTag|NewVsMapRest],CondMid,
	   CondRest,Condition,CType,AleFS,TMaps,NewVsMap) :-
  ( maximal(VType) -> CondMid = CondMid2, NewVTag = VTag, NewVType = VType
  ; var(VTag) -> NewVTag = VTag, CondMid = (nonvar(NewVTag);CondMid2), NewVType = VType
  ; functor(VTag,NewVType,_), 
    ( maximal(NewVType) -> NewVTag = NewVType, CondMid = CondMid2
    ; arg(1,VTag,NewVTag), CondMid = (nonvar(NewVTag);CondMid2))
  ),
  update_map(VsMap,NewVsMapRest,CondMid2,CondRest,Condition,CType,AleFS,TMaps,NewVsMap).

%% fully_deref_map/4: update a feature value map for a FS that has just undergone full
%%  dereferencing, create a new conditional, and resuspend.  This is identical to
%%  project_map/4 except that it delays rather than dereferences.
%fully_deref_map([_],[],CondRest,CondRest) :- !.  % extra position is for intensionality variable.
%fully_deref_map([V|Vs],[VType-VTag|VsMap],(nonvar(VTag);CondMid),CondRest) :-
%  when(nonvar(V),(arg(1,V,VTag), arg(2,V,VSVs),
%		  when(nonvar(VSVs),(( functor(VSVs,'a_',1) -> VType = VSVs
%				     ; functor(VSVs,VType,_)
%				     ))))),
%  fully_deref_map(Vs,VsMap,CondMid,CondRest).

%% trale_deref/4: dereference a feature structure, but stop at tags 
%%  instantiated to fully/2 (used by fully_deref) or numbers (used by pp_fs).
%trale_deref(Tag-SVs,TagOut,SVsOut) :-
%  var(Tag) -> TagOut = Tag,
%              SVsOut = SVs
%; functor(Tag,-,2) -> trale_deref(Tag,TagOut,SVsOut)
%; TagOut = Tag,  % stop at number or fully/2.
%  SVsOut = SVs. 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Utilities
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

a_atoms([]).
a_atoms([(a_ _)|AsRest]) :-
  a_atoms(AsRest).

subsume_map([],[]).
subsume_map([T|Ts],[S|Ss]) :-
  sub_type(T,S),
  subsume_map(Ts,Ss).

unify_map([],[],[]).
unify_map([T1|T1s],[T2|T2s],[T3|T3s]) :-
  unify_type(T1,T2,T3),
  unify_map(T1s,T2s,T3s).

subsume_tmap([],[]).
subsume_tmap([T-_|Ts],[S-_|Ss]) :-  % dont need to check tags - if Ss is
  sub_type(T,S),                    %  subsumed, then structure-sharing must
  subsume_tmap(Ts,Ss).              %  be allowed

% must call this predicate inside \+ \+ in order not to bind a_ atoms or
%  tags permanently
unify_tmap_check([],[]).
unify_tmap_check([T1-Tag|T1s],[T2-Tag|T2s]) :-
  unify_type(T1,T2,_),
  unify_tmap_check(T1s,T2s).

build_root_map([],[],[]).
build_root_map([F:R|FRs],[R|Root],[F|MapFeats]) :-
  build_root_map(FRs,Root,MapFeats).

build_cover_maps([],_,[]).
build_cover_maps([Type|Types],MapFeats,[Map|Maps]) :-
  build_cover_map(MapFeats,Type,Map),
  build_cover_maps(Types,MapFeats,Maps).

build_cover_map([],_,[]).
build_cover_map([F|Feats],Type,[R|Map]) :-
  approp(F,Type,R),
  build_cover_map(Feats,Type,Map).

map_max([],SsMax,SsMax).
map_max([S|Ss],SsTemp,SsMax) :-
  insert_if_maximal(SsTemp,S,SsTempNew),
  map_max(Ss,SsTempNew,SsMax).

insert_if_maximal([],S,[S]).
insert_if_maximal([S2|Ss2],S,[S2|Ss2]):-
  sub_type(S,S2), !.
insert_if_maximal([S2|Ss2],S,Ss3):-
  sub_type(S2,S), !,
  insert_if_maximal(Ss2,S,Ss3).
insert_if_maximal([S2|Ss2],S,[S2|Ss3]):-
  insert_if_maximal(Ss2,S,Ss3).

