%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% SWI-Prolog compatibility suite
% Copyright 2009--2010 Martin Lazarov
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

switrale_version(0,9,100).
switrale_version :-
    switrale_version(X,Y,Z),
    format("swi trale version: ~w.~w.~w~n", [X,Y,Z]).


:- ensure_loaded(ugraph_coloring).
:- ensure_loaded(ugraph_min_path).

:- multifile user:goal_expansion/2.
user:goal_expansion(if(If, Then, Else), (If *-> Then ; Else)).


stub_warning(Pred) :-
    (   clause(stub_warned_about_(Pred), true) -> true
    ;   format(user_error, "-- WARNING: ~w is a stub~n", [Pred]),
        assert(stub_warned_about_(Pred))
    ).

not_implemented(Pred) :-
    throw(not_implemented(Pred)).

:- multifile abolish_user_preds/1.
abolish_user_preds(all) :-
    first_ale_record_id(First),
    next_ale_id(ale_record, Next),
    retract_ale_records(First, Next),
    retractall(next_ale_id(ale_record,_)),
    assert(next_ale_id(ale_record, First)).

:- dynamic next_ale_id/2.
:- dynamic first_post_grammar_record_id/1.
first_ale_record_id(10000). % not to interfere with edge ids (for performace reasons)
next_ale_id(ale_record, N) :-
    first_ale_record_id(N).

:- multifile compile_gram_hook/0.
compile_gram_hook :-
    next_ale_id(ale_record, Id),
    retractall(first_post_grammar_record_id(_)),
    assert(first_post_grammar_record_id(Id)).

:- multifile clear_hook/0.
clear_hook :-
    first_post_grammar_record_id(Id1),
    next_ale_id(ale_record, Id2),
    %N is Id2 - Id1,
    %trace_message("clear_hook: retracting ~w ale records~n", [N]),
    set_prolog_flag(gc, true),
    retract_ale_records(Id1, Id2),
    set_prolog_flag(gc, false),
    retractall(next_ale_id(ale_record,_)),
    assert(next_ale_id(ale_record, Id1)).

retract_ale_records(Id, IdMax) :-
    Id < IdMax, !,
    recorded(Id, Rec, Ref),
    (   functor(Rec, ale_record, _)
    ->  erase(Ref)
    ;   true
    ),
    Id1 is Id + 1,
    retract_ale_records(Id1, IdMax).
retract_ale_records(IdMax, IdMax).

ale_next_id(Ctx, Id) :-
    (   retract(next_ale_id(Ctx, Id))
    ->  Id1 is Id + 1
    ;   Id = 0, Id1 = 1
    ),
    assert(next_ale_id(Ctx, Id1)).


ale_when(Condition, Goal) :-
    when(Condition, Goal).

ale_dif(X, Y) :-
    when(?=(X, Y), X \== Y).
    %dif(X, Y). % some lex entries won't compile with this

ale_frozen(V, Goal) :-
    trace_term(ale_frozen),
    '$attvar':term_residuals(V, Residuals, []),
    (   Residuals == []
    ->  Goal = true
    ;   Residuals = [H|T],
        (   T == []
        ->  Goal = H
        ;   list_conj(Residuals, Goal)
        )
    ).

list_conj([X,Y], ','(X, Y)) :- !.
list_conj([X|T], ','(X, Y)) :-
    list_conj(T, Y).

:- dynamic edge_data/5.
edge(N,Left,Right,FS,Dtrs,RuleName) :-
    edge_data(N,Left,Right,Dtrs,RuleName),
    recorded(N,edge_fs(FS),_).
    
add_edge_to_chart(edge(N,Left,Right,FS,Dtrs,RuleName)) :-
    %trace_message("edge ~w-~w [~w]~n", [Left, Right, RuleName]),
    recorda(N, edge_fs(FS), _),
    asserta(edge_data(N,Left,Right,Dtrs,RuleName)).

clear_edges :-
    (   retract(edge_data(N,_Left,_Right,_Dtrs,_RuleName)),
        recorded(N, edge_fs(_), Ref),
        erase(Ref),
        fail
    ;   true
    ).

retract_edge(I,Left) :-
    retract(edge_data(I,Left,_Right,_Dtrs,_RuleName)),
    recorded(I, edge_fs(_), Ref),
    erase(Ref).

ale_assert((Head :- Body)) :- !,
    prepare_clause(Head, Body, Clause),
    assert(Clause).
ale_assert(Head) :-
    prepare_clause(Head, true, Clause),
    assert(Clause).

ale_assert((Head :- Body), Ref) :- !,
    not_implemented(ale_assert((head :- non_empty_body), -ref)),
    prepare_clause(Head, Body, Clause),
    assert(Clause, Ref).
ale_assert(Head, Ref) :-
    prepare_clause(Head, true, Clause),
    assert(Clause, Ref).

ale_instance(Ref, (Head :- true)) :-
    clause(Head, Body, Ref),
    call(Body).

nonground_or_cyclic_to_var(X, X) :- ground(X), \+ cyclic_term(X), !.
nonground_or_cyclic_to_var(_, _).

make_head_retriever(F, Args0, Head1) :-
    maplist(nonground_or_cyclic_to_var, Args0, Args1),
    Head1 =.. [F | Args1].

prepare_clause(Head, Body, Clause) :-
    term_variables(Head, HeadVars),
    include(attvar, HeadVars, HeadAttVars),
    Head =.. [F | HeadArgs],
    include(cyclic_term, HeadArgs, CyclicArgs),
    (   HeadAttVars == [], CyclicArgs == []
    ->  Clause = (Head :- Body)
    ;   make_head_retriever(F, HeadArgs, Head1),
        %trace_message("deranged clause: ~w~n", [Head1]),
        ale_next_id(ale_record, RecId),
        %trace_term(ale_next_id(ale_record, RecId)),
        (   Body == true
        ->  recordz(RecId, ale_record(Head), Ref),
            Clause = (Head1 :- recorded(RecId, ale_record(Head1), Ref))
        ;   recordz(RecId, ale_record(Head, Body), Ref),
            Clause = (Head1 :- recorded(RecId, ale_record(Head1, Body), Ref), call(Body))
        )
    ).


% -------------- diagnostics

list_edges_(Left,Right) :-
    edge(_N,Left,Right,_FS,_Dtrs,RuleName),writeln((Left-Right,RuleName)),fail.
list_edges_(_,_).
list_edges_ :- list_edges_(_,_).

num_rules_ :-
    findall(_,rule_compiled(_),Xs),
    length(Xs,N),
    writeln(N).

num_lex_ :-
    findall(_,lex(_,_),Xs),
    length(Xs,N),
    writeln(N).

log_message(Message) :-
    datime(datime(Year,Month,Day,Hour,Minute,Second)),
    open('/tmp/switrale.log', append, S),
    format(S, "~w-~w-~w ~w:~w: ", [Year,Month,Day,Hour,Minute,Second]),
    write(S, Message),
    nl(S),
    close(S).

% -------------- attributed variables

'$put_attributes'(V, TFS) :-
	%trace_message('$put_attributes'(V, TFS)),
    put_attr(V, user, TFS).

'$get_attributes'(V, TFS, Type) :-
	(   get_attr(V, user, TFS)
    ->  arg(1, TFS, Type)
    ;   Type = 0
    )
    %,trace_message('$get_attributes'(V, TFS, Type))
    .

'$delete_attributes'(V) :-
	%trace_message('$delete_attributes'(V)),
    del_attr(V, user).

user:attribute_goals(V, ['$put_attributes'(V, TFS) | GsRest], GsRest) :-
    get_attr(V, user, TFS).

user:attr_portray_hook(Att, _Var) :-
    write_term(Att, [max_depth(2)]).

%verify_attributes(FS1,FS2,PostGoals) :-
%  '$get_attributes'(FS1,TFS1,Type1),
%  (var(FS2) -> '$get_attributes'(FS2,TFS2,Type2),
%               u(Type1,Type2,TFS1,TFS2,FS2,PostGoals)
%  ; 
%    u(Type1,FS2,TFS1,_,FS2,PostGoals)
%  ).

user:attr_unify_hook(TFS1, FS2) :-
	%trace_message(attr_unify_hook:enter),
    arg(1, TFS1, Type1),
    (   var(FS2)
    ->  '$get_attributes'(FS2,TFS2,Type2),
        u(Type1,Type2,TFS1,TFS2,FS2,PostGoals)
    ;   u(Type1,FS2,TFS1,_,FS2,PostGoals)
    ),
    call_list(PostGoals)
    %,trace_message(attr_unify_hook:leave)
    .

call_list([]).
call_list([G|Gs]) :-
    call(G),
    call_list(Gs).


% -------------- call_residue/2

call_residue(Goal, Residue) :-
    call_residue_vars(Goal, RVars),
    '$attvar':term_residuals(RVars, Residue0, []),
    dummy_keys_goals(Residue0, Residue),
    findall(Module, current_module(Module), Modules),
    detach_attributes(RVars, Modules).

dummy_keys_goals([], []).
dummy_keys_goals([G|Gs], [[_]-G|Rest]) :-
    dummy_keys_goals(Gs, Rest).

detach_attributes([], _).
detach_attributes([V|Vs], Modules) :-
    maplist(del_attr(V), Modules),
    detach_attributes(Vs, Modules).

% -------------- compilation

process_alec_throw :-
    expand_eof_to_code(Code), !,
    compile_code(Code).
process_alec_throw.
    
compile('.alec_throw') :- 
    !, process_alec_throw.
compile(Spec) :-
    consult(Spec).

ale_consult('.alec_throw') :-
    !, process_alec_throw.
ale_consult(Spec) :-
    consult(Spec).

%abolish_if_not_multifile(Name/Arity) :-
%    functor(F, Name, Arity),
%    (   predicate_property(F, multifile) ->  true
%    ;   abolish(Name/Arity)
%    ).

compile_code(Code0) :-
    prepare_code(Code0, Code, [], Predicates),
    %maplist(abolish_if_not_multifile, Predicates),
    maplist(abolish, Predicates),
    maplist(ale_assert, Code),
    compile_predicates(Predicates).

prepare_code([], [], Ps0, Ps) :-
    !, sort(Ps0, Ps).
prepare_code([end_of_file|_], [], Ps0, Ps) :-
    !, sort(Ps0, Ps).
prepare_code([(:- _Directive)|Cs], Code, Ps0, Ps) :-
    !, prepare_code(Cs, Code, Ps0, Ps).
prepare_code([C|Cs], [C|Code], Ps0, Ps) :-
    clause_head_predspec(C, Head, PredSpec),
    (   predicate_property(Head, dynamic)
    ->  prepare_code(Cs, Code, Ps0, Ps)
        %,trace_message("will not compile dynamic predicate: ~w~n", [PredSpec])
    ;   prepare_code(Cs, Code, [PredSpec|Ps0], Ps)
    ).
    %prepare_code(Cs, Code, [PredSpec|Ps0], Ps).

clause_head_predspec((Head :- _Body), Head, F/A) :- !, functor(Head, F, A).
clause_head_predspec(Head, Head, F/A) :- functor(Head, F, A).


% -------------- exceptions

:- meta_predicate
    on_exception(+, :, :).
on_exception(Ex, Goal, Recover) :-
    catch(Goal, Ex, Recover).

raise_exception(Ex) :-
    throw(Ex).


% -------------- lists

nth(Index, List, Elem) :-
    nth1(Index, List, Elem).

same_length([], []).
same_length([_|T1], [_|T2]) :-
    same_length(T1, T2).

remove_duplicates([], []).
remove_duplicates([Head|Tail1], [Head|Tail2]) :- 
    delete(Tail1, Head, Residue),
    remove_duplicates(Residue, Tail2).


% -------------- flags
:- dynamic '$ale_swidefs__prolog_flag_version'/1.
:-  (   '$ale_swidefs__prolog_flag_version'(_)
    ->  true
    ;   prolog_flag(version, Version),
        assert('$ale_swidefs__prolog_flag_version'(Version)),
        abolish(prolog_flag/2)
    ).

prolog_flag(F, V) :-
    (   var(F) 
    ->  F = version,
        '$ale_swidefs__prolog_flag_version'(V)
    ;   F == version
    ->  '$ale_swidefs__prolog_flag_version'(V), !
    ;   fail
    ).
prolog_flag(F, V) :-
    current_prolog_flag(F, V), F \== version.

prolog_flag(F, V0, V1) :-
    current_prolog_flag(F, V0),
    set_prolog_flag(F, V1).


% -------------- streams

skip_line :-
    skip(10).
skip_line(Stream) :-
    skip(Stream, 10).

ttyflush :-
    flush_output.

ttyget0(Ch) :-
    get0(Ch).

ttynl :-
    nl.

at_end_of_line(Stream) :-
    peek_byte(Stream, 10).

% -------------- etc

term_subsumer(_Term1, _Term2, _General) :-
    not_implemented(term_subsumer/3).

del_assoc(_,_,_,_) :-
    % missing from swi
    % it came as a surprise that there's no means to diassociate keys
    not_implemented(del_assoc/4).

ale_if(If, Then, Else) :-
    If *-> Then ; Else.

ale_statistics(walltime, Value) :-
    !, statistics(real_time, Value). % or should we use system_time instead?
ale_statistics(choice, [0,1]) :-
    !. %% todo
ale_statistics(heap, [Used,0]) :-
    !, statistics(heapused, Used).
ale_statistics(trail, [Used,0]) :-
    !, statistics(trailused, Used).
ale_statistics(Key, Value) :- statistics(Key, Value).


file_exists(F) :-
    exists_file(F).
file_exists(F, [Mode]) :-
    access_file(F, Mode).


environ(Name, Value) :-
    getenv(Name, Value).


ale_abolish(Term) :-
    var(Term), !,
    throw(instantiation_error(ale_abolish(Term),1)).
ale_abolish([]).
ale_abolish([P|Ps]) :-
    ale_abolish(P),
    ale_abolish(Ps).
ale_abolish(Pred/Arity) :-
    abolish(Pred/Arity).
ale_abolish(Pred,Arity) :-
    abolish(Pred/Arity).


now(X) :-
    system:get_time(X).

datime(Datime) :-
    now(Time),
    datime(Time,Datime).
datime(Time,datime(Year,Month,Day,Hour,Minute,Second)) :-
    stamp_date_time(Time, date(Year,Month,Day,Hour,Minute,SecondF,_Off,_TZ,_DST), local),
    Second is floor(SecondF).


host_name(X) :-
    gethostname(X).


% differs from SICS' version in that it executes the goal using once/1,
% not call/1
time_out(Goal, TimeMs, Result) :-
    Time is TimeMs / 1000.0,
    catch((
        call_with_time_limit(Time, Goal),
        Result = success
        ),
        time_limit_exceeded,
        Result = time_out
    ).


% copied over from trale_emacs.pl
number_of(Predicate,Number):-
    (   current_predicate(_,Predicate)
    ->  findall(_,Predicate,List),
        length(List,Number)
    ;   Number = 0
	).


:- use_module(library(charsio)).
:- ( current_predicate(charsio:format_to_chars/3) -> true
   ; ensure_loaded(format_to_chars)
   ).

read_line(Line) :-
    read_line_to_codes(user_input, Line).


get_mutable(_,_) :-
    fail.


% swi_to_sics_layout(+TermPos, +PosToLine, -Layout)
%
% Adapts subterm positions as returned by SWI's read_term/3
% (character positions) to SICStus' layout positions (line positions).
% The current implementation uses a complex term to map char to line positions.
% It's a somewhat memory inefficient solution, which is likely to break on
% larger files (it works fine for file sizes of a couple of megs) but it's
% also a trivial one to implement in comparision to the alternatives.

swi_to_sics_layout(TermPos, PosToLine, Layout) :-
    map_swi_to_sics_layout(TermPos, Layout, PosToLine).

map_swi_to_sics_layout([], [], _PosToLine).
map_swi_to_sics_layout(none, [], _PosToLine).
map_swi_to_sics_layout(From-_To, FromLine, PosToLine) :-
    pos_to_line(From, PosToLine, FromLine).
map_swi_to_sics_layout(string_position(From, _To), FromLine, PosToLine) :-
    pos_to_line(From, PosToLine, FromLine).
map_swi_to_sics_layout(brace_term_position(From, _To, _Arg), FromLine, PosToLine) :-
    pos_to_line(From, PosToLine, FromLine).
map_swi_to_sics_layout(term_position(From, _To, _FFrom, _FTo, SubPos), [FromLine|T], PosToLine) :-
    pos_to_line(From, PosToLine, FromLine),
    map_swi_to_sics_layout_list(SubPos, T, PosToLine).
map_swi_to_sics_layout(
    list_position(From, To, [FirstElt | RestElts], Tail),
    [LBrackLine, EltLayout, TailLayout],
    PosToLine
    ) :-
    pos_to_line(From, PosToLine, LBrackLine),
    map_swi_to_sics_layout(FirstElt, EltLayout, PosToLine),
    (   RestElts == []
    ->  (   Tail == none                    % only one element in list
        ->  pos_to_line(To, PosToLine, TailLayout)
        ;   map_swi_to_sics_layout(Tail, TailLayout, PosToLine)
        )
    ;   map_swi_to_sics_layout_rest(RestElts, TailLayout, RestLayoutTail, PosToLine),
        (   Tail == none
        ->  pos_to_line(To, PosToLine, ToLine),
            RestLayoutTail = [ToLine]
        ;   map_swi_to_sics_layout(Tail, TailRes, PosToLine),
            RestLayoutTail = [TailRes]
%            (   Tail = list_position(_,_,_,_)
%            ->  RestLayoutTail = [TailRes]
%            ;   RestLayoutTail = TailRes
%            )
        )
    ).

term_left_edge([H|_], H1) :- !,
    term_left_edge(H, H1).
term_left_edge(H, H).

map_swi_to_sics_layout_rest([H], [LeftEdge,H1|Tail], Tail, PosToLine) :- !,
    map_swi_to_sics_layout(H, H1, PosToLine),
    term_left_edge(H1, LeftEdge).
map_swi_to_sics_layout_rest([H|T], [LeftEdge,H1,T1], RestLayoutTail, PosToLine) :-
    map_swi_to_sics_layout(H, H1, PosToLine),
    term_left_edge(H1, LeftEdge),
    map_swi_to_sics_layout_rest(T, T1, RestLayoutTail, PosToLine).

map_swi_to_sics_layout_list([], [], _PosToLine).
map_swi_to_sics_layout_list([H|T], [H1|Rest], PosToLine) :-
    map_swi_to_sics_layout(H, H1, PosToLine),
    map_swi_to_sics_layout_list(T, Rest, PosToLine).

make_pos_to_line_map(File, [[[PosToLine]]]) :-
    open(File, read, FS),
    get_code(FS, Code),
    make_pos_to_line_map(Code, 1, FS, PosToLineList),
    PosToLine =.. ['$arg_to_line' | PosToLineList].

make_pos_to_line_map(-1, Line, FS, [Line]) :- !,
    close(FS).
make_pos_to_line_map(10, Line, FS, [Line|Rest]) :- !,
    Line1 is Line + 1,
    get_code(FS, Code),
    make_pos_to_line_map(Code, Line1, FS, Rest).
make_pos_to_line_map(_Code, Line, FS, [Line|Rest]) :-
    get_code(FS, Code),
    make_pos_to_line_map(Code, Line, FS, Rest).

pos_to_line(I0, [[[PosToLine]]], A) :-
    I1 is I0 + 1,
    arg(I1, PosToLine, A).
