% $Id: feature_ordering.pl,v 1.10 2003/05/08 16:39:18 dm Exp $
%
% This file defines code for ordering features in feature structures
% that are displayed.
%
% Author: Detmar Meurers
% Modified by: Gerald Penn
%
% The user specifies feature order as follows:
%
% f <<< g.    %% meaning: f will be ordered before g.
% <<< h.      %% meaning: h will be ordered last. --- no support for this in rec/1 yet
% >>> i.      %% meaning: i will be ordered first. 
%
% Note: These statements are identical to the ones that were
% used by the ConTroll system. The ConTroll feature_ordering code
% was buggy though, so this code is rewritten from scratch.
%
% Added feature hiding based on hidden_feat/1 statements in the grammar
% or the standard ale commands no_write_feat/1.
%
% sort_keyed_feats/3 is now the sole entry point from ghooks.pl.  Also, transitive closures
%  are the wrong operation here - should be a topological sort.
% 1/22/09 - G. Penn
%
% Bug corrected: topo_sort_feats/3 was returning unsorted template.
% 5/5/09 - G. Penn
%
% Added preference for lexicographic ordering in the absence of ordering constraints.
% 5/5/09 - G. Penn

:- use_module(library(lists),[member/2]).
:- use_module(library(ugraphs),[transitive_closure/2,transpose/2]).

:- use_module('sicstus_additions/samsort', [samsort/3]).

:- op(850,xfy,<<<).
:- op(850,fx,<<<).
:- op(850,fx,>>>).

:- multifile ale_exception/1, recoverable_ale_exception/1.
ale_exception(feat_order_before_undef(Before,After)) :-
  prolog_flag(toplevel_print_options,Options,Options),
  write_term(user_error,(Before <<< After),Options),
  format(user_error,' - ~a not declared in signature',[Before]).
ale_exception(feat_order_after_undef(Before,After)) :-
  prolog_flag(toplevel_print_options,Options,Options),
  write_term(user_error,(Before <<< After),Options),
  format(user_error,' - ~a not declared in signature',[After]).
ale_exception(feat_order_cycle(Feat,Cycle)) :-
  prolog_flag(toplevel_print_options,Options,Options),
  format(user_error,'cycle at ~a exists in feature order declarations: ~@',[Feat,write_term(Cycle,Options)]).
ale_exception(feat_order_sisters_initial(InitFeat1,InitFeat2,AppropType)) :-
  prolog_flag(toplevel_print_options,Options,Options),
  write_term(user_error,[(>>> InitFeat1), (>>> InitFeat2)],Options),
  format(user_error,' - features are both appropriate to ~a, and therefore cannot both be initial',[AppropType]).
ale_exception(feat_order_before_initial(Before,InitFeat)) :-
  prolog_flag(toplevel_print_options,Options,Options),
  write_term(user_error,[(>>> InitFeat), (Before <<< InitFeat)],Options),
  format(user_error,' - a feature cannot precede an initial feature',[]).
recoverable_ale_exception(feat_order_before_undef(_,_)).
recoverable_ale_exception(feat_order_after_undef(_,_)).
recoverable_ale_exception(feat_order_cycle(_,_)).
recoverable_ale_exception(feat_order_sisters_initial(_,_,_)).
recoverable_ale_exception(feat_order_before_initial(_,_,_)).

:- dynamic feat_topo_order/2, topo_sort_feats_cached/3.

abolish_feature_ordering :-
%   abolish(hidden_feat/1),  % GBP - we should register features as hidden, shown, or "shown only when interesting"
   retractall(feat_topo_order(_,_)),
   retractall(topo_sort_feats_cached(_,_)),
   abolish((<<<)/1),
   abolish((<<<)/2),
   abolish((>>>)/1).  % maybe this should be relativized by type.

compute_feature_ordering :-
  on_exception(ale(Exception),compute_feature_ordering_act,alex(Exception)).
	       
compute_feature_ordering_act :-
  features_are_ordered ->
    ( ( feat_precedes(Before,After), \+ approp(Before,_,_) ) -> raise_exception(ale(feat_order_before_undef(Before,After)))
    ; ( feat_precedes(Before,After), \+ approp(After,_,_) ) -> raise_exception(ale(feat_order_after_undef(Before,After)))
    ; ( feat_initial(InitFeat), feat_precedes(Before,InitFeat) ) -> raise_exception(ale(feat_order_before_initial(Before,InitFeat)))
    ; ( feat_initial(InitFeat1), feat_initial(InitFeat2), InitFeat1 \== InitFeat2,
	 approp(InitFeat1,T,_), approp(InitFeat2,T,_) )
      -> raise_exception(ale(feat_order_sisters_initial(InitFeat1,InitFeat2,T)))
    ; findall(Before-After,(feat_precedes(Before,After),
			    feat_not_initial(Before)),RelFeatOrders),
      findall(F,(feature(F),
		 feat_not_initial(F)),NonInitFeats),
      vertices_edges_to_ugraph(NonInitFeats,RelFeatOrders,FeatOrderCover),
      ( alpha_top_sort(FeatOrderCover,FeatTopoOrderRest) -> esetof(InitFeat,feat_initial(InitFeat),InitFeats),
	                                                    append(InitFeats,FeatTopoOrderRest,FeatTopoOrder),
	                                                    topo_enumerate_feats(FeatTopoOrder,1)
      ; min_path(F,F,FeatOrderCover,Cycle,_),
	raise_exception(ale(feat_order_cycle(F,Cycle)))
      )
    )
  ; true.

features_are_ordered :-
  current_predicate('<<<',(_ <<< _)),
  (_ <<< _),
  !.
features_are_ordered :-
  current_predicate('>>>',(>>> _)),
  (>>> _),
  !.

feat_precedes(Before,After) :-
  current_predicate('<<<',(_ <<< _)),
  (Before <<< After).

feat_initial(Feat) :-
  current_predicate('>>>',(>>> _)),
  (>>> Feat).

feat_not_initial(Feat) :-
  current_predicate('>>>',(>>> _)) -> \+ (>>> Feat) ; true.

topo_enumerate_feats([],_).
topo_enumerate_feats([F|Fs],N) :-
  assert(feat_topo_order(F,N)),
  NPlus1 is N + 1,
  topo_enumerate_feats(Fs,NPlus1).

topo_sort_feats(Type,KeyedFeats,SortedFeatsTemplate) :-
  topo_sort_feats_template(KeyedFeats,TopoFeatsTemplate,KeyedFeatsTemplate),
  keysort(KeyedFeatsTemplate,SortedKeyedFeatsTemplate),
  strip_keys(SortedKeyedFeatsTemplate,SortedFeatsTemplate),
  assert(topo_sort_feats_cached(Type,TopoFeatsTemplate,SortedFeatsTemplate)),
  TopoFeatsTemplate=KeyedFeats.

topo_sort_feats_template([],[],[]).
topo_sort_feats_template([fval(F,_,_)|KeyedFeats],[TemplVar|TVs],[N-TemplVar|KeyedTVs]) :-
  clause(feat_topo_order(F,N),true),
  topo_sort_feats_template(KeyedFeats,TVs,KeyedTVs).

strip_keys([],[]).
strip_keys([_-Datum|KeyedDs],[Datum|Ds]) :-
  strip_keys(KeyedDs,Ds).

sort_keyed_feats(Type,KeyedFeats,SortedKeyedFeats) :-
  features_are_ordered -> ( clause(topo_sort_feats_cached(Type,KeyedFeats,SortedKeyedFeats),true) -> true
			  ; topo_sort_feats(Type,KeyedFeats,SortedKeyedFeats)
			  )
  ; SortedKeyedFeats = KeyedFeats.

% this is a variant of the SICStus built-in top_sort/2 that defaults to lexicographic ordering
alpha_top_sort([],[]) :- !.
alpha_top_sort(Graph,AlphaTopo) :-
  build_marked_ugraph(Graph,MarkedGraph),
  ord_list_to_assoc(MarkedGraph,MarkedGraphAssoc),
  transpose(Graph,Transpose),      % by traversing nodes in reverse lexicographic order, we ensure that high
  reverse(Transpose,RevTranspose), % elements in the order are pushed onto the accumulator first
  visit_root_nodes(RevTranspose,MarkedGraphAssoc,[],AlphaTopo),
  AlphaTopo \== []. % if so, then there are no root nodes, and therefore this is a cycle.

build_marked_ugraph([],[]).
build_marked_ugraph([V-Neibs|G],[V-m(_Marker,RevNeibs)|Ms]) :-
  reverse(Neibs,RevNeibs),  % traverse these in reverse, too
  build_marked_ugraph(G,Ms).

visit_root_nodes([],_,Accum,Accum).
visit_root_nodes([V-Neibs|RevTpose],MGAssoc,Accum0,Accum) :-
  ( Neibs==[] -> depth_first_traverse_marked_ugraph(V,MGAssoc,Accum0,Accum1)
  ; Accum1 = Accum0
  ),
  visit_root_nodes(RevTpose,MGAssoc,Accum1,Accum).

depth_first_traverse_marked_ugraph(V,MGAssoc,Accum0,Accum) :-
  get_assoc(V,MGAssoc,M), arg(1,M,Marker), arg(2,M,RevNeibs),
  ( ground(Marker) -> Accum = Accum0
  ; var(Marker) -> Marker = open(X),
                   depth_first_traverse_marked_ugraph_act(RevNeibs,MGAssoc,Accum0,Accum1),
                   X = closed, Accum = [V|Accum1]
% ; fail - there is a cycle       
  ).

depth_first_traverse_marked_ugraph_act([],_,Accum,Accum).
depth_first_traverse_marked_ugraph_act([V|Vs],MGAssoc,Accum0,Accum) :-
  depth_first_traverse_marked_ugraph(V,MGAssoc,Accum0,Accum1),
  depth_first_traverse_marked_ugraph_act(Vs,MGAssoc,Accum1,Accum).

% GBP: Above this line is what new ghooks call
% --------------------------------------------------------------------------------

:- dynamic feature_order/2, feature_order_first/1, feature_order_last/1.

order_features(Fs0,Fs) :-
   samsort(feat_order_rel,Fs0,Fs).

% initialization routines -----

% clear_feature_output/0  abolishes source declarations
clear_feature_output :-
   abolish(hidden_feat/1),
   abolish((<<<)/1),
   abolish((<<<)/2),
   abolish((>>>)/1).

process_feature_output :-
   process_feature_hiding,
   process_feature_order.

process_feature_hiding :-
   retractall(no_write_feat_flag(_)),
   initialize_feature_hiding.

process_feature_order :-
   retractall(feature_order(_,_)),
   retractall(feature_order_first(_)),
   retractall(feature_order_last(_)),
   initialize_feature_order.

% -----

initialize_feature_hiding :-
   fail_if_undefined(hidden_feat(F)),
   no_write_feat(F),
   fail;
   true.

initialize_feature_order :-
   % Deal with the binary order specifications:
   (   % get a graph G from user specified order predicates
       setof(X-YList,setof(Y,feat_order_spec(X,Y),YList),G)
   ->  % compute transitive closure of G
       transitive_closure(G,TransG),
       assert_trans_feat_ord(TransG)
   ;   true			% no feature order specified
   ),
   % Transfer the first and last specifications into an asserted
   % internal representation, which allows us to check for errors here.
   (  feat_first_spec(F),
      assert(feature_order_first(F)),
      fail
   ;
      true
   ),
   (  feat_last_spec(F),
      assert(feature_order_first(F)),
      fail
   ;
      true
   ).
   
   
  
assert_trans_feat_ord([]).
assert_trans_feat_ord([X-L|Rest]) :-
        member(Y,L),
	assert(feature_order(X,Y)),
	fail
	;
	assert_trans_feat_ord(Rest).

% -----

feat_order_rel(fval(X,_,_),fval(Y,_,_)) :-
   feat_order_rel_aux(X,Y).

% Feature order relation:
feat_order_rel_aux(X,_) :-	% enforce feat_first order statement
   feature_order_first(X),
   !.
feat_order_rel_aux(_,Y) :-	% enforce feat_last order statement
   feature_order_last(Y),
   !.
feat_order_rel_aux(X,Y) :-	% enforce feature order statement
   \+ feature_order_first(Y),	% unless conflicting with feat_first
   \+ feature_order_last(X),	%     or conflicting with feat_last
   feature_order(X,Y),
   !.
feat_order_rel_aux(X,Y) :-	% if it doesn't contradict
   \+ feature_order_first(Y),	% - feat_first statement
   \+ feature_order_last(X),	% - feat_last statement
   \+ feature_order(Y,_),	% - captures contradiction with Y < X
				%   as well as breaking transitivity chain
   X @=< Y.			% then enforce standard term order

% The breaking of the transitivity chain occurs at last element, e.g.
% if we have features a, b, c, and d  and the statement d <<< a in the
% grammar, then we get d < a < b < c since a @=<b @=< c but since d <<< _
% is in the grammar the \+ feature_order(Y,_) stops us from enforcing c @=< d. 

% -----

% predicates for looking up the user specifications:

feat_order_spec(X,Y) :-
   fail_if_undefined(X<<<Y),
   (  feature(X),feature(Y)
   -> true
   ;  (  feature(X)
      -> format(user_error,'~N**Warning: ignoring feature ordering "~w" for unintroduced feature "~w".~n',[X<<<Y,Y]),fail
      ;  (  feature(Y)
	 -> format(user_error,'~N**Warning: ignoring feature ordering "~w" for unintroduced feature "~w".~n',[X<<<Y,X]),fail
	 ;  format(user_error,'~N**Warning: ignoring feature ordering "~w" for unintroduced features "~w" and "~w".~n',[X<<<Y,X,Y]),fail
	 )
      )
   ).

feat_first_spec(First) :-
   fail_if_undefined((>>> First)),
   (  feature(First)
   -> true
   ;  format(user_error,'~N**Warning: ignoring feature ordering "~w" for unintroduced feature "~w".~n',[>>>First,First]),fail
   ).

feat_last_spec(Last) :-
   fail_if_undefined((<<< Last)),
   (  feature(Last)
   -> true
   ;  format(user_error,'~N**Warning: ignoring feature ordering "~w" for unintroduced feature "~w".~n',[<<<Last,Last]),fail
   ).

fs2keyed_feats(FS,KeyedFeats,Type) :-
  get_type(FS,Type),
  approps(Type,FRs,_),
  build_keyed_feats(FRs,FS,KeyedFeats).


%% -*-Mode: prolog;-*-

/* % some useful predicates that currently aren't used:

% sister_feats(+Feat,list(Feats): given a feature, returns the list of 
% other features appropriate for the type or its subtypes
sister_feats(F,L) :-
   setof(SisterF,T^T1^T2^(approp(F,T,T1),approp(SisterF,T,T2)),L).

% list of all features:
all_feats(AllFeats) :-
   setof(F,T0^T^approp(F,T0,T),AllFeats).
*/
