% $Id: fs2desc.pl,v 1.3 2002/10/14 20:31:52 dm Exp $
% 
% Defines a pretty printer which prints TRALE descriptions.
% Author: Gerald Penn 
%
% GBP - Updated to new data structure, and changed fs2desc/2
%  so that it doesn't strip off inequations.
%
% $Log: fs2desc.pl,v $
% Revision 1.3  2002/10/14 20:31:52  dm
% extensionalise/3 is now extensionalise/2 - changed call to it here
%
% Revision 1.2  2002/08/07 04:42:58  dm
% Added fs2desc/2 predicate as top-level goal here.
%

:- use_module(library(assoc)).
:- use_module(library(terms),[subsumes_chk/2]).

% ================================================
% fs2desc(+FS:<fs>, -Desc:<desc>) 
%
% Returns the ALE description Desc which is the
% most general satisfier of FS with inequations Iqs
%

fs2desc(FS,Desc) :-
  fs2desc(FS,Desc,_).

fs2desc(FS,Desc,IqGoal) :-
   frozen_term(FS,Frozen),
   pp_desc(FS,Frozen,Desc,_FSFrozen,IqGoal).

% ------------------------------------------------------------------------------
% pp_desc(+Tag,+SVs,-Desc)
% pp_desc(+Tag,+SVs,+Iqs,-Desc,+PredName,-IqGoal,-IqPreds)
% ------------------------------------------------------------------------------
% Convert the feature structure Tag-SVs/Iqs to a description, Desc, logic
% program IqPreds, and goal, IqGoal, such that query((X=Desc,IqGoal))
% over IqPreds iteratively succeeds with X bound to an alpha-variant of
% Tag-SVs, non-deterministically enumerating every satisfying selection of
% conjuncts from Iqs.
%   In the absence of extensional types, there is a unique selection, in which
% case X is isomorphic to Tag-SVs/Iqs in the inequational algebra.  In the
% absence of inequations, the most general satisfier of Desc is Tag-SVs.  Desc
% is the minimal description, relative to the type system cum appropriateness,
% whose most general satisfier is Tag-SVs, although there may exist smaller
% descriptions in the presence of universal constraints.
%
% Tag-SVs/Iqs is assumed to have been extensionalised, fully dereferenced,
% and pruned.  This can be achieved by calling:
%   extensionalise(Tag0,SVs0),
%   fully_deref_prune(Tag0,SVs0,Tag,SVs,Iqs0,Iqs)
% before calling pp_desc(Tag,SVs,Iqs,_,_,_,_).
%
% The caller must choose the name of the logic programming predicate, PredName,
% to accompany the description.
% ------------------------------------------------------------------------------

pp_desc(FS,Residue,Desc,FSFrozen,IqGoal) :-
  build_iqs_restore_atts(Residue,Iqs,FSFrozen),
  pp_desc(FS,Iqs,Desc,IqGoal).

pp_desc(FS,Iqs,Desc,IqGoal) :-
  empty_assoc(VarsIn),
  ppd_dups(FS,Iqs,VarsIn,VarsMid2),
% DESCRIPTION GENERATION
  ppd_fs(FS,bot,VarsMid2,VarsOut,Desc,desc),
  (Iqs == [] -> IqGoal = true
% INEQUATION GENERATION
  ; ppd_iqs(Iqs,VarsOut,IqGoal)
  ).

% ------------------------------------------------------------------------------
% ppd_dups(+Tag,+SVs,+Iqs,+VarsIn,-VarsOut)
% ------------------------------------------------------------------------------
% Traverse Tag-SVs/Iqs to create an association list of reference tags.  If
% a tag appears more than once, its value in VarsOut is bound to a pair
% [DVar|Visited], where DVar is a description variable, and Visited is a
% flag used to indicate whether a tag is visited first by the description
% generator or the inequation generator.
% ------------------------------------------------------------------------------
ppd_dups(FS,Iqs,VarsIn,VarsOut) :-
  ppd_dups_fs(FS,VarsIn,VarsMid),
  ppd_dups_iqs(Iqs,VarsMid,VarsOut).

ppd_dups_fs(FS,VarsIn,VarsOut) :-
  get_assoc(FS,VarsIn,var(_,_,_)) -> VarsOut = VarsIn
  ; put_assoc(FS,VarsIn,_,VarsMid),
    get_type_or_bot(FS,Type), approps(Type,FRs,_),
    ppd_dups_vs(FRs,FS,VarsMid,VarsOut).

ppd_dups_vs([],_,Vars,Vars).
ppd_dups_vs([F:_|FRs],FS,VarsIn,VarsOut) :-
  clause(fcolour(F,K,_),true), arg(K,FS,V),
  ppd_dups_fs(V,VarsIn,VarsMid),
  ppd_dups_vs(FRs,FS,VarsMid,VarsOut).

ppd_dups_iqs([],Vars,Vars).
ppd_dups_iqs([ineq(FS1,FS2)|Iqs],VarsIn,VarsOut) :-
  ppd_dups_fs(FS1,VarsIn,VarsMid),
  ppd_dups_fs(FS2,VarsMid,VarsMid2),
  ppd_dups_iqs(Iqs,VarsMid2,VarsOut).

% ------------------------------------------------------------------------------
% ppd_fs(+Tag,+SVs,+R,+VarsIn,-VarsOut,-Desc,-VisFlag)
% ------------------------------------------------------------------------------
% Desc is a minimal description whose most general satisfier is Tag-SVs,
% in the context of type inferencing mode, R.  VisFlag indicates whether
% this predicate was called in the context of description or inequation
% generation.  VarsOut is VarsIn, with the type modes of variables in Desc
% updated.
% ------------------------------------------------------------------------------

ppd_fs(FS,R,VarsIn,VarsOut,Desc,VisFlag) :-
  get_assoc(FS,VarsIn,TagBind),
% if this is a singular tag, don't bind a variable to it
  ( TagBind = nobind -> ppd_fs_act(FS,R,VarsIn,VarsOut,Descs,VisFlag),
                        list_to_desc(Descs,Desc)
  ; TagBind = var(DVar,VMode,Visited),
% otherwise, if this is our first visit, mark with VisFlag, and bind DVar
    ( var(VMode) -> VMode = bot ; true),  % initialise type mode of DVar
    unify_type(R,VMode,RNew),  % unify with current type mode
    ( sub_type(RNew,VMode) -> VarsMid = VarsIn  
    ; put_assoc(FS,VarsIn,var(DVar,RNew,Visited),VarsMid) 
    ),  % update type mode of DVar
    ( var(Visited) -> Visited = VisFlag,
	              ppd_fs_act(FS,RNew,VarsMid,VarsOut,Descs,VisFlag),
                      list_to_desc(Descs,DVar,Desc)
% if it's not our first visit, use DVar only - the first visit is 
%  responsible for the rest of the description.
    ; Desc = DVar, VarsOut = VarsMid
    )
  ).

ppd_fs_act(FS,R,VarsIn,VarsOut,Descs,VisFlag) :-
  get_type_or_bot(FS,Type),
  approps(Type,FRs,_),
  ppd_vs(FRs,FS,R,TypInf,VarsIn,VarsOut,FDescs,VisFlag),
% if the type can be inferred from the mode plus type inferencing on
% feature names, leave the type out.
  (sub_type(Type,TypInf) -> Descs = FDescs ; Descs = [Type|FDescs]).

ppd_vs([],_,TypInf,TypInf,Vars,Vars,[],_).
ppd_vs([F:R|FRs],FS,TypIn,TypOut,VarsIn,VarsOut,FDescs,VisFlag) :-
  clause(fcolour(F,K,_),true), arg(K,FS,V),
  ppd_fs(V,R,VarsIn,VarsMid,FDesc,VisFlag),
% bot means that it is not necessary to add information to this feature
% in this context.
  ( FDesc == bot -> ppd_vs(FRs,FS,TypIn,TypOut,VarsMid,VarsOut,FDescs,VisFlag)
  ; introduce(F,FIntro),
    unify_type(FIntro,TypIn,TypMid),
    FDescs = [F:FDesc|FDescsRest],
    ppd_vs(FRs,FS,TypMid,TypOut,VarsMid,VarsOut,FDescsRest,VisFlag)
  ).

% ------------------------------------------------------------------------------
% ppd_iqs(+Iqs,+Vars,-IqBody)
% ------------------------------------------------------------------------------
% Convert a non-empty set of inequations into a CNF of ineq/2 calls that
% will enumerate the acceptable conjunctions of inequations.  This is the
% best we can do in general, because a disjunction may arise from decomposition
% of an inequation attached to an extensionally typed FS not accessible from
% Tag-SVs, and there is no two-place inequation relation in the description
% language.  We could be a little more careful about variable type mode here,
% but that would involve generalising the types of variables after every
% disjunction.
% ------------------------------------------------------------------------------

ppd_iqs([Iq|Iqs],Vars,IqGoal) :-
  ppd_iqs(Iqs,Iq,Vars,IqGoal).

ppd_iqs([],Iq,Vars,IqGoal) :-
  ppd_iq(Iq,Vars,_VarsOut,IqGoal).
ppd_iqs([Iq|Iqs],Iq0,VarsIn,(IqGoal0,IqGoal)) :-
  ppd_iq(Iq0,VarsIn,VarsOut,IqGoal0),
  ppd_iqs(Iqs,Iq,VarsOut,IqGoal).

ppd_iq(ineq(FS1,FS2),VarsIn,VarsOut,add_ineq(Desc1,Desc2)) :-
  ppd_fs(FS1,bot,VarsIn,VarsMid,Desc1,ineq),
  ppd_fs(FS2,bot,VarsMid,VarsOut,Desc2,ineq).

% ==============================================================================
% UTILITIES
% ==============================================================================

% convert list to conjunctive ALE description
list_to_desc([],bot).     
list_to_desc([Desc|Descs],D0) :-
  list_to_desc(Descs,Desc,D0).

list_to_desc([],Desc,Desc).
list_to_desc([Desc|Descs],D0,(D0,DRest)) :-
  list_to_desc(Descs,Desc,DRest).  

% isolate those Vs that were first visited by the description generator.
desc_bound_vars([],_,[]).
desc_bound_vars([V|Vs],Vars,DBs) :-
  get_assoc(V,Vars,var(_,_,Visited)),
  (Visited == desc -> DBs = [V|DBsRest] ; DBsRest = DBs),
  desc_bound_vars(Vs,Vars,DBsRest).

% this is an interpreted version of the ineq/2 predicate in the ALE library.
add_ineq(Desc1,Desc2) :-
  add_to(Desc1,FS1),
  add_to(Desc2,FS2),
  ineq(FS1,FS2).

% MLA: wraps get_type, returning bot instead of 0
get_type_or_bot(FS, Type) :-
    get_type(FS, TypeOrZero),
    (   TypeOrZero == 0
    ->  Type = bot
    ;   Type = TypeOrZero
    ).
