:- use_module(library(socket), [
    tcp_socket/1
    , tcp_bind/2
    , tcp_accept/3
    , tcp_open_socket/3
    , tcp_close_socket/1
    ]).
% :- use_module(library(charsio), [write_to_chars/2]).
% :- use_module(library(lists), [substitute/4]).

:- ensure_loaded(trale_home(ghooks)).
%:- ensure_loaded(tokenization).

trale_server_start(Port) :-
    init_tokenization,
    tcp_socket(ServerSocket),
    tcp_bind(ServerSocket, Port),
    tcp_listen(ServerSocket, 1),
    format("waiting for client on port ~w~n", [Port]),
    tcp_accept(ServerSocket, ClientSocket, IPAddr),
    format("-- client: ~w~n", [IPAddr]),
    tcp_close_socket(ServerSocket),
    tcp_open_socket(ClientSocket, InStream, OutStream),
    asserta(socket_info(_, OutStream)),
    trale_server_toplevel(InStream),
    close(InStream),
    close(OutStream),
    % tcp_close_socket(ClientSocket),
    retractall(socket_info(_, _)), !.

trale_server_loop(Port) :-
    trale_server_start(Port),
    trale_server_loop(Port).

trale_server_toplevel(Stream) :-
    repeat,
    catch((
        read(Stream, Term),
        writeln(read_term(Term)),
        (   Term = rec(X, MaxResults, MaxWords)
                            ->  Command = (wt_recmax(X, MaxResults, MaxWords), fail)
        ;   Term = rec(X, MaxResults)
                            ->  Command = (wt_recmax(X, MaxResults, 10), fail)
        ;   Term = rec(X)   ->  Command = (wt_recmax(X, 100, 100), fail)
        ;   Term = lex(X)   ->  Command = (wt_lexall(X), fail)
        ;   Term = words    ->  Command = (wt_lexicon, fail)
        ;   Term = quit     ->  Command = true
        ;   Term = end_of_stream ->  Command = true
        ;                       Command = fail
        ),
        once(Command)
        ),
        Exception,
        (advice_exception(Exception), fail)
    ).

substitute(X1,Chars1,X2,Chars2):-
    substitute_(Chars1,Chars2,X1,X2).
substitute_([],[],_X1,_X2).
substitute_([X1|T1],[X2|T2],X1,X2) :-!,
    substitute_(T1,T2,X1,X2).
substitute_([H|T1],[H|T2],X1,X2) :-
    substitute_(T1,T2,X1,X2).
advice_exception(Ex) :-
    writeln(exception:Ex),
    get_output_stream(S),
    write(S, '*'),
    with_output_to(chars(ExChars0), write(Ex)),
    substitute(0'., ExChars0, 0'?, ExChars), % '
    atom_codes(ExAtom, ExChars),
    write(S, ExAtom), nl(S),
    terminate_output(S).

terminate_output :-
    get_output_stream(S),
    terminate_output(S).
    
terminate_output(S) :-
    get_output_stream(S),
    write(S, '.'), nl(S),
    flush_output(S).

:- dynamic wt_recmax_counter/0.

wt_recmax(String, MaxResults, MaxWords) :-
    general_tokenize_sentence_string(String, Words, _IgnoredDesc),
    length(Words, NumWords),
    (   NumWords > MaxWords
    ->  throw(webtrale_rec_error('word count limit exceeded', NumWords, limit:MaxWords))
    ;   true
    ),
    retractall(wt_recmax_counter),
    Desc=bot,
    wt_recmax(Words=Desc, MaxResults),
    terminate_output.

wt_recmax(Words=Desc, MaxResults) :-
    rec(Words, FS, Desc, Residue, Index),
    portray_cat(Words, Desc, FS, Residue, Index),
    assert(wt_recmax_counter),
    findall(_, wt_recmax_counter, Counters),
    length(Counters, NumCounters),
    NumCounters >= MaxResults,
    !.
wt_recmax(_, _).

wt_lexall(String) :-
    general_tokenize_sentence_string(String, Words, _),
    (   Words = [Word], nonvar(Word)        % exactly one word, nonvar
    ->  true
    ;   throw(webtrale_lex_error(String, Words))
    ),
    (   lex(Word, FS),
        portray_lex(Word, FS, _),
        fail
    ;   true
    ),
    terminate_output.

wt_lexicon :-
    get_output_stream(S),
    (   --->(Word, _FS),
        write(S, Word), nl(S),
        fail
    ;   true
    ),
    terminate_output.

init_tokenization :-
    current_predicate(general_tokenize_sentence_string/3) -> true
    ; ensure_loaded(tokenization).
