% ------------------------------------------------------------------------------
%
% tralesld_hooks4.pl
%
% The Prolog side of the TRALE/Kahina bridge.
%
% ------------------------------------------------------------------------------

:- use_module(library(atts)).
:- use_module(library(avl)).
:- use_module(library(file_systems),[current_directory/2]).
:- use_module(library(jasper)).
:- use_module(library(codesio)).
:- use_module(library(lists),[nth1/3]).
:- use_module(library(system)).

:- attribute fs_of_mother/1,
             fs_of_lex/1,
             varnum/1,
             vars/1,
             varscopestack/1,
             ra/4,
             dtrs/2,
             goalstack/1.

% Load ghooks:
:- ensure_loaded(trale_home(ghooks4)).

% ------------------------------------------------------------------------------
% DEBUGGER INTERFACE
% Catch debugger hooks to give commands and disable console output.
% ------------------------------------------------------------------------------

get_reply_hook(Reply) :-
  await_gui_guidance(Reply).

dif_nl_hook.

dif_flush_hook.

dif_format_hook(_,_).

dif_write_hook(_).

dif_write_command_hook(_).

dquery_proceed_hook :-
  fail.

% ------------------------------------------------------------------------------
% HOOK IMPLEMENTATIONS
% These are implementations of the step/port announcement hooks in TRALE's
% debugger. They catch unwanted failure/nondeterminism and delegate to
% appropriate tralesld_* predicates.
% ------------------------------------------------------------------------------

:- multifile clear_hook/0.

clear_hook :-
    retractall(solutions_found(_)),
    asserta(solutions_found(0)),
    retractall(ra(_,_,_)),
    retractall(ra_retrieved(_,_)),
    retractall(ra_retrieved_step(_,_,_)),
    retractall(ra_position_index(_,_,_)),
    retractall(new_prospective_edge_needed(_)),
    start_new_kahina_session.

announce_warning_hook(Warning) :-
    r2ed(tralesld_warning(Warning)).

announce_solution_found_hook(Words,Solution,Residue,Index) :-
    r2ed(tralesld_solution_found(Words,Solution,Residue,Index)).

announce_step_hook(StepID,Command,Line,Goal) :-
    r2ed(tralesld_step(StepID,Command,Line,Goal)).

announce_call_hook(StepID,Command,Line,Goal) :-
    r2ed(tralesld_call(StepID,Command,Line,Goal)).

announce_fail_hook(StepID,Command,Line,Goal) :-
    r2ed(tralesld_fail(StepID,Command,Line,Goal)).

announce_finished_hook(StepID,Command,Line,Goal) :-
    r2ed(tralesld_finished(StepID,Command,Line,Goal)).

announce_exit_hook(StepID,Command,Line,Goal,DetFlag) :-
    r2ed(tralesld_exit(StepID,Command,Line,Goal,DetFlag)).

announce_redo_hook(StepID,Command,Line,Goal) :-
    r2ed(tralesld_redo(StepID,Command,Line,Goal)).

announce_exception_hook(StepID,Command,Line,Goal,Exception) :-
    r2ed(tralesld_exception(StepID,Command,Line,Goal,Exception)).
  
announce_edge_added_hook(Number,Left,Right,RuleName) :-
    r2ed(tralesld_edge_added(Number,Left,Right,RuleName)).

announce_edge_retrieved_hook(Number,Left,Right,RuleName) :-
    r2ed(tralesld_edge_retrieved(Number,Left,Right,RuleName)).

% require a goal to exit deterministically, otherwise raise exception
r2ed(Goal) :-
    (call(Goal)
    -> ! % TODO this forces rather than checks for determinism
     ; raise_exception(tralesld_hook_failed(Goal))).

% ------------------------------------------------------------------------------
% TRACKING THE PARSING PROCESS
% These predicates keep track of what is going on in the parsing process and
% communicate the necessary information to the GUI.
% ------------------------------------------------------------------------------

:- dynamic solutions_found/1.

tralesld_grammar(AbsName) :-
    get_jvm(JVM),
    get_instance(Instance),
    write_to_codes(AbsName,AbsNameChars),
    call_foreign_meta(JVM,register_grammar(Instance,AbsNameChars)).

tralesld_solution_found(Words,Solution,Residue,Index) :-
    send_solution_to_gui(Words,Solution,Residue,Index),
    jvm_store(JVM),
    get_bridge(Bridge),
    call_foreign_meta(JVM,select(Bridge,0)).

% Called before a new step first appears on the stack to transmit information
% about this step to the GUI. The purpose is to keep the stack lean, with just
% step IDs and no further information about the steps on it.
tralesld_step(StepID,rule(RuleName),Line,d_add_dtrs(_,_,_,_,LeftmostDaughterIndex,_,_,_,_,_,_,_)) :-
    !,
    jvm_store(JVM),
    get_bridge(Bridge),
    write_to_codes(RuleName,RuleNameChars),
    with_output_to_codes(write_command(rule(RuleName)),MessageChars),
    call_foreign_meta(JVM,register_rule_application(Bridge,StepID,RuleNameChars,LeftmostDaughterIndex,MessageChars)),    
    ((nonvar(Line),
      Line = [AbsolutePath|LineNumber],
      integer(LineNumber))
     -> (write_to_codes(AbsolutePath,AbsolutePathChars),
         call_foreign_meta(JVM,register_step_source_code_location(Bridge,StepID,AbsolutePathChars,LineNumber)))
      ; true).
tralesld_step(StepID,Command,Line,_Goal) :-
    tralesld_root_step(Command,StepID),
    jvm_store(JVM),
    get_bridge(Bridge),
    command_nodelabel(Command,NodeLabel),
    functor(Command,Functor,_),
    write_to_codes(Functor,FunctorChars),
    write_to_codes(NodeLabel,NodeLabelChars),
    with_output_to_codes(write_command(Command),MessageChars),
    call_foreign_meta(JVM,register_step_information(Bridge,StepID,FunctorChars,NodeLabelChars,MessageChars)),
    ((nonvar(Line),
      Line = [AbsolutePath|LineNumber],
      integer(LineNumber))
     -> (write_to_codes(AbsolutePath,AbsolutePathChars),
         call_foreign_meta(JVM,register_step_source_code_location(Bridge,StepID,AbsolutePathChars,LineNumber)))
      ; true).

% Called when a step is first called. Stack already contains the ID of this
% step.
tralesld_call(StepID,Command,Line,Goal) :-
    tralesld_ra_enter(StepID,Command,Line,Goal),
    tralesld_ra_call(StepID,Command,Line,Goal),
    tralesld_lex_enter(StepID,Command,Line,Goal),
    uniftrace_call(Command),
    varscope_call(Command,StepID),
    goaltrace_call(Command,Goal),
    varwatch_call(Command),
    send_fss_to_gui(StepID,call,Command),
    jvm_store(JVM),
    get_bridge(Bridge),
    call_foreign_meta(JVM,register_step_location(Bridge,StepID)).

% Called when a step fails.
tralesld_fail(StepID,Command,Line,Goal) :-
    tralesld_ra_fail(StepID,Command,Line,Goal),
    tralesld_ra_leave(StepID,Command,Line,Goal),
    jvm_store(JVM),
    get_bridge(Bridge),
    call_foreign_meta(JVM,register_step_failure(Bridge,StepID)).

% Called when a failure-driven step completes.
tralesld_finished(StepID,Command,Line,Goal) :-
    tralesld_ra_leave(StepID,Command,Line,Goal),
    jvm_store(JVM),
    get_bridge(Bridge),
    call_foreign_meta(JVM,register_step_finished(Bridge,StepID)).

% Called when the exception port of a step is reached.
tralesld_exception(StepID,Command,Line,Goal,Exception) :-
    tralesld_ra_fail(StepID,Command,Line,Goal),
    tralesld_ra_leave(StepID,Command,Line,Goal),
    exception_message(Exception,MessageChars),
    jvm_store(JVM),
    get_bridge(Bridge),
    call_foreign_meta(JVM,register_step_exception(Bridge,StepID,MessageChars)).

% Called when a step completes successfully. Stack  still contains the step.
tralesld_exit(StepID,Command,Line,Goal,DetFlag) :-
    tralesld_ra_leave(StepID,Command,Line,Goal),
    varscope_exit(Command),
    goaltrace_exit(Command,Goal),
    send_fss_to_gui(StepID,exit,Command),
    jvm_store(JVM),
    get_bridge(Bridge),
    call_foreign_meta(JVM,register_step_exit(Bridge,StepID,DetFlag)).

% Called when a previously successful step is redone.
tralesld_redo(StepID,Command,Line,Goal) :-
    tralesld_ra_enter(StepID,Command,Line,Goal),
    tralesld_ra_redo(StepID,Command,Line,Goal),
    tralesld_lex_enter(StepID,Command,Line,Goal),
    jvm_store(JVM),
    get_bridge(Bridge),
    call_foreign_meta(JVM,register_step_redo(Bridge,StepID)).

% Called when an edge is added to the chart (happens as a side effect during
% application of a rule).
tralesld_edge_added(Number,Left,Right,RuleName) :-
    jvm_store(JVM),
    get_bridge(Bridge),
    write_to_codes(RuleName,RuleNameChars),
    call_foreign_meta(JVM,register_chart_edge(Bridge,Number,Left,Right,RuleNameChars)),
    edge_by_index(Number,_,_,_,_,Dtrs,_),
    register_edge_dependencies(Number,Dtrs).

tralesld_edge_retrieved(Number,Left,Right,RuleName) :-
    tralesld_ra_edge_retrieved(Number,Left,Right,RuleName).

tralesld_warning(Warning) :-
    warning_message(Warning,MessageChars),
    jvm_store(JVM),
    get_bridge(Bridge),
    call_foreign_meta(JVM,warning(Bridge,MessageChars)).

% ------------------------------------------------------------------------------
% TRACKING ROOTS
% ------------------------------------------------------------------------------

:- dynamic root_id/1.

tralesld_root_step(rec(_),RootID) :-
    !,
    retractall(root_id(_)),
    asserta(root_id(RootID)).
tralesld_root_step(_,_).

% ------------------------------------------------------------------------------
% TRACKING RULE APPLICATIONS
% Manages four dynamic predicates that store information about the current rule
% application(s). All of these four predicates have RAID, the StepID of the
% associated RA (rule application), as their first argument.
% ------------------------------------------------------------------------------

:- dynamic ra/3.                % current rule application (stacked assertions)
:- dynamic ra_retrieved/2.      % how many edges (2nd arg) for this RA have already been retrieved
:- dynamic ra_retrieved_step/3. % how many edges (3rd arg) had been retrieved before a certain step (ID in 2nd arg)
:- dynamic ra_position_index/3. % stores N (2nd arg) and the edge index of the N-th daughter (3rd arg)
:- dynamic new_prospective_edge_needed/1.

tralesld_ra_enter(RAID,rule(RuleName),_,d_add_dtrs(_,_,_,_,LeftmostDaughterIndex,_,_,PrevDtrs,_,_,DtrStore,_)) :-
    !,
    asserta(ra(RAID,RuleName,-1)), % TODO third arg no longer needed. Actually, ra/3 no longer needed. Use attributes instead.
    %write('entering rule application '),write(RAID),nl,
    asserta(ra_retrieved(RAID,1)),
    asserta(ra_position_index(RAID,1,LeftmostDaughterIndex)),
    get_tralesld_atts(Atts),
    put_atts(Atts,fs_of_mother(_)), % delete old mother
    put_atts(Atts,dtrs(PrevDtrs,DtrStore)),
    put_atts(Atts,ra(RAID,RuleName,PrevDtrs,LeftmostDaughterIndex)).
tralesld_ra_enter(_,_,_,_) :-
    ra(RAID,_,_),
    !,
    register_prospective_edge_if_needed(RAID).
tralesld_ra_enter(_,_,_,_).

tralesld_ra_leave(RAID,_,_,_) :-
    retract(ra(RAID,_,_)),
    !,
    %write('leaving rule application '),write(RAID),nl,
    retractall(ra_retrieved(RAID,_)),
    retractall(ra_retrieved_step(RAID,_,_)),
    retractall(ra_position_index(RAID,_,_)).
tralesld_ra_leave(_,_,_,_).

tralesld_ra_call(StepID,_,_,_) :-
    ra(RAID,_,_),
    !,
    % store number of daughter edges that have already been retrieved at this step:
    \+ \+ ( ra_retrieved(RAID,EdgeCount),%write(EdgeCount),write(' edges retrieved for rule application '),write(RAID),write(' at step '),write(StepID),nl,
            asserta(ra_retrieved_step(RAID,StepID,EdgeCount)) ).
tralesld_ra_call(_,_,_,_).

tralesld_ra_fail(_,_,_,_) :-
    ra(RAID,_,_),
    !,
    (\+ new_prospective_edge_needed(RAID)
    -> asserta(new_prospective_edge_needed(RAID))
     ; true).
tralesld_ra_fail(_,_,_,_).

tralesld_ra_redo(StepID,_,_,_) :-
    ra(RAID,_,_),
    !,
    % look up number of daughter edges that had already been retrieved at this step:
    \+ \+ ( %write('edges retrieved for rule application '),write(RAID),write(' at step '),write(StepID),write(': '),
            ra_retrieved_step(RAID,StepID,EdgeCount),%write(EdgeCount),nl,
            retractall(ra_retrieved(RAID,_)),
            asserta(ra_retrieved(RAID,EdgeCount)) ).
tralesld_ra_redo(_,_,_,_).

tralesld_ra_edge_retrieved(EdgeIndex,_,_,_) :-
    ra(RAID,_,_),
    !,
    retract(ra_retrieved(RAID,OldEdgeCount)),
    NewEdgeCount is OldEdgeCount + 1,
    asserta(ra_retrieved(RAID,NewEdgeCount)),
    asserta(ra_position_index(RAID,NewEdgeCount,EdgeIndex)),
    (integer(EdgeIndex)									% TODO retrieval of empty elements currently not registered with GUI
    -> register_prospective_edge_if_needed(RAID),
       register_edge_retrieval(EdgeIndex)
     ; true).
tralesld_ra_edge_retrieved(_,_,_,_).

% ------------------------------------------------------------------------------
% TRACKING LEXICAL LOOKUPS
% Lexical lookups are similar to rule applications but much easier to deal with
% since they don't nest.
% ------------------------------------------------------------------------------

:- dynamic lex_current_word/1.

tralesld_lex_enter(_,lex(Word,_),_,_) :-
    !,
    retractall(lex_current_word(_)),
    asserta(lex_current_word(Word)).
tralesld_lex_enter(_,_,_,_).

% ------------------------------------------------------------------------------
% CONTROL
% ------------------------------------------------------------------------------

await_gui_guidance(Pressed) :-
    wait_until_button_pressed(Pressed),
    !.

wait_until_button_pressed(PressedButton) :-
    repeat,
    pressed_button(PressedButton),
    (PressedButton == 110 % n
    -> (sleep(0.1),
        fail)
     ; true).

pressed_button(Button) :-
    jvm_store(JVM),
    get_bridge(Bridge),
    call_foreign_meta(JVM, get_pressed_button(Bridge, Button)).

% ------------------------------------------------------------------------------
% NODE LABELS
% ------------------------------------------------------------------------------

command_nodelabel(unify(_,XName,_,_),unify(XName)) :-
    !.
command_nodelabel(featval(Loc,Feat,_),featval(Desc:Feat)) :-
    !,
    loc_desc(Loc,Desc).
command_nodelabel(type(Loc,Type,_),type(Desc,Type)) :-
    !,
    loc_desc(Loc,Desc).
command_nodelabel(comp(Functor,Arity),goal(Functor/Arity)) :-
    !.
command_nodelabel(lex(Word,WordStart),lex(Word,WordStart)) :-
    !.
command_nodelabel(match_cats(_),match_cats) :-
    !.
command_nodelabel(cons(Type),cons(Type)) :-
    !.
command_nodelabel(comp_clause(_,_),clause) :-
    !.
command_nodelabel(retrieve,retrieve_edge) :-
    !.
command_nodelabel(now(BlockingStepID),now(BlockingStepID)) :-
    !.
command_nodelabel(rec(Words),rec(Words)) :-
    !.
command_nodelabel(compile_gram(AbsPath),compile_gram(AbsPath)) :-
    !.
command_nodelabel(Command,Label) :-
    functor(Command,Label,_).

loc_desc(empty,'empty cat') :-
    !.
loc_desc(ineq,'inequated desc') :-
    !.
loc_desc(feat(F),F) :-
    !.
loc_desc(lex,'lex entry') :-
    !.
loc_desc(cons(T),Desc) :-
    !,
    atom_concat(T,'-constrained FS',Desc).
loc_desc(lrin,'LR input') :-
    !.
loc_desc(lrout,'LR output') :-
    !.
loc_desc(left,'left arg') :-
    !.
loc_desc(right,'right arg') :-
    !.
loc_desc(arg(N),Desc) :-
    !,
    number_codes(N,NCodes),
    atom_codes(NAtom,NCodes),
    atom_concat('arg',NAtom,Desc).
loc_desc(query_desc,'query desc') :-
    !.
loc_desc(edge,'edge') :-
    !.
loc_desc(dtrlist,'dtr list') :-
    !.
loc_desc(mother,'mother') :-
    !.

% ------------------------------------------------------------------------------
% FS DISPLAY
% sending tree fragments for the current rule application — more specifically
% for the current step — and variable bindings to the GUI
%
% KE: ghooks.pl, and then this section, are in dire need of despaghettifying.
% Also, residues are not yet dealt with properly.
% ------------------------------------------------------------------------------

send_fss_to_gui(StepID,Port,rule_close(FS)) :-
    !,
    port_fskey(Port,FSKey),
    send_fs_to_gui(FS,[],'Edge',StepID,FSKey). % TODO residue?
send_fss_to_gui(StepID,Port,Command) :-
    goalstack_peek(Functor,Arity,OpenEndedArgs),
    !,
    % Build DiffAssoc:
    empty_avl(Empty),
    (command_fs(Command,FS)
    -> avl_store(different(FS),Empty,_,DiffAssoc)
     ; Empty = DiffAssoc),
    % Create re-entrancy list:
    close_list(Arity,OpenEndedArgs,Args),
    Goal =.. [Functor|Args],
    duplicates_list(Args,Empty,DupsMid,Empty,VisMid,0,NumMid),
    fetch_vars(VarTerms,Vars),
    duplicates_list(Vars,DupsMid,Dups,VisMid,_,NumMid,Num),
    % Portray goal:
    port_fskey(Port,FSKey),
    asserta(portraying_fs(StepID,FSKey)),
    tralesld_portray_goal(Goal,DiffAssoc,Dups,Num),
    retractall(portraying_fs(_,_)),
    % Portray vars:
    send_vars_to_gui(VarTerms,StepID,FSKey,DiffAssoc,Dups).
send_fss_to_gui(StepID,Port,Command) :-
    ra(RAID,RuleName,_),
    get_tralesld_atts(HookAtts),
    !,
    \+ \+ ( % Get position of leftmost word covered by active edges:			% TODO replace this double negation by appropriate cuts
            ra_position_index(RAID,1,LeftmostEdgeIndex),
            get_edge_ref(LeftmostEdgeIndex,Left,_,_,_,_),
            % Get position of rightmost word covered by active edges:
            ra_retrieved(RAID,EdgeCount),
            ra_position_index(RAID,EdgeCount,RightmostEdgeIndex),
            get_edge_ref(RightmostEdgeIndex,_,Right,_,_,_),
            % Build label:
            parsing(Words),
            list_from_to(Words,Left,Right,WordsLabel),
            % Build subtrees:
            get_atts(HookAtts,dtrs(PrevDtrs,DtrStore)),
            build_local_subtrees(PrevDtrs,DtrStore,Subtrees),
            % Get mother FS (or leave uninstantiated):
            (get_atts(HookAtts,fs_of_mother(MotherFS))
            -> true
             ; true),
            % Build DiffAssoc:
            empty_avl(Empty),
            (command_fs(Command,FS)
            -> avl_store(different(FS),Empty,_,DiffAssoc)
             ; Empty = DiffAssoc),
            % Create re-entrancy list:
            trees_fss(Subtrees,SubtreesFSs),						% TODO get duplicates on the fly in build_local_subtrees
            duplicates_list([MotherFS|SubtreesFSs],Empty,DupsMid,Empty,VisMid,0,NumMid),
            fetch_vars(HookAtts,VarTerms,Vars),
            duplicates_list(Vars,DupsMid,Dups,VisMid,_,NumMid,Num),
            % Portray tree:
            port_fskey(Port,FSKey),
            asserta(portraying_fs(StepID,FSKey)),
            tralesld_portray_tree(WordsLabel,MotherFS,tree(RuleName,WordsLabel,MotherFS,Subtrees),SubtreesFSs,DiffAssoc,Dups,Num),
            retractall(portraying_fs(_,_)),
            % Portray vars:
            send_vars_to_gui(VarTerms,StepID,FSKey,DiffAssoc,Dups)
            ).
send_fss_to_gui(StepID,Port,Command) :-
    get_tralesld_atts(HookAtts),
    get_atts(HookAtts,fs_of_lex(LexFS)),
    !,
    lex_current_word(Word),
    empty_avl(Empty),
    (command_fs(Command,FS)
    -> avl_store(different(FS),Empty,_,DiffAssoc)
     ; Empty = DiffAssoc),
    % Create re-entrancy list:
    fetch_vars(HookAtts,VarTerms,Vars),
    duplicates_list([LexFS|Vars],Empty,Dups,Empty,_,0,Num),
    % Portray entry:
    port_fskey(Port,FSKey),
    asserta(portraying_fs(StepID,FSKey)),
    tralesld_portray_tree([Word],LexFS,tree(lex,[Word],LexFS,[]),[],DiffAssoc,Dups,Num),
    retractall(portraying_fs(_,_)),
    % Portray vars:
    send_vars_to_gui(VarTerms,StepID,FSKey,DiffAssoc,Dups).
send_fss_to_gui(_,_,_).

port_fskey(call,start).
port_fskey(exit,end).
port_fskey(redo,start).

build_local_subtrees([PrevDtr|PrevDtrs],[FS|DtrStore],[tree(RuleName,Covered,FS,[])|Subtrees]) :-
    nonvar(PrevDtr),
    !,
    get_edge_ref(PrevDtr,Left,Right,_OrigFS,_DtrNums,RuleName),
    parsing(Words),
    list_from_to(Words,Left,Right,Covered),
    build_local_subtrees(PrevDtrs,DtrStore,Subtrees).
build_local_subtrees(_,_,[]).

tralesld_portray_tree(Words,FS,Tree,SubtreesFSs,DiffAssoc,DupsIn,NumIn) :- % TODO call filter_iqs/3?
    grale_flag,
    list_to_double_quoted_string(Words,DQWords),
    empty_avl(Empty),
    append("!newdata",DQWords,GraleCommandPrefix),
    grale_write_chars(GraleCommandPrefix),
    \+ \+ (insert_duplicates_list(SubtreesFSs,DupsIn,Dups,NumIn,Num),
           avl_store(top_index,DiffAssoc,Num,HDMid),
           avl_store(tree_struc,HDMid,Tree,HD),
           pp_fs(FS,0,Dups,_,Empty,_,0,HD,_)),
    grale_nl,
    grale_flush_output.

tralesld_portray_goal(Goal,DiffAssoc,DupsIn,NumIn) :-
    grale_flag,
    functor(Goal,Name,Arity),
    write_to_codes(Name,NameChars),
    write_to_codes(Arity,ArityChars),
    grale_write_chars("!newdata"),
    grale_write_chars("\""),
    grale_write_chars(NameChars),
    grale_write_chars("/"),
    grale_write_chars(ArityChars),
    grale_write_chars("\""),
    avl_store(id_index,DiffAssoc,NumIn,HDIn),
    empty_avl(Empty),
    \+ \+ (grale_portray_goal(Goal,DupsIn,DupsMid,Empty,VisMid,HDIn,HDMid),
           portray_reentrancies(DupsMid,_,VisMid,_,0,HDMid,_)),
    grale_nl,
    grale_flush_output.

trees_fss([],[]).
trees_fss([Tree|Trees],FSs) :-
  tree_fss(Tree,TreeFSs),
  trees_fss(Trees,TreesFSs),
  append(TreeFSs,TreesFSs,FSs).								% TODO use difference lists

tree_fss(tree(_,_,FS,Children),[FS|ChildrenFSs]) :-
  trees_fss(Children,ChildrenFSs).

send_solution_to_gui(Words,Solution,Residue,_Index) :-
  join_words(Words,Joint),
  send_fs_to_gui(Solution,Residue,Joint,0,end).

send_vars_to_gui([],_,_,_,_). % TODO call filter_iqs/3?
send_vars_to_gui([var(VarName,Var)|VarTerms],StepID,Key,DiffAssoc,Dups) :-
    get_type(Var,Type),
    asserta(portraying_variable_binding(StepID,Key,VarName,Type)),
    list_to_double_quoted_string([VarName],DQName),
    append("!newdata",DQName,GraleCommandPrefix),
    grale_write_chars(GraleCommandPrefix),
    empty_avl(Empty),
    \+ \+ pp_fs(Var,0,Dups,_,Empty,_,0,DiffAssoc,_),
    grale_nl,
    grale_flush_output,
    retractall(portraying_variable_binding(_,_,_,_)),
    send_vars_to_gui(VarTerms,StepID,Key,DiffAssoc,Dups).

send_fs_to_gui(FS,Residue,Title,StepID,Key) :-
    filter_iqs(Residue,_Iqs,_FSResidue),
    asserta(portraying_fs(StepID,Key)),
    list_to_double_quoted_string([Title],DQName),
    append("!newdata",DQName,GraleCommandPrefix),
    grale_write_chars(GraleCommandPrefix),
    empty_avl(Empty),
    \+ \+ ( duplicates_list([FS],Empty,Dups,Empty,_,0,_),
            pp_fs(FS,0,Dups,_,Empty,_,0,Empty,_) ), % TODO residues
    grale_nl,
    grale_flush_output,
    retractall(portraying_fs(_,_)).

fetch_vars(VarTerms,Vars) :-
    get_tralesld_atts(Atts),
    fetch_vars(Atts,VarTerms,Vars).

fetch_vars(HookAtts,VarTerms,Vars) :-
    varscopestack_peek(HookAtts,ScopeID),
    get_atts(HookAtts,vars(RawVarTerms)),
    !,
    fetch_vars_from_open_ended_list(ScopeID,RawVarTerms,VarTerms,Vars).
fetch_vars(_,[],[]).

fetch_vars_from_open_ended_list(_,X,[],[]) :-
    var(X),
    !.
fetch_vars_from_open_ended_list(ScopeID,[var(ScopeID,VarName,Var)|RawVarTerms],[var(VarName,Var)|VarTerms],[Var|Vars]) :-
    !,
    fetch_vars_from_open_ended_list(ScopeID,RawVarTerms,VarTerms,Vars).
fetch_vars_from_open_ended_list(ScopeID,[_|RawVarTerms],VarTerms,Vars) :-
    fetch_vars_from_open_ended_list(ScopeID,RawVarTerms,VarTerms,Vars).

% ------------------------------------------------------------------------------
% FS DISPLAY - CALLBACK
% These hooks catch ghooks.pl's attempts to write to a stream and instead
% collect the respective bits and pieces of Grisu strings in asserted
% grale_message_chunk/1 clauses.
% ------------------------------------------------------------------------------

:- dynamic portraying_fs/2.
:- dynamic portraying_variable_binding/4.
:- dynamic grale_message_chunk/1.

grale_write_chars_hook(Chars) :-
    portraying_variable_binding(_,_,_,_),
    !,
    asserta(grale_message_chunk(Chars)).
grale_write_chars_hook(Chars) :-
    portraying_fs(_,_),
    !,
    asserta(grale_message_chunk(Chars)).
    
grale_nl_hook :-
    portraying_variable_binding(StepID,Key,VarName,FSType),
    !,
    grale_write_chars_hook([10]),
    r2ed(tralesld_grale_message_end(StepID,Key,VarName,FSType)).
grale_nl_hook :-
    portraying_fs(StepID,Key),
    !,
    grale_write_chars_hook([10]),
    r2ed(tralesld_grale_message_end(StepID,Key)).

grale_flush_output_hook.

collect_grale_message(MessageChars) :-
    collect_grale_message([],MessageChars).

collect_grale_message(Rest,MessageChars) :-
    retract(grale_message_chunk(Chars)),
    !,
    append(Chars,Rest,NewRest),
    collect_grale_message(NewRest,MessageChars).
collect_grale_message(MessageChars,MessageChars).

% NEW VERSION WITH ALTERNATIVE OUTPUT TO FILE
tralesld_grale_message_end(StepID,Key) :-
    collect_grale_message(MessageChars),
    ((
        fs_to_textfile(TempFile), !,
        open(TempFile,write,OutputStream),
        put_string(MessageChars,OutputStream),
        close(OutputStream)
    );
    (
        write_to_codes(Key,KeyChars),
        jvm_store(JVM),
        get_bridge(Bridge),
        call_foreign_meta(JVM,register_message(Bridge,StepID,KeyChars,MessageChars))
    )).

put_string([],_).
put_string([Char|Chars],OutputStream) :-
    put(OutputStream,Char),
    put_string(Chars,OutputStream).

tralesld_grale_message_end(StepID,Key,VarName,FSType) :-
    jvm_store(JVM),
    get_bridge(Bridge),
    collect_grale_message(MessageChars),
    write_to_codes(Key,KeyChars),
    write_to_codes(VarName,VarNameChars),
    write_to_codes(FSType,FSTypeChars),
    call_foreign_meta(JVM,register_message(Bridge,StepID,KeyChars,VarNameChars,FSTypeChars,MessageChars)).

% ------------------------------------------------------------------------------
% GRAPHICAL TRACING OF UNIFICATION
% ------------------------------------------------------------------------------

uniftrace_call(Command) :-
    command_key_fs(Command,Key,FS),
    !,
    get_tralesld_atts(HookAtts),
    uniftrace_put(Key,FS,HookAtts).
uniftrace_call(_).

uniftrace_put(fs_of_mother,FS,HookAtts) :-
    put_atts(HookAtts,fs_of_mother(FS)).
uniftrace_put(fs_of_lex,FS,HookAtts) :-
    put_atts(HookAtts,fs_of_lex(FS)).

command_key_fs(type(mother,_,FS),fs_of_mother,FS) :-
    !.
command_key_fs(featval(mother,_,FS),fs_of_mother,FS) :-
    !.
command_key_fs(type(lex,_,FS),fs_of_lex,FS) :-
    !.
command_key_fs(featval(lex,_,FS),fs_of_lex,FS) :-
    !.

% ------------------------------------------------------------------------------
% VARIABLE WATCHING
% ------------------------------------------------------------------------------

varscope_call(rule(_),StepID) :-
    !,
    get_tralesld_atts(HookAtts),
    varscopestack_push(StepID,HookAtts).
varscope_call(comp_clause(_,_),StepID) :-
    !,
    get_tralesld_atts(HookAtts),
    varscopestack_push(StepID,HookAtts).
varscope_call(fun_clause(_,_),StepID) :-
    !,
    get_tralesld_atts(HookAtts),
    varscopestack_push(StepID,HookAtts).
varscope_call(cons(_),StepID) :-
    !,
    get_tralesld_atts(HookAtts),
    varscopestack_push(StepID,HookAtts).
varscope_call(lex(_,_),StepID) :-
    !,
    get_tralesld_atts(HookAtts),
    varscopestack_push(StepID,HookAtts).
varscope_call(lr(_,_,_),StepID) :-
    !,
    get_tralesld_atts(HookAtts),
    varscopestack_push(StepID,HookAtts).
varscope_call(empty,StepID) :-
    !,
    get_tralesld_atts(HookAtts),
    varscopestack_push(StepID,HookAtts).
varscope_call(apply_once(_RuleName,_Empty,_DtrNum),StepID) :-
    !,
    get_tralesld_atts(HookAtts),
    varscopestack_push(StepID,HookAtts).
varscope_call(_,_).

varscope_exit(rule(_)) :-
    !,
    get_tralesld_atts(HookAtts),
    varscopestack_pop(HookAtts).
varscope_exit(comp_clause(_,_)) :-
    !,
    get_tralesld_atts(HookAtts),
    varscopestack_pop(HookAtts).
varscope_exit(fun_clause(_,_)) :-
    !,
    get_tralesld_atts(HookAtts),
    varscopestack_pop(HookAtts).
varscope_exit(cons(_)) :-
    !,
    get_tralesld_atts(HookAtts),
    varscopestack_pop(HookAtts).
varscope_exit(lex(_,_)) :-
    !,
    get_tralesld_atts(HookAtts),
    varscopestack_pop(HookAtts).
varscope_exit(lr(_,_,_)) :-
    !,
    get_tralesld_atts(HookAtts),
    varscopestack_pop(HookAtts).
varscope_exit(_).

varscopestack_push(ScopeID,HookAtts) :-
    (get_atts(HookAtts,varscopestack(Stack))
    -> true
     ; Stack = []),
    put_atts(HookAtts,varscopestack([ScopeID|Stack])).

varscopestack_pop(HookAtts) :-
    get_atts(HookAtts,varscopestack([_|Stack])),
    put_atts(HookAtts,varscopestack(Stack)).

varscopestack_peek(HookAtts,ScopeID) :-
    get_atts(HookAtts,varscopestack([ScopeID|_])).

varwatch_call(unify(_,VarName,_,Var)) :-
    !,
    get_tralesld_atts(HookAtts),
    (get_atts(HookAtts,varnum(VarNum))
    -> NewVarNum is VarNum + 1,
       put_atts(HookAtts,varnum(NewVarNum)),
       get_atts(HookAtts,vars(Vars))
     ; put_atts(HookAtts,varnum(1)),
       NewVarNum = 1,
       put_atts(HookAtts,vars(Vars))),
    varscopestack_peek(HookAtts,ScopeID),
    nth1(NewVarNum,Vars,var(ScopeID,VarName,Var)).
varwatch_call(_).

% ------------------------------------------------------------------------------
% TRACING GOALS
% ------------------------------------------------------------------------------

goaltrace_call(comp(_,_),d_query_goal0_lcomp(_,_,OpenEndedArgs,_,_,Functor,Arity)) :-
    !,
    goalstack_push(Functor,Arity,OpenEndedArgs).
goaltrace_call(_,_).

goaltrace_exit(comp(_,_),_) :-
    !,
    goalstack_pop.
goaltrace_exit(_,_).

goalstack_push(Functor,Arity,OpenEndedArgs) :-
    get_tralesld_atts(Atts),
    (get_atts(Atts,goalstack(Stack))
    -> true
     ; Stack = []),
    put_atts(Atts,goalstack([goal(Functor,Arity,OpenEndedArgs)|Stack])).

goalstack_pop :-
    get_tralesld_atts(Atts),
    get_atts(Atts,goalstack([_|Stack])),
    put_atts(Atts,goalstack(Stack)).

goalstack_peek(Functor,Arity,OpenEndedArgs) :-
    get_tralesld_atts(Atts),
    get_atts(Atts,goalstack([goal(Functor,Arity,OpenEndedArgs)|_])).

% ------------------------------------------------------------------------------
% SIGNATURE READOUT (report subtype relations etc. to bridge; crude first version)
% ------------------------------------------------------------------------------

send_signature :-
  get_jvm(JVM),
  get_bridge(Bridge),
  send_type_info(JVM,Bridge,bot).

% do not try to traverse the atoms in the type hierarchy
send_type_info(_,_,a_ _) :-
  !.
send_type_info(JVM,Bridge,Type) :-
  approp_feats(Type,Fs),
  send_approp_feats(JVM,Bridge,Type,Fs),
  immed_subtypes(Type,Subtypes),
  send_subtypelist_info(JVM,Bridge,Type,Subtypes),
  send_typelist_info(JVM,Bridge,Subtypes).

send_typelist_info(_,_,[]).
send_typelist_info(JVM,Bridge,[Type|Rest]) :-
  send_type_info(JVM,Bridge,Type),
  send_typelist_info(JVM,Bridge,Rest).

send_approp_feats(_,_,_,[]).
send_approp_feats(JVM,Bridge,Type,[F|Rest]) :-
  approp(F,Type,ValRestr),
  write_to_codes(F,FChars),
  write_to_codes(Type,TypeChars),
  write_to_codes(ValRestr,ValRestrChars),
  call_foreign_meta(JVM,register_appropriate_feature(Bridge,TypeChars,FChars,ValRestrChars)),
  send_approp_feats(JVM,Bridge,Type,Rest).

send_subtypelist_info(_,_,_,[]).
send_subtypelist_info(JVM,Bridge,Type,[Subtype|Rest]) :-
  write_to_codes(Type,TypeChars),
  write_to_codes(Subtype,SubtypeChars),
  call_foreign_meta(JVM,register_subtype_relation(Bridge,TypeChars,SubtypeChars)),
  send_subtypelist_info(JVM,Bridge,Type,Rest).

% ------------------------------------------------------------------------------
%  LEX and MGS versions that store results as GRISU string in a temporary file
% ------------------------------------------------------------------------------

:- dynamic fs_to_textfile/1.

lex_to_tempfile(Word,TempFile):-
  \+ \+ on_exception(ale(Exception),
	             (asserta(fs_to_textfile(TempFile)),
                      current_predicate(lex,lex(_,_))
                     -> call_residue((if(lex(Word,FS),
   		                      true,
             		              raise_exception(ale(unk_words([Word]))))),Residue),
                        ((current_predicate(portray_lex,portray_lex(_,_,_)),
                          portray_lex(Word,FS,Residue)) -> true
                        ; nl, write('WORD: '), write(Word), nl,
                          send_fs_to_gui(FS,Residue,'current',-2,start)
                        ),
                        retract(fs_to_textfile(TempFile))
	               ; raise_exception(ale(no_lex))),
	             alex(Exception)).


mgs_to_tempfile(Desc,TempFile):-
   asserta(fs_to_textfile(TempFile)),
   %write('Call: infer_mgsat_type('), write(Desc), write(',bot,MGType)'),nl,
   infer_mgsat_type(Desc,bot,MGType),
   %write('Exit: infer_mgsat_type('), write(Desc), write(',bot,'), write(MGType), write(')'),nl,
   \+ \+ (call_residue(add_to(Desc,FS),Residue),write(FS),nl,
	  ((current_predicate(portray_mgsat,portray_mgsat(_,_,_,_)),
	    portray_mgsat(Desc,FS,Residue,MGType)) -> true
	  ; nl, write('MOST GENERAL SATISFIER OF: '), write(Desc), nl,
	    send_fs_to_gui(FS,Residue,'current',-2,start)
	  ),
          retract(fs_to_textfile(TempFile))).

% ------------------------------------------------------------------------------
% HELPER PREDICATES
% ------------------------------------------------------------------------------

atoms_concat([],'').
atoms_concat([Head|Tail],Atom) :-
    atoms_concat(Tail,TailAtom),
    atom_concat(Head,TailAtom,Atom).

register_edge_dependencies(_,[]) :-
    !.
register_edge_dependencies(Mother,[Daughter|Daughters]) :-
    (integer(Daughter)
    -> (jvm_store(JVM),
        get_bridge(Bridge),
        call_foreign_meta(JVM,register_edge_dependency(Bridge,Mother,Daughter)))
     ; true),
    register_edge_dependencies(Mother,Daughters).

register_edge_retrievals(Daughters) :-
    var(Daughters),
    !.
register_edge_retrievals([empty(_,_)|Daughters]) :-
    !,
    register_edge_retrievals(Daughters).
register_edge_retrievals([Daughter|Daughters]) :-
    !,
    register_edge_retrieval(Daughter),
    register_edge_retrievals(Daughters).
register_edge_retrievals(_).

register_edge_retrieval(EdgeIndex) :-
    jvm_store(JVM),
    get_bridge(Bridge),
    call_foreign_meta(JVM,register_edge_retrieval(Bridge,EdgeIndex)).

register_prospective_edge_if_needed(RAID) :-
    retract(new_prospective_edge_needed(RAID)),
    !,
    register_prospective_edge.
register_prospective_edge_if_needed(_).

register_prospective_edge :-
    get_tralesld_atts(Atts),
    get_atts(Atts,ra(RAID,RuleName,PrevDtrs,LeftmostDaughterIndex)),
    jvm_store(JVM),
    get_bridge(Bridge),
    write_to_codes(RuleName,RuleNameChars),
    call_foreign_meta(JVM,register_prospective_edge(Bridge,RAID,RuleNameChars,LeftmostDaughterIndex)),
    register_edge_retrievals(PrevDtrs).

list_from_to([],_,_,[]) :-
    !.
list_from_to(List,0,To,SubList) :-
    !,
    list_to(List,To,SubList).
list_from_to([_|Rest],From,To,SubList) :-
    NewFrom is From - 1,
    NewTo is To - 1,
    list_from_to(Rest,NewFrom,NewTo,SubList).

list_to(_,0,[]) :-
    !.
list_to([First|Rest],To,SubList) :-
    NewTo is To -1,
    list_to(Rest,NewTo,NewRest),
    SubList = [First|NewRest].

command_fs(unify(_,_,FS,_),FS) :-
    !.
command_fs(patheq(_,_,_,FS),FS) :-
    !.
command_fs(ineq_add(_,FS),FS) :-
    !.
command_fs(featval(_,_,FS),FS) :-
    !.
command_fs(macro(_,FS),FS) :-
    !.
command_fs(type(_,_,FS),FS) :-
    !.
command_fs(atom(_,_,FS),FS) :-
    !.
command_fs(fun(_,_,FS),FS) :-
    !.
command_fs(lr(_,_,FS),FS) :-
    !.
command_fs(match_cats(FS),FS) :-
    !.

get_tralesld_atts(Atts) :-
    % HACK This is SICStus Prolog's global variable facility, so to speak.
    % The SWI port will be able to use a cleaner solution.
    execution_state(private(P)),
    memberchk(tralesld_atts(Atts),P).

close_list(0,_,[]) :-
    !.
close_list(Length,[Element|OpenEndedList],[Element|List]) :-
    NewLength is Length - 1,
    close_list(NewLength,OpenEndedList,List).

split_at_first([],_,[],[]).
split_at_first([First|Rest],Target,[],Rest) :-
    First == Target,
    !.
split_at_first([First|Rest],Target,[First|BeforeRest],After) :-
    split_at_first(Rest,Target,BeforeRest,After).

join_words([],'').
join_words([Atom],Atom) :-
  !.
join_words([Atom|Rest],Joint) :-
  atom_concat(Atom,' ',Prefix),
  join_words(Rest,Suffix),
  atom_concat(Prefix,Suffix,Joint).

my_cd(Path) :-
  current_directory(_,Path).

% ------------------------------------------------------------------------------
% INVERSION OF CONTROL
% ------------------------------------------------------------------------------

dgo :-
    start_new_kahina_session,
    jvm_store(JVM),
    get_instance(Instance),
    repeat,
    call_foreign_meta(JVM,get_command(Instance,CommandLine)),
    wait_and_fail_if_no_command(CommandLine),
    split_at_first(CommandLine,32,Command,Argument), % space
    atom_codes(CommandAtom,Command),
    execute_command(CommandAtom,Argument).

wait_and_fail_if_no_command([]) :-
    !,
    sleep(0.1),
    fail.
wait_and_fail_if_no_command(_).

execute_command(query,Query) :-
    read_from_codes(Query,QueryTerm),
    toggle_abort_by_exception,
    \+ \+ catch(call(QueryTerm),ale_abort,true),
    fail.
execute_command(quit,_).

toggle_abort_by_exception :-
    retract(abort_by_exception).
toggle_abort_by_exception :-
    assert(abort_by_exception).
toggle_abort_by_exception :-
    retract(abort_by_exception),
    fail.

% ------------------------------------------------------------------------------
% MESSAGES
% ------------------------------------------------------------------------------

exception_message(Exception,ExceptionChars) :-
  write_to_codes(Exception,ExceptionChars).
% TODO use messages from ale_exception/1.

warning_message(feat_order_before_undef(Before,After),MessageChars) :-
  !,
  prolog_flag(toplevel_print_options,Options,Options),
  format_to_codes('~@ <<< ~@ - ~a not declared in signature',[write_term(Before,Options),write_term(After,Options),Before],MessageChars).
warning_message(feat_order_after_undef(Before,After),MessageChars) :-
  !,
  prolog_flag(toplevel_print_options,Options,Options),
  format_to_codes('~@ <<< ~@ - ~a not declared in signature',[write_term(Before,Options),write_term(After,Options),After],MessageChars).
warning_message(feat_order_cycle(Feat,Cycle),MessageChars) :-
  !,
  prolog_flag(toplevel_print_options,Options,Options),
  format_to_codes('cycle at ~a exists in feature order declarations: ~@',[Feat,write_term(Cycle,Options)],MessageChars).
warning_message(feat_order_sisters_initial(InitFeat1,InitFeat2,AppropType),MessageChars) :-
  !,
  prolog_flag(toplevel_print_options,Options,Options),
  format_to_codes('>>> ~@, >>> ~@ - features are both appropriate to ~a, and therefore cannot both be initial',[write_term(InitFeat1,Options),write_term(InitFeat2,Options),AppropType],MessageChars).
warning_message(feat_order_before_initial(Before,InitFeat),MessageChars) :-
  !,
  prolog_flag(toplevel_print_options,Options,Options),
  format_to_codes('>>> ~@, ~@ <<< ~@ - a feature cannot precede an initial feature',[write_term(InitFeat,Options),write_term(Before,Options),write_term(InitFeat,Options)],MessageChars).
warning_message(Warning,WarningChars) :-
  write_to_codes(Warning,WarningChars).

% ------------------------------------------------------------------------------
% JASPER INTERFACE
% TODO cleanup
% ------------------------------------------------------------------------------

:- dynamic jvm_store/1.
:- dynamic bridge_store/1.
:- dynamic instance_store/1.

% wrapper predicate for easier foreign calls
% TODO make this even easier by retrieving the JVM here, not before the call
% TODO http://kahina.org/trac/ticket/18
call_foreign_meta(JVM, Goal) :-
   functor(Goal, Name, Arity),  % extract predicate name
   functor(ArgDesc, Name, Arity), % build template
   foreign(Method, java, ArgDesc), % look it up
   !,
   catch(jasper_call(JVM, Method, ArgDesc, Goal),
         Excp,
         (is_java_exception(JVM, Excp) ->
          print_exception_info(JVM, Excp);
          throw(Excp))).

% Kahina instance methods
foreign(method('org/kahina/tralesld/TraleSLDInstance','startNewSession',[instance]),java,start_new_session(+object('org/kahina/tralesld/TraleSLDInstance'),[-object('org/kahina/tralesld/bridge/TraleSLDBridge')])).
foreign(method('org/kahina/tralesld/TraleSLDInstance','getCommand',[instance]),java,get_command(+object('org/kahina/tralesld/TraleSLDInstance'),[-chars])).

% bridge methods
foreign(method('org/kahina/tralesld/bridge/TraleSLDBridge','registerChartEdge',[instance]),java,register_chart_edge(+object('org/kahina/tralesld/bridge/TraleSLDBridge'),+integer,+integer,+integer,+chars)).
foreign(method('org/kahina/tralesld/bridge/TraleSLDBridge','registerEdgeDependency',[instance]),java,register_edge_dependency(+object('org/kahina/tralesld/bridge/TraleSLDBridge'),+integer,+integer)).
foreign(method('org/kahina/tralesld/bridge/TraleSLDBridge','registerEdgeRetrieval',[instance]),java,register_edge_retrieval(+object('org/kahina/tralesld/bridge/TraleSLDBridge'),+integer)).
foreign(method('org/kahina/tralesld/bridge/TraleSLDBridge','registerProspectiveEdge',[instance]),java,register_prospective_edge(+object('org/kahina/tralesld/bridge/TraleSLDBridge'),+integer,+chars,+integer)).
foreign(method('org/kahina/tralesld/bridge/TraleSLDBridge','step',[instance]),java,register_step_information(+object('org/kahina/tralesld/bridge/TraleSLDBridge'),+integer,+chars,+chars,+chars)).
foreign(method('org/kahina/tralesld/bridge/TraleSLDBridge','registerRuleApplication',[instance]),java,register_rule_application(+object('org/kahina/tralesld/bridge/TraleSLDBridge'),+integer,+chars,+integer,+chars)).
foreign(method('org/kahina/tralesld/bridge/TraleSLDBridge','registerStepSourceCodeLocation',[instance]),java,register_step_source_code_location(+object('org/kahina/tralesld/bridge/TraleSLDBridge'),+integer,+chars,+integer)).
foreign(method('org/kahina/tralesld/bridge/TraleSLDBridge','call',[instance]),java,register_step_location(+object('org/kahina/tralesld/bridge/TraleSLDBridge'),+integer)).
foreign(method('org/kahina/tralesld/bridge/TraleSLDBridge','fail',[instance]),java,register_step_failure(+object('org/kahina/tralesld/bridge/TraleSLDBridge'),+integer)).
foreign(method('org/kahina/tralesld/bridge/TraleSLDBridge','finished',[instance]),java,register_step_finished(+object('org/kahina/tralesld/bridge/TraleSLDBridge'),+integer)).
foreign(method('org/kahina/tralesld/bridge/TraleSLDBridge','exception',[instance]),java,register_step_exception(+object('org/kahina/tralesld/bridge/TraleSLDBridge'),+integer,+chars)).
foreign(method('org/kahina/tralesld/bridge/TraleSLDBridge','exit',[instance]),java,register_step_exit(+object('org/kahina/tralesld/bridge/TraleSLDBridge'),+integer,+boolean)).
foreign(method('org/kahina/tralesld/bridge/TraleSLDBridge','redo',[instance]),java,register_step_redo(+object('org/kahina/tralesld/bridge/TraleSLDBridge'),+integer)).
foreign(method('org/kahina/tralesld/bridge/TraleSLDBridge','select',[instance]),java,select(+object('org/kahina/tralesld/bridge/TraleSLDBridge'),+integer)).
foreign(method('org/kahina/tralesld/bridge/TraleSLDBridge','registerMessage',[instance]),java,register_message(+object('org/kahina/tralesld/bridge/TraleSLDBridge'),+integer,+chars,+chars)).
foreign(method('org/kahina/tralesld/bridge/TraleSLDBridge','registerMessage',[instance]),java,register_message(+object('org/kahina/tralesld/bridge/TraleSLDBridge'),+integer,+chars,+chars,+chars,+chars)).
foreign(method('org/kahina/tralesld/bridge/TraleSLDBridge','warning',[instance]),java,warning(+object('org/kahina/tralesld/bridge/TraleSLDBridge'),+chars)).
foreign(method('org/kahina/tralesld/bridge/TraleSLDBridge','getAction',[instance]),java,get_pressed_button(+object('org/kahina/tralesld/bridge/TraleSLDBridge'),[-char])).
foreign(method('org/kahina/tralesld/bridge/TraleSLDBridge','registerSubtype',[instance]),java,register_subtype_relation(+object('org/kahina/tralesld/bridge/TraleSLDBridge'),+chars,+chars)).
foreign(method('org/kahina/tralesld/bridge/TraleSLDBridge','registerAppropriateFeature',[instance]),java,register_appropriate_feature(+object('org/kahina/tralesld/bridge/TraleSLDBridge'),+chars,+chars,+chars)).
foreign(method('org/kahina/tralesld/bridge/TraleSLDBridge','signatureFinished',[instance]),java,signature_finished(+object('org/kahina/tralesld/bridge/TraleSLDBridge'))).

:- multifile file_search_path/2.

file_search_path(kahinatrunk,KahinaTrunk) :-
    environ('KAHINA_TRUNK',KahinaTrunk).
file_search_path(gralejdev,Gralejdev) :-
    environ('GRALEJ_TRUNK_DIR',Gralejdev).

tralesld_classpath(Classpath) :-
    kahina_classpath(Classpath,GralejClasspath),
    gralej_classpath(GralejClasspath,[]).

% Set KAHINA_TRUNK to use a development version of Kahina rather than the one included in TRALE.
kahina_classpath([kahinatrunk('bin'),trale_home('trale-sld/lib/jedit-textArea.jar')|Rest],Rest) :-
    environ('KAHINA_TRUNK',TTD),
    \+ TTD = '',
    !.
kahina_classpath([trale_home('kahina/bin'),trale_home('kahina/lib/jedit-textArea.jar')|Rest],Rest).

% Set GRALEJ_TRUNK_DIR to use a development version of Gralej rather than the one included in TRALE.
gralej_classpath([gralejdev('bin'),
        gralejdev('lib/tomato.jar'),
        gralejdev('lib/batik-awt-util.jar'),
        gralejdev('lib/batik-svggen.jar'),
        gralejdev('lib/plugin.jar'),
        gralejdev('lib/jlibeps.jar'),
        gralejdev('lib/batik-util.jar')|Rest],Rest) :-
    environ('GRALEJ_TRUNK_DIR',TGDD),
    \+ TGDD = '',
    !.
gralej_classpath([trale_home('gralej/dist/gralej.jar'),
        trale_home('gralej/dist/lib/batik-awt-util.jar'),
        trale_home('gralej/dist/lib/batik-svggen.jar'),
        trale_home('gralej/dist/lib/batik-util.jar'),
        trale_home('gralej/dist/lib/jlibeps.jar'),
        trale_home('gralej/dist/lib/tomato.jar')|Rest],Rest).

get_jvm(JVM) :-
    jvm_store(JVM),
    !.
get_jvm(JVM) :-
    create_jvm(JVM),
    asserta(jvm_store(JVM)).

create_jvm(JVM) :-
    tralesld_classpath(Classpath),
    write('Using classpath: '),
    write(Classpath),
    nl,
    jasper_initialize([classpath(Classpath)/*,'-javaagent:/home/ke/workspace/kahina/lib/shiftone-jrat.jar'*/,'-Xmx1024m'],JVM).

get_instance(Instance) :-
    instance_store(Instance),
    !.
get_instance(Instance) :-
    create_instance(Instance),
    asserta(instance_store(Instance)).

create_instance(Instance) :-
    get_jvm(JVM),
    call_and_print_java_exception_info(
            jasper_new_object(JVM,'org/kahina/tralesld/TraleSLDInstance',init,init,LocalInstance)),
    jasper_create_global_ref(JVM,LocalInstance,Instance).
% instance stays around for the rest of the Prolog session, so we never have to
% delete the global reference

get_bridge(Bridge) :-
    bridge_store(Bridge),
    !.
% no lazy creation of bridge, start session explicitly

start_new_kahina_session :-
    get_jvm(JVM),
    clear_old_kahina_session(JVM),
    get_instance(Instance),
    call_and_print_java_exception_info(
            call_foreign_meta(JVM,start_new_session(Instance,Bridge))),
    jasper_create_global_ref(JVM,Bridge,GlobalBridge),
    assert(bridge_store(GlobalBridge)).

clear_old_kahina_session(JVM) :-
    retract(bridge_store(Bridge)),
    jasper_delete_global_ref(JVM,Bridge),
    fail. % delete all bridge references in a failure-driven loop
clear_old_kahina_session(_).
    
call_and_print_java_exception_info(Goal) :-
    catch(Goal,E,
            (print_java_exception_info(E),
             throw(E))).

print_java_exception_info(E) :-
    jvm_store(JVM),
    is_java_exception(JVM,E),
    !,
    print_exception_info(JVM,E).
print_java_exception_info(_).            

% ------------------------------------------------------------------------------
% EXCEPTION HANDLING
% ------------------------------------------------------------------------------

is_java_exception(_JVM, Thing) :- var(Thing), !, fail.
is_java_exception(_JVM, Thing) :-
   Thing = java_exception(_),      % misc error in Java/Prolog glue
   !.
is_java_exception(JVM, Thing) :-
   jasper_is_object(JVM, Thing),
   jasper_is_instance_of(JVM, Thing, 'java/lang/Throwable').

print_exception_info(_JVM, java_exception(Message)) :- !,
   format(user_error, '~NJasper exception: ~w~n', [Message]).
print_exception_info(JVM, Excp) :-
   /*
   // Approximate Java code
   {
      String messageChars = excp.getMessage();
   }
   */
   jasper_call(JVM,
               method('java/lang/Throwable', 'getMessage', [instance]),
               get_message(+object('java/lang/Throwable'), [-chars]),
               get_message(Excp, MessageChars)),
   /* // Approximate Java code
   {
      StringWriter stringWriter = new StringWriter();
      PrintWriter printWriter = new PrintWriter(stringWriter);
      excp.printStackTrace(printWriter);
      printWriter.close();
      stackTraceChars = StringWriter.toString();
   }
   */
   jasper_new_object(JVM, 'java/io/StringWriter',
                     init, init, StringWriter),
   jasper_new_object(JVM, 'java/io/PrintWriter',
                     init(+object('java/io/Writer')),
                     init(StringWriter), PrintWriter),
   jasper_call(JVM,
               method('java/lang/Throwable', 'printStackTrace', [instance]),
               print_stack_trace(+object('java/lang/Throwable'),
                                 +object('java/io/PrintWriter')),
               print_stack_trace(Excp, PrintWriter)),
   jasper_call(JVM,
               method('java/io/PrintWriter','close',[instance]),
               close(+object('java/io/PrintWriter')),
               close(PrintWriter)),
   jasper_call(JVM,
               method('java/io/StringWriter','toString',[instance]),
               to_string(+object('java/io/StringWriter'),[-chars]),
               to_string(StringWriter, StackTraceChars)),
   jasper_delete_local_ref(JVM, PrintWriter),
   jasper_delete_local_ref(JVM, StringWriter),
   %% ! exceptions are thrown as global references
   jasper_delete_global_ref(JVM, Excp),
   format(user_error, '~NJava Exception: ~s\nStackTrace: ~s~n',
          [MessageChars, StackTraceChars]).
